/**************************************************************************
4-gon functions

This module contains:

void write_region_data(region_data& region,ostream& os, string prefix="")
void write_chord_data(chord_data& pattern,ostream& os, string prefix="")
bool reset_interior_choice(region_data& region, vector<int>& valency, int max_valency)
bool increment_interior_choice(region_data& region, vector<int>& valency, int max_valency)
void enumerate_interior_choices(chord_data& pattern, vector<int>& valency, int max_valency,result_data& search_results)
void enumerate_interior_vertices(chord_data& pattern, int num_interior_vertices, vector<int>& region_allocation, int remaining_vertices, int first_region, vector<int>& valency, int max_valency,result_data& search_results)
void add_interior_edges(matrix<int> incidence, chord_data& pattern, result_data& search_results)
bool planar_graph(matrix<int>& incidence,int num_boundary_vertices,	vector<list<int> >& petal_list)
bool valid_4_gon_graph(matrix<int>& incidence, vector<list<int> >& petal_list, int num_boundary_vertices, int num_crossings, int solution_count, string& gauss_code)
string doodle_gauss_code(vector<pair<int,int> >& edge_pair,matrix<int>& cycle_vertices, int num_edges, int num_boundary_edges)
int succeeding_edge(matrix<int>& region_edges, int num_4_gons_in_disc, int edge, int& region)
bool four_connected (string gauss_code)

**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>

#include <pthread.h>

/******************** Global Variables **********************/
extern ofstream    debug;
extern ofstream    output;
extern bool	DRAW_GRAPHS;
extern bool	TEST_MODE;
extern bool	INTERIOR_CHOICES;
extern bool	sigusr1_request;

#include <4-gon.h>

#include <ogdf/fileformats/GraphIO.h>
#include <ogdf/orthogonal/OrthoLayout.h>
#include <ogdf/planarity/EmbedderMinDepthMaxFaceLayers.h>
#include <ogdf/planarity/PlanarSubgraphFast.h>
#include <ogdf/planarity/PlanarizationLayout.h>
#include <ogdf/planarity/SubgraphPlanarizer.h>
#include <ogdf/planarity/VariableEmbeddingInserter.h>

namespace util
{
	string itos(long n);
}
 
using namespace ogdf;

bool planar_graph(matrix<int>& incidence,int num_boundary_vertices,	vector<list<int> >& petal_list);
bool valid_4_gon_graph(matrix<int>& incidence, vector<list<int> >& petal_list, int num_boundary_vertices, int num_crossings, int solution_count, string& gauss_code);
float edge_angle(float v_x, float v_y, float e_x, float e_y);
string doodle_gauss_code(vector<pair<int,int> >& edge_pair,matrix<int>& cycle_vertices, int num_edges, int num_boundary_edges);
int succeeding_edge(matrix<int>& region_edges, int num_4_gons_in_disc, int start, int& region);
void add_interior_edges(matrix<int> incidence, chord_data& pattern, result_data& search_results);
bool gauss_to_peer_code(generic_code_data gauss_code_data, generic_code_data& peer_code_data, bool optimal, vector<int>* gauss_crossing_map, bool evaluate_gauss_crossing_perm);

void write_region_data(region_data& region,ostream& os, string prefix="")
{
	os << prefix << "index = " << region.index << endl;
	os << prefix << "num_boundary_vertices = " << region.num_boundary_vertices << endl;
	os << prefix << "num_boundary_edges = " << region.num_boundary_edges << endl;
	os << prefix << "num_inside_vertices = " << region.num_inside_vertices << endl;
	os << prefix << "num_interior_vertices = " << region.num_interior_vertices << endl;
	os << prefix << "first_interior_vertex = " << region.first_interior_vertex << endl;
	os << prefix << "boundary_vertices: ";
	for (size_t i=0; i< region.boundary_vertices.size(); i++)
		os << region.boundary_vertices[i] << ' ';
	os << endl;
	os << prefix << "inside_vertices: ";
	for (size_t i=0; i< region.inside_vertices.size(); i++)
		os << region.inside_vertices[i] << ' ';
	os << endl;
	os << prefix << "interior_choice: ";
	for (size_t i=0; i< region.interior_choice.size(); i++)
		os << region.interior_choice[i] << ' ';
	os << endl;
	os << prefix << "max_interior_choice: ";
	for (size_t i=0; i< region.max_interior_choice.size(); i++)
		os << region.max_interior_choice[i] << ' ';
	os << endl;
}

void write_chord_data(chord_data& pattern,ostream& os, string prefix="")
{
	os << prefix << "chord_data: " << endl;
	os << prefix << "  number of chords = " << pattern.chords.size() << ": ";
	list<pair<int,int> >::iterator next_cptr = pattern.chords.begin();
	while (next_cptr != pattern.chords.end())
	{
		os << '(' << next_cptr->first << ',' << next_cptr->second << ") ";
		next_cptr++;
	}
	os << endl;
	os << prefix << "  incidence matrix: " << endl;
	print (pattern.incidence,os, 2,prefix+" ");	
	os << prefix << "  number of chord regions = " << pattern.chords.size()+1 << endl;
	os << prefix << "  number of remaining interior edges required = " << pattern.num_edges_to_add << endl;
	os << prefix << "  total number of inside vertices = " << pattern.total_num_inside_vertices << endl;
	os << prefix << "  min_num_interior_vertices: ";
	for (size_t i=0; i< pattern.min_num_interior_vertices.size(); i++)
		os << pattern.min_num_interior_vertices[i] << ' ';
	os << endl;
	os << prefix << "  regions: " << endl;
	for (size_t i=0; i< pattern.region.size(); i++)
		write_region_data(pattern.region[i],os, prefix+"    ");
	os << prefix << "  region_map: ";
	for (size_t i=0; i< pattern.region_map.size(); i++)
		os << pattern.region_map[i] << ' ';
	os << endl;
}

/* reset_interior_choice initializes the region's interior_choice and checks that we can join all the 2-valent inside boundary vertices 
   of the region to interior vertices without exceeding the maximum allowed valencey.  
*/
bool reset_interior_choice(region_data& region, vector<int>& valency, int max_valency)
{

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "reset_interior_choice: presented with region_data" << endl;
    write_region_data(region,debug,"reset_interior_choice:   ");
}
	bool realizable = true;

	for (int i=0; i< region.num_inside_vertices; i++)
	{
		int interior_vertex = i/max_valency;
		if (interior_vertex == region.num_interior_vertices)
		{
			output << "Doodle code is unrealizable.  Cannot join inside vertices to interior vertices without exceeding maximum allowed valency" << endl;
			realizable = false;
			break;
		}
		else
		{
			region.interior_choice[i] = region.first_interior_vertex+interior_vertex;
			valency[region.first_interior_vertex+interior_vertex]++;
			
			/* set the max_interior_choice so that we do not effectively renumber interior vertices.  If there are
			   more interior vertices available, we don't need to consider each one in turn, we just need to select
			   a different one to those already joined to an inside vertex
			*/
			if (i < region.num_interior_vertices)
				region.max_interior_choice[i] = region.first_interior_vertex+i;
		}
	}

if (debug_control::DEBUG >= debug_control::SUMMARY && realizable)
{
	debug << "reset_interior_choice: reset region_data" << endl;
    write_region_data(region,debug,"reset_interior_choice:   ");
	debug << "reset_interior_choice:   valency: ";
	for (size_t i=0; i< valency.size(); i++)
		debug << valency[i] << ' ';
	debug << endl;
}	

	return realizable;
}

/* increment_interior_choice enumarates all possible interior choices without effectively renumbering interior vertices.  It emulates the increase of an n-digit 
   base-k integer, where the maximum value for each digit is set to avoid renumbering.  Having incremented a particular 'digit' we reset subsequent 'digits' to
   their 'zero state', which is to assign them to the smallest available interior vertex.
*/
bool increment_interior_choice(region_data& region, vector<int>& valency, int max_valency)
{	
	/* set the new max_interior_choice to avoid effectively re-numbering the interior vertices.  If there are unattached interior vertices and we are going to increment 
	   interior_choice so that an unattached vertex becomes joined to the boundary, then we do not need to select each available unattached interior vertex in turn, since
	   doing so effectively renumbers the unattached interior vertices.  We would get the same graph layouts with the interior vertices renumbered.  Therefore, we can set 
	   the maximum interior choice for boundary vertex i (well, 2*i but the ith element if interior_choice) to be one more than the current value of interior_choice[i-1].
	   Note however, that if we have a current interior choice such as 8 9 8 9, with three interior vertices 8,9,10 available, then we have to allow the next interior_choice
	   8 9 8 10, so the vector max_interior_choice must always be monotonically increasing.
	*/		
	for (int i=1; i<region.num_inside_vertices; i++)
		region.max_interior_choice[i] = max(region.max_interior_choice[i-1],min(region.interior_choice[i-1]+1,region.first_interior_vertex+region.num_interior_vertices-1));

		/* increment interior_choice for the even numbered vertices in the disc boundary */

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "increment_interior_choice: incrementing interior_choice for region " << region.index << endl;
if (debug_control::DEBUG >= debug_control::DETAIL)
    write_region_data(region,debug,"increment_interior_choice:   ");
	
	bool incremented = false;
			
	for (int i = region.num_inside_vertices-1; i>0; i--) // always fix the first inside vertex of the region to the first interior vertex
	{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "increment_interior_choice:   inside vertex " << i << ", interior_choice[" << i << "] = " << region.interior_choice[i] << endl;
	
		if (region.interior_choice[i] < region.max_interior_choice[i])
		{
			for (int j=i; j<region.num_inside_vertices; j++)
				valency[region.interior_choice[j]]--;

			/* we could already have max_valency edges at interior_choice[i]+1 in amongst the vertices 0,...,i, if not, this is a valid next choice 
			   for the ith even numbered boundary vertex
			*/
			if(valency[region.interior_choice[i]+1] < max_valency)
			{
				
				region.interior_choice[i]++;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "increment_interior_choice:   new interior_choice[" << i << "] = " << region.interior_choice[i] << endl;
	
				valency[region.interior_choice[i]]++;
						
				/* reset the interior choice of all inside vertices after the i-th one to the smallest available interior vertex.
				   We know there are enough interior vertices to join all the even edges of the disc boundary to an interior verrtex 
				*/
				for (int j=0; j< region.num_inside_vertices-i-1; j++)
				{
					int interior_vertex = 0;
					while (valency[region.first_interior_vertex+interior_vertex] >= max_valency)
						interior_vertex++;
						
					region.interior_choice[i+1+j] = region.first_interior_vertex+interior_vertex;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "increment_interior_choice:   new interior_choice[" << i+1+j << "] = " << region.interior_choice[i+1+j] << endl;
					valency[region.first_interior_vertex+interior_vertex]++;
				}
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "increment_interior_choice:   updated valency: ";
	for (size_t i=0; i< valency.size(); i++)
		debug << valency[i] << ' ';
	debug << endl;
}
				incremented = true;
				break;
			}
			else
			{
				/* restore the valency counts ready for testing the preceeding inside vertex's interior choice.*/
				for (int j=i; j<region.num_inside_vertices; j++)
					valency[region.interior_choice[j]]++;
					
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "increment_interior_choice:   restored valency: ";
	for (size_t i=0; i< valency.size(); i++)
		debug << valency[i] << ' ';
	debug << endl;
}
			}
		}
	}

	if (incremented)
	{	
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "increment_interior_choice: incremented interior_choice for region" << endl;
    write_region_data(region,debug,"increment_interior_choice:   ");
}
	}
	else
	{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "increment_interior_choice: region " << region.index << " could not be incremented" << endl;
}
	}	
	
	return incremented;
}

/* Within each chord region we must have an edge joing each inside boundary vertex to an interior vertex.  The functions reset_interior_choice and 
   increment_interior_choice systematically enumerate each possibility for a given region, whilst enumerate_interior_choices enumerates all of the 
   combinations of choice for each region that has interior vertices.
*/
void enumerate_interior_choices(chord_data& pattern, vector<int>& valency, int max_valency,result_data& search_results)
{
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "enumerate_interior_choices: presented with chord_data:" << endl;
	write_chord_data(pattern, debug, "enumerate_interior_choices: ");
}	
	int num_chord_regions = pattern.chords.size()+1;
	
	bool realizable = true;
			
	for (int i=0; i< num_chord_regions; i++)
	{
		realizable = reset_interior_choice(pattern.region[i], valency, max_valency);
		if (!realizable)
			break;
	}
	
	if(realizable)
	{
		bool complete = false;
		do
		{
			if (INTERIOR_CHOICES)
			{
				for (int i=0; i< num_chord_regions; i++)
				{
					if (pattern.region[i].num_inside_vertices > 0)
					{
						cout << "  region " << i << ": ";
						
						for (int j=0; j< pattern.region[i].num_inside_vertices; j++)
							cout << setw(3) << pattern.region[i].interior_choice[j];
	
						cout << ", max: ";
						for (int j=0; j< pattern.region[i].num_inside_vertices; j++)
							cout << setw(3) << pattern.region[i].max_interior_choice[j];
							
						cout << "  ";
					}
				}
				cout << endl;
			}
			else
			{
			
				matrix<int> local_incidence = pattern.incidence;
				
				/* assign interior choices to local incidence */
				for (int i=0; i< num_chord_regions; i++)
				{
					for (int j=0; j< pattern.region[i].num_inside_vertices; j++)
					{
						local_incidence[pattern.region[i].inside_vertices[j]][pattern.region[i].interior_choice[j]] = 1;
						local_incidence[pattern.region[i].interior_choice[j]][pattern.region[i].inside_vertices[j]] = 1;
					}
				}
						
				if (TEST_MODE)
				{
					cout << "do you want to assign an interior choice? y/n: ";
					char ch;
					cin >> ch;
		
					if (ch == 'y')
					{
						for (int i=0; i< num_chord_regions; i++)
						{
							cout << "interior choice for regions " << i << ": (" << pattern.region[i].num_inside_vertices << "): ";

if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << "TEST_MODE: specified interior choice for regions " << i << ": (" << pattern.region[i].num_inside_vertices << "): ";

							for (int j=0; j< pattern.region[i].num_inside_vertices; j++)
							{
								cin >>  pattern.region[i].interior_choice[j];

if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << pattern.region[i].interior_choice[j] << ' ';
							}

if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << endl;							
						}
						
						for (int i=0; i< num_chord_regions; i++)
						{
							for (int j=0; j< pattern.region[i].num_inside_vertices; j++)
							{
								local_incidence[pattern.region[i].inside_vertices[j]][pattern.region[i].interior_choice[j]] = 1;
								local_incidence[pattern.region[i].interior_choice[j]][pattern.region[i].inside_vertices[j]] = 1;
							}
						}						
					}
				}
				
				/* add interior edges */
//				add_interior_edges(local_incidence, pattern.num_edges_to_add, pattern.total_num_inside_vertices, search_results);				
				add_interior_edges(local_incidence, pattern, search_results);				
			}
			
			/* increment interior choices */
			complete = true;
			for (int i=num_chord_regions-1; i>=0; i--)
			{
				if (pattern.region[i].num_interior_vertices == 0)
					continue;
					
				if (increment_interior_choice(pattern.region[i], valency, max_valency))
				{
					complete = false;
					for (int j=i+1; j<num_chord_regions; j++)
					{
						
						for (int k=0; k< pattern.region[j].num_interior_vertices; k++)
							valency[pattern.region[j].first_interior_vertex+k] = 0;
						
						reset_interior_choice(pattern.region[j], valency, max_valency);
					}
					break;
				}
			}			
			
			if (TEST_MODE)
				break;
				
		} while (!complete);
	}
}

/* enumerate_interior_vertices adds the remaining interior vertices to the minimum number of interior vertices over the chord regions
   in all possible ways.  This ammounts to distributing k objects amongst n buckets in all possible ways.  It is tempting to consider 
   each object having n possible locations, a total of n^k combinations, which can be enumerated by incrementing a k-digit base-n number 
   from 0 to n^k-1.  However, this approach duplicates some allocations, since the k objects are unordered: if we write b^i_j to indicate
   that the ith object is allocated to bucket j, then the allocation b^1_0,..,b^{k-1}_0,b^k_1 is the same as b^1_0,...b^{k-1}_1,b^k_0 and 
   b^1_0,...,b^{k-1}_2,b^k_1 is the same as b^1_0,...,b^{k-1}_1,b^k_2 etc.
   
   The function therefore works recursively, sequentially placing m=0,...,k objects in the first bucket and distributing k-m objects
   amongst n-1 buckets.  If n-1=1, all k-m objects go into the last bucket.
*/
void enumerate_interior_vertices(chord_data& pattern, int num_interior_vertices, vector<int>& region_allocation, int remaining_vertices, int first_region, vector<int>& valency, int max_valency,result_data& search_results)
{
	int num_chord_regions = pattern.chords.size()+1;
	int num_boundary_vertices = search_results.num_boundary_vertices;
	
	if (TEST_MODE || remaining_vertices == 0 || first_region == num_chord_regions-1) // first_region is the last region in the pattern
	{
		if (remaining_vertices > 0)
		{
			region_allocation[num_chord_regions-1] = remaining_vertices;
		}
		else
		{
			for (int i=first_region; i< num_chord_regions; i++)
				region_allocation[i]=0;
		}
			
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "enumerate_interior_vertices: remaining vertex region_allocation: " ;
	for (int i=0; i< num_chord_regions; i++)
		debug << region_allocation[i] << ' ';
	debug << endl;
}	

		vector<int> interior_vertex_allocation = pattern.min_num_interior_vertices;
		
		if (TEST_MODE)
		{
			cout << "do you want to change the interior_vertex allocation? y/n: ";
			char ch;
			cin >> ch;

			if (ch == 'y')
			{
			
				cout << "number of interior vertices for regions: ";
				for (int i=0; i< num_chord_regions; i++)
					cin >> interior_vertex_allocation[i];

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "TEST_MODE: specified number of interior vertices for regions: ";
	for (int i=0; i< num_chord_regions; i++)
		debug << interior_vertex_allocation[i] << ' ';
	debug << endl;
}
					
				int new_num_vertices_in_dual_graph = num_boundary_vertices;
				for (int i=0; i< num_chord_regions; i++)
					new_num_vertices_in_dual_graph += interior_vertex_allocation[i];
					
				cout << "total number of vertices in dual graph = " << new_num_vertices_in_dual_graph << endl;
				valency = vector<int>(new_num_vertices_in_dual_graph);
				
				cout << "default max_valency = " << max_valency << ", do you want to change it? y/n: ";
				cin >> ch;
				
				if (ch == 'y')
				{
					cout << "new max_valency: ";
					cin >> max_valency;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "TEST_MODE: specified new max_valency " << max_valency << endl;
				}
			}
			else
			{
				for (int i=0; i< num_chord_regions; i++)
					interior_vertex_allocation[i] += region_allocation[i];
			}
		}
		else
		{
			for (int i=0; i< num_chord_regions; i++)
				interior_vertex_allocation[i] += region_allocation[i];
		}
			
		int cumultive_interior_vertex_index = num_boundary_vertices;
		
		for (int i=0; i< num_chord_regions; i++)
		{
			pattern.region[i].num_interior_vertices = interior_vertex_allocation[i];
			pattern.region[i].first_interior_vertex = cumultive_interior_vertex_index;
			cumultive_interior_vertex_index += pattern.region[i].num_interior_vertices;
			pattern.region[i].max_interior_choice = vector<int>(pattern.region[i].num_inside_vertices, pattern.region[i].first_interior_vertex+pattern.region[i].num_interior_vertices-1);
		}
			
		if (INTERIOR_CHOICES)
		{
			cout << "  interior vertex allocation ";
			for (int i=0; i< num_chord_regions; i++)
				cout << interior_vertex_allocation[i] << ' ';
			cout << endl;			
		}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "enumerate_interior_vertices: interior vertex allocation ";
	for (int i=0; i< num_chord_regions; i++)
		debug << interior_vertex_allocation[i] << ' ';
	debug << endl;			
}

	
		/* reset the valency of any earlier interior vertex allocations */
		for (size_t i=0; i< valency.size(); i++)
			valency[i] = 0;
		
		enumerate_interior_choices(pattern,valency,max_valency,search_results);
	}
	else
	{
		for (int i=0; i<= remaining_vertices; i++)
		{
			/* Only recurse if the allocation remains valid.  If first_region has no inside vertices and four edges, the min_num_interior_vertices is zero for this region.
			   In this case, the region allocation must be >= 4
			*/ 
			   
			if (pattern.region[first_region].num_boundary_vertices == 4 && pattern.region[first_region].num_inside_vertices == 0 && i < 4)
			{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "enumerate_interior_vertices: not allocating fewer than four interior vertices to region " << first_region << " of pattern:" << endl;
	write_chord_data(pattern, debug, "enumerate_interior_vertices: ");
}	
				i = 3; // jump over those cases we're not going to consider
				region_allocation[first_region] = 0;
				enumerate_interior_vertices(pattern,num_interior_vertices,region_allocation, remaining_vertices, first_region+1,valency, max_valency, search_results);
			}
			else
			{			  
				region_allocation[first_region] = i;
				enumerate_interior_vertices(pattern,num_interior_vertices,region_allocation, remaining_vertices-i, first_region+1,valency, max_valency, search_results);
			}
		}
	} 
}

/* for i >=3 doodle_code[i] indicates the maximum number of i-valent vertices permitted in the interior graph */
//void add_interior_edges(matrix<int> incidence, int num_edges_to_add, int num_inside_vertex_edges, result_data& search_results)
void add_interior_edges(matrix<int> incidence, chord_data& pattern, result_data& search_results)
{
	int num_edges_to_add = pattern.num_edges_to_add;
	int num_inside_vertex_edges = pattern.total_num_inside_vertices;
	bool no_block_increment = false; // used by TEST_MODE
	int num_vertices = incidence.numrows();
	int num_boundary_vertices = search_results.num_boundary_vertices;
	int num_interior_vertices = num_vertices - num_boundary_vertices;

	int max_valency = static_cast<int>(search_results.doodle_code.size())-1;
	while (search_results.doodle_code[max_valency] == 0)
		max_valency--;
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	print (incidence,debug, 2,"add_interior_edges: ");
	debug << "add_interior_edges: num_edges_to_add = " << num_edges_to_add << ", num_inside_vertex_edges = " << num_inside_vertex_edges << endl;
	debug << "add_interior_edges: num_interior_vertices = " << num_interior_vertices << ", num_boundary_vertices = " << num_boundary_vertices << endl;

}
	vector<int> valency(num_vertices);
	for (int i=0; i< num_vertices; i++)
	{
		for (int j=0; j< num_vertices; j++)
			valency[i] += incidence[i][j];
	}
	
	for (int i=0; i<num_boundary_vertices/2; i++)
		valency[2*i+1]++; // edge to the infinite vertex.

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: current vertex valency: ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
}	

	vector<int> valency_count(max_valency+1); // we want to use the valency as an index into valency_count, so +1
	for (int i=0; i< num_vertices; i++)
		valency_count[valency[i]]++;
		
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: initial valency_count: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}	
	
	/* We have to place num_edges_to_add edges in incidence to the intersection of the right side of the leading diagonal and the right-most 
	   num_interior_vertices columns.  We have already added num_inside_vertex_edges edges in this region of incidence for the inside vertices 
	   of our interior choice, so we start by calculating the number of places left and establish a map from the integers 0,...,num_places-1 
	   to the location of the free places.
	*/
	int num_places = (num_boundary_vertices * num_interior_vertices - num_inside_vertex_edges) + (num_interior_vertices * (num_interior_vertices -1) / 2);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: num_places to fill = " << num_places << endl;

	vector<pair<int,int> > place_map(num_places);
	int index = 0;
	for (int i=0; i<num_boundary_vertices; i++)
	{
		for (int j=num_boundary_vertices; j< num_vertices; j++)
		{
			if (incidence[i][j] == 0)
			{
				place_map[index] = {i,j};
				index++;
			}
		}
	}
	
	for (int i=num_boundary_vertices; i< num_vertices; i++)
	{
		for (int j=i+1; j< num_vertices; j++)
		{
			if (incidence[i][j] == 0)
			{
				place_map[index] = {i,j};
				index++;
			}
		}
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: place_map:" << endl;
	for (int i=0; i< num_places; i++)
	debug << "add_interior_edges:   " << setw(2) << i << ": " << place_map[i].first << ' ' << place_map[i].second << endl;
}
	/* work through all options for filling the available places */
	vector<int> block(num_edges_to_add);
	vector<int> max(num_edges_to_add);
	for (int i=0; i< num_edges_to_add; i++)
	{
		block[i] = i;
		max[i] = num_places-num_edges_to_add+i;
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: block max: ";
	for (int i=0;i< num_edges_to_add; i++)
		debug << max[i] << ' ';
	debug << endl;
}

if (TEST_MODE)
{
	cout << "test block? y/n/r ";
	char ch;
	cin >> ch;

	if (ch == 'y')
		no_block_increment = true;
	
	if (ch == 'y' || ch == 'r')
	{
		cout << "enter test block (" << num_edges_to_add << "): ";
		for (int i=0; i< num_edges_to_add; i++)
			cin >> block[i];

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "TEST_MODE: specified test block: ";
	for (int i=0; i< num_edges_to_add; i++)
		debug << block[i] << ' ';
	debug << endl;
}

	}
}	

	bool finished = false;
	do // test block
	{		
		if (sigusr1_request)
		{						
//			vector<int> original_doodle_code = doodle_code;
			vector<int> original_doodle_code = search_results.doodle_code;
			original_doodle_code[num_boundary_vertices/2]++;
			
			clog << "doodle code: ";
			for (size_t i=0; i < original_doodle_code.size(); i++)
				clog << original_doodle_code[i] << ' ';
			clog << endl;
//			clog << "interior_choice: ";
//			for (size_t i=0; i< interior_choice.size(); i++)
//				clog << setw(3) << interior_choice[i];
//			clog << endl;
			clog << "num_boundary_vertices = " << num_boundary_vertices << endl;
			clog << "num_interior_vertices = " << num_interior_vertices << endl;
			
			if (pattern.chords.size())
			{
				clog << "boundary chords: " ;
				list<pair<int,int> >::iterator next_cptr = pattern.chords.begin();
				while (next_cptr != pattern.chords.end())
				{
					clog << '(' << next_cptr->first << ',' << next_cptr->second << ") ";
					next_cptr++;
				}
				clog << endl;
				
				clog << "interior vertex allocation: ";
				for (size_t i=0; i< pattern.chords.size()+1; i++)
					clog << pattern.region[i].num_interior_vertices << ' ';
				clog << endl;
			}
			clog << "interior choice: " << endl;

			for (size_t i=0; i< pattern.chords.size()+1; i++)
			{
				if (pattern.region[i].num_inside_vertices > 0)
				{
					clog << "  region " << i << ": ";
					
					for (int j=0; j< pattern.region[i].num_inside_vertices; j++)
						clog << setw(3) << pattern.region[i].interior_choice[j];
						
					clog << "  ";
				}
			}
			clog << endl;

			clog << "block size = " << num_edges_to_add << endl;
			clog << "num_places for test block = " << num_places << endl;
			clog << "test block: ";
			for (int i=0;i< num_edges_to_add; i++)
				clog << block[i] << ' ';
			clog << endl;
			
			sigusr1_request = false;
		}
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: test block ";
	for (int i=0;i< num_edges_to_add; i++)
		debug << block[i] << ' ';
	debug << endl;
}
		
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: current vertex valency : ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges: current valency_count: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}

		bool skip_block = false;
		
		for (int i=0; i< num_edges_to_add; i++)
		{	
			int vr=place_map[block[i]].first;  // vertex row, this will be the row in incidence in which we set the next 1
			int vc=place_map[block[i]].second;  // vertex column, the column in incidence where we set the next 1
			
			/* The regions of the dual graph have to be four-sided, so if u is joined by an edge to v and v is joinred to w, 
			   then u cannot be joined to w.  Here we are proposing to vertex vr to vc, so first check that rows vr and vc
			   do not already have a common set column entry, indicting that vertex vr and vc are both already joined to a
			   common vertex.
			*/
			bool common_vertex = false;
			for (int j=0; j< num_vertices; j++)
			{
				if (incidence[vr][j] == 1 && incidence[vc][j] == 1)
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: vertices " << vr << " and " << vc << " (block[" << i << "]) are joined to common vertex = " << j << endl;
					common_vertex = true;
					break;
				}
			}

			bool valency_limit_reached = false;
			if (valency[vr] == max_valency || valency[vc] == max_valency)
			{				
				valency_limit_reached = true;
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: cannot add another edge to vertex " << (valency[vr] == max_valency? vr:vc) << " (block[" << i << "]) without exceeding max_valency = " << max_valency << endl;
			}
			
			if (common_vertex || valency_limit_reached)
			{				

				/* Un-set any edges we've already set in incidence and adjust valency and 
				   valency _count accordingly, then advance the block if we can. 
				*/
				for (int j=0; j< i; j++)
				{
					int vvr = place_map[block[j]].first;
					int vvc = place_map[block[j]].second;
					valency_count[valency[vvr]]--;
					valency_count[valency[vvc]]--;
					incidence[vvr][vvc] = 0;
					incidence[vvc][vvr] = 0;
					valency[vvr]--;
					valency[vvc]--;
					valency_count[valency[vvr]]++;
					valency_count[valency[vvc]]++;
				}
					
				finished = true;
				for (int j=i; j>=0; j--)
				{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "add_interior_edges: j = " << j << "block[" << j << "] =  " << block[j] << " max[" << j << "] =  " << max[j] << endl;
					if (block[j] < max[j])
					{
						finished = false;
						block[j]++;
						for (int k=j+1; k<num_edges_to_add; k++)
							block[k] = block[k-1]+1;
						break;
					}
				}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{			
	if (finished)
	{
		debug << "add_interior_edges: cannot advance block, finished" << endl;
	}
	else
	{
		debug << "add_interior_edges: advanced block to ";
		for (int j=0;j< num_edges_to_add; j++)
			debug << block[j] << ' ';
		debug << endl;
		
		debug << "add_interior_edges: vertex valency reset to : ";
		for (int i=0; i< num_vertices; i++)
			debug << valency[i] << ' ';
		debug << endl;
		debug << "add_interior_edges: valency_count reset to: ";
		for (int i=0; i<= max_valency; i++)
			debug << valency_count[i] << ' ';
		debug << endl;
	}
}				
				skip_block = true;
				break;
			}
			else
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: set incidence [" << vr << "][" << vc << "] = 1, together with its transpose" << endl;

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "add_interior_edges:   valency[" << vr << "] = " << valency[vr] << ", valency[" << vc << "] = " << valency[vc] << endl;
	debug << "add_interior_edges:   valency_count[" << valency[vr] << "] = " << valency_count[valency[vr]] << ", valency_count[" << valency[vc] << "] = " << valency_count[valency[vc]] << endl;
}
				valency_count[valency[vr]]--;
				valency_count[valency[vc]]--;
				incidence[vr][vc] = 1;
				incidence[vc][vr] = 1;
				valency[vr]++;
				valency[vc]++;
				valency_count[valency[vr]]++;
				valency_count[valency[vc]]++;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges:   vertex valency stands at: ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges:   valency_count stands at: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}
	
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "add_interior_edges:   new valency[" << vr << "] = " << valency[vr] << ", new valency[" << vc << "] = " << valency[vc] << endl;
	debug << "add_interior_edges:   new valency_count[" << valency[vr] << "] = " << valency_count[valency[vr]] << ", new valency_count[" << valency[vc] << "] = " << valency_count[valency[vc]] << endl;
}

			}
		}

		if (skip_block)
		{
			continue; // already reset incidence and incremented block
		}
		else
		{
			/* test incidence and valency */
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: test vertex valency: ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges: test valency_count: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}
			bool valid = true;

			/* Reject any block that leaves vertices with valency less than 3.  Advance block past these cases where possible.  All
			   vertices up to num_boundary_vertices have valency at least 3, so only need check after that.
			*/
	
			for (int i=num_vertices-1; i >= num_boundary_vertices; i--)
			{
				if (valency[i] < 3)
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: vertex " << i << " has valency less than 3, reject block" << endl;					

					valid = false;
					
					/* look for the greatest block element on or before row i, this is the edge we're going to advance */
					int block_index = -1;
					for (int j=num_edges_to_add-1; j>=0 ; j--)
					{
						if (place_map[block[j]].first <= i)
						{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   block index " << j << ", element " << block[j] << " is the largest index contributing an edge before row " << i << endl;					
	
							block_index = j;
							break;
						}
					}
					
					if (block_index == -1)
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   no block index contributing an edge before row " << i << ", finished" << endl;											
						finished = true;
						break;
					}
					else
					{					
						/* advance the block starting at block_index, re-setting incidence and valency and valency_count */
	
						for (int j=0; j< num_edges_to_add; j++)
						{
							int vvr = place_map[block[j]].first;
							int vvc = place_map[block[j]].second;
							valency_count[valency[vvr]]--;
							valency_count[valency[vvc]]--;
							incidence[vvr][vvc] = 0;
							incidence[vvc][vvr] = 0;
							valency[vvr]--;
							valency[vvc]--;
							valency_count[valency[vvr]]++;
							valency_count[valency[vvc]]++;
						}
	
						/* adjust block */
						finished = true;
						for (int i=block_index; i>=0; i--)
						{
							if (block[i] < max[i])
							{
								finished = false;
								block[i]++;
								for (int j=i+1; j<num_edges_to_add; j++)
									block[j] = block[j-1]+1;
								break;
							}
						}
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{			
	debug << "add_interior_edges: vertex valency reset to : ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges: valency_count reset to: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}				
						break;					
					}
				}
			}

			if (!valid)
				continue;


			for (int i=3; i<= max_valency; i++)
			{
				/* search_results.doodle_code[i] correctly identifies the number of vertices in the dual graph of valency i, since we
				   subracted one from the maximum i-gon count in the doodle code to account for the infinite region.
				*/
				if (valency_count[i] != search_results.doodle_code[i])
				{
					valid = false;
					break;
				}
			}
			
			if (valid)
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: valency_count correct, block may define a 4-gon graph" << endl;					
		
				vector<list<int> > petal_list(num_vertices+1);

				if (planar_graph(incidence,num_boundary_vertices,petal_list))
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: interior graph is planar" << endl;					

					string gauss_code;
					if (valid_4_gon_graph(incidence, petal_list, num_boundary_vertices, num_vertices-1,search_results.solution_count,gauss_code))
					{
						search_results.solution_count++;			
						
						/* check whether this soltion is distinct from the others on doodle_list, we check twice, once with 
						   the gauss_code provided and once mirror image
						*/
						bool distinct = true;
						string test_string = gauss_code;
						
						for (int i=0; i < 2; i++)
						{
							if (i == 1)
							{
								for (size_t j=0; j< test_string.length(); j++)
								{
									if (test_string[j] == 'L')
										test_string[j] = 'R';
									else if (test_string[j] == 'R')
										test_string[j] = 'L';
								}
								
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   mirror image of gauss code " << test_string << endl;

								ofstream tmp("tmpfile");
								tmp << "[raw-output,silent,gauss,ulpgd]" << endl;
								tmp << test_string << endl;
								
								system("braid tmpfile");
								ifstream braid("braid.out");
								
								if (!getline(braid, test_string))
								{
									cerr << "\nError! Braid programe failed to calculate left_preferred Gauss data." << endl;
									exit(0);
								}
								braid.close();
								
								test_string += gauss_code.substr(gauss_code.find('/'));
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   left preferred mirror image of gauss code " << test_string << endl;
								
							}
							
							list<string>::iterator sptr = search_results.doodle_list.begin();
							while (sptr != search_results.doodle_list.end())
							{
								if (*sptr == test_string)
								{
									distinct = false;
									break;
								}
								
								sptr++;
							}
							if (!distinct)
								break;
						}
						
						if (distinct)
						{
							search_results.doodle_list.push_back(gauss_code);													
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   gauss code is distinct" << endl;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "add_interior_edges:   found distinct gauss code " << gauss_code << endl;
						}
						else
						{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   gauss code is not distinct" << endl;
						}
					}
					else
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: interior graph is not a 4-gon graph" << endl;					
					}
				}
				else
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: interior graph is non-planar" << endl;					
				}										
			}
			else
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: incorrect valency_count, block does not define a 4-gon graph" << endl;					
			}
		}
	
		if (no_block_increment)
			break;
			
		/* re-set incidence, valency and valency_count for next test */
		for (int i=0; i< num_edges_to_add; i++)
		{
			int vvr = place_map[block[i]].first;
			int vvc = place_map[block[i]].second;
			valency_count[valency[vvr]]--;
			valency_count[valency[vvc]]--;
			incidence[vvr][vvc] = 0;
			incidence[vvc][vvr] = 0;
			valency[vvr]--;
			valency[vvc]--;
			valency_count[valency[vvr]]++;
			valency_count[valency[vvc]]++;
		}
		
		// increment block 
		finished = true;
		for (int i=num_edges_to_add-1; i>=0; i--)
		{
			if (block[i] < max[i])
			{
				finished = false;
				block[i]++;
				for (int j=i+1; j<num_edges_to_add; j++)
					block[j] = block[j-1]+1;
				break;
			}
		}
		
	} while (!finished);	
}


/* planar_graph uses the OGDF library function planarEmbed to decide whether the the dual graph determined by incidence is planar.  If it is, then
   the function determines the list of petals for each vertex moving anti-clockwise around each vertex.
*/
bool planar_graph(matrix<int>& incidence,int num_boundary_vertices,	vector<list<int> >& petal_list)
{
	int num_vertices = incidence.numrows();

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	print (incidence,debug, 2,"planar_graph: ");
	
	Graph G;

	ostringstream incidence_DOT;

	/* first the disc boundary */
	incidence_DOT << "graph G {0";
	for (int i=0; i< num_boundary_vertices; i++)
		incidence_DOT << "--" << (i+1)%num_boundary_vertices;

	/* next any boundary chords */
	for (int i=0; i< num_boundary_vertices; i++)
	{
		for (int j=i+1; j< num_boundary_vertices; j++)
		{
			if (incidence[i][j] == 1  && j != (i+1)%num_boundary_vertices && j != (i-1+num_boundary_vertices)%num_boundary_vertices)
			{
				incidence_DOT << ' ' << i << "--" << j;
			}
		}
	}	

	/* now edges from the boundary to the interior vertices, including interior choice edges.  
	   The OGDF will assign an index to each vertex in the order that they appear in the DOT description, which means we cannot avoid vertices being re-numbered.  
	   If, for example edge i -- j is the first occurrence amongst the edges of vertex i but j> i+1, we might not be listing vertex i+1 until later, since we 
	   consider only incidence entries to the right of the leading diagonal.
	   
	   Therefore, we create a list of edge pairs before writing the DOT description so we can evaluate the vertex permutation introduced by the OGDF.  This allows
	   to reverse the permutation when we read the adjacency list of planar graphs and continue working with our own vertex numbering.
	*/	
	list<pair<int,int> > edge_list;
	for (int i=0; i<num_boundary_vertices; i++)
	{
		for (int j=num_boundary_vertices; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
			{
				edge_list.push_back(pair<int,int>(i,j));
			}
		}
	}

	/* edges between interior vertices */
	for (int i=num_boundary_vertices; i< num_vertices; i++)
	{
		for (int j=i+1; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
			{
				edge_list.push_back(pair<int,int>(i,j));
			}
		}
	}		

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "planar_graph: initial edge_list:";
	list<pair<int,int> >::iterator eptr = edge_list.begin();
	while (eptr != edge_list.end())
	{
		debug << ' ' << eptr->first << "--" << eptr->second;
		eptr++;
	}	
	debug << endl;
}
	
	vector<int> vertex_perm(num_vertices+1);
	for (int i=0; i<= num_vertices; i++)
		vertex_perm[i] = i;
	
	vector<bool> visited(num_vertices,false);
	for (int i=0; i< num_boundary_vertices; i++)
		visited[i] = true;
	
	visited[num_vertices] = true;
	
	list<pair<int,int> >::iterator eptr = edge_list.begin();
	int next_vertex = num_boundary_vertices;
	while (eptr != edge_list.end())
	{
		if(!visited[eptr->first])
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: vertex " << eptr->first << " becomes " << next_vertex << endl;
			vertex_perm[eptr->first] = next_vertex++;
			visited[eptr->first] = true;
		}
		
		if(!visited[eptr->second])
		{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: vertex " << eptr->second << " becomes " << next_vertex << endl;
			vertex_perm[eptr->second] = next_vertex++;
			visited[eptr->second] = true;
		}

		eptr++;
	}
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "planar_graph: vertex_perm: ";
	for (int i=0; i<= num_vertices; i++)
		debug << vertex_perm[i] << ' ';
	debug << endl;
}	
	vector<int> inverse_perm(num_vertices+1);
	for (int i=0; i<= num_vertices; i++)
		inverse_perm[vertex_perm[i]] = i;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "planar_graph: inverse_perm: ";
	for (int i=0; i<= num_vertices; i++)
		debug << inverse_perm[i] << ' ';
	debug << endl;
}	
	
	
	eptr = edge_list.begin();
	while (eptr != edge_list.end())
	{
		incidence_DOT << ' ' << eptr->first << "--" << eptr->second;
		eptr++;
	}			
			
	/* outside edges to infinite vertex */	
	for (int i=0; i<num_boundary_vertices/2; i++)
		incidence_DOT << ' ' << 2*i+1 << "--" << num_vertices;
	
	incidence_DOT << '}';

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: incidence_DOT = " << incidence_DOT.str() << endl;

	istringstream input(incidence_DOT.str());
	
	if (GraphIO::readDOT(G,input))
	{
 		bool planar = ogdf::planarEmbed(G);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: incidence " << (planar? "verified to " : "does not ") << " define a planar graph" << endl;
		
		if (planar)
		{
			/* The planar embedding produced by the OGDF could be the same as our canonical view of the dual disc, or it could be it's reflection in a line in the plane.
			   As a consequence, we do not know whether the petal list corresponding to the embedding is clockwise or anti-clockwise with respect to our canonical view.
			   We shall require an anti-clockwise order, since valid_4_gon_graph and doodle_gauss_code are written based on anti-clockwise flowers, which requires that
			   for vertex 1, vertex zero follows the infinite vertex in the petal list.  We can use this fact to check the lists produced by the OGDF and reverse them 
			   if necessary.
			*/
			for (node n: G.nodes)
			{								
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: node " << n->index() << " adjEntries: ";
	
				for(adjEntry a: n->adjEntries)
				{
					petal_list[inverse_perm[n->index()]].push_back(inverse_perm[a->twinNode()->index()]);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << a->twinNode()->index() << ' ';
				}
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << endl;
			}				
			
			/* check vertex 1 */
			bool anticlockwise = true;
			list<int>::iterator iptr = find(petal_list[1].begin(),petal_list[1].end(),num_vertices); // location of infinite vertex.

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: infinite vertex index in petal_list[1] = " << distance(iptr,petal_list[1].begin()) << endl;
			
			list<int>::iterator zptr = iptr;
			zptr++;
			if (zptr == petal_list[1].end())
				zptr = petal_list[1].begin();
				
			if (*zptr != 0)
				anticlockwise = false;
			
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: subsequent petal = " << *zptr <<", anticlockwise lists = " << anticlockwise << endl;

			if (!anticlockwise)
			{
				for (size_t i=0; i< petal_list.size(); i++)
					petal_list[i].reverse();
			}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "planar_graph: petal_list:" << endl;
	for (size_t i=0; i< petal_list.size(); i++)
	{
		debug << "planar_graph:   " << i << ": ";
		list<int>::iterator lptr = petal_list[i].begin();
		while(lptr != petal_list[i].end())
		{
			debug << *lptr << ' ';
			lptr++;
		}
		debug << endl;
	}				
}				
		}

		return planar;	
	}
	else
	{
		cerr << "planar_graph error! Failed to read graph from DOT string" << endl;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: Error! Failed to read graph from incidence_DOT"<< endl;
		exit(0);
	}
}


/* valid_4_gon_graph checks whether the interior graph determined by incidence has 4-gon regions.  Since the OGDF planar embedding can 
   permute the numbering of vertices, we work from the petal list rather than our original incidence matrix.
*/
bool valid_4_gon_graph(matrix<int>& incidence, vector<list<int> >& petal_list, int num_boundary_vertices, int num_crossings, int solution_count, string& gauss_code)
{
	bool valid = true;
	
	int num_vertices = incidence.numrows();
	
	vector<pair<int,int> > edge_pair(2*num_crossings);  // here for the call to doodle_gauss_code
		
	/* create edge_pair from incidence: first the disc boundary */
	int index = 0;
	for (int i=0; i< num_boundary_vertices; i++)
		edge_pair[index++] = pair<int,int>(i,(i+1)%num_boundary_vertices);

	/* next any boundary chords */
	for (int i=0; i< num_boundary_vertices; i++)
	{
		for (int j=i+1; j< num_boundary_vertices; j++)
		{
			if (incidence[i][j] == 1  && j != (i+1)%num_boundary_vertices && j != (i-1+num_boundary_vertices)%num_boundary_vertices)
				edge_pair[index++] = pair<int,int>(i,j);
		}
	}	

	/* now interior choice edges */	
	for (int i=0; i<num_boundary_vertices; i++)
	{
		for (int j=num_boundary_vertices; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
				edge_pair[index++] = pair<int,int>(i,j);
		}
	}
			
	/* edges between interior vertices */
	for (int i=num_boundary_vertices; i< num_vertices; i++)
	{
		for (int j=i+1; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
				edge_pair[index++] = pair<int,int>(i,j);
		}
	}		
			
	/* outside edges to infinite vertex */	
	for (int i=0; i<num_boundary_vertices/2; i++)
			edge_pair[index++] = pair<int,int>(2*i+1,num_vertices);
	
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: edge_pair: " << endl;
	for (int i = 0; i< 2* num_crossings; i++)
		debug << "valid_4_gon_graph:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}	
	
	int num_4_gons_in_disc = num_crossings-num_boundary_vertices/2;
	matrix<int> cycle_vertices(num_4_gons_in_disc,4);
	
		/* calculate the turning cycles on the disc determined by the flowers at each vertex.  We know the number of cycles in the 
		   disc, since num_4_gons_in_disc = num_crossings - num_edges_in_infinite_cycle = num_crossings-num_boundary_vertices/2.
		   We will only complete the record cycles when they all have length 4.
		   
		   We calculate right turning cycles, since this allows us to move anticlockwise around the vertices on the disc, the same
		   order in which the petals are stored.  Moreover, always turning right means that for an interior edge joining u to v, we
		   traverse the edge in the direction u to v and v to u, so we can use this fact to identify the next edge on which to start a cycle.
		   The use of right turning cycles means that the cycle vertices are recorded in a clckwise direction with respect to the chosen point 
		   at infinity.
		   
		   The interior edges are indicated in edge_pair, corresponding to indices [num_boundary_vertices,...,2*num_crossings), we set
		   edge_pair[i].first to be negative if we have traversed the edge in the direction first->second and set edge_pair[i].second
		   negative if we have traversed the edge in the other direction.  Since we may start a cycle anywhere , we need not worry about 
		   attempting to set vertex zero negative.
		*/
		int num_edges = 2*num_crossings;

		int start = edge_pair[0].first;		
		edge_pair[0].first *= -1;	
		int last_vertex = start;
		int next_vertex = edge_pair[0].second;
		int edge_count = 1;
		int cycle_index = 0;
		cycle_vertices[cycle_index][0] = start;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: start vertex " << start << " at edge_pair[" << num_boundary_vertices << "].second, next_vertex = " << next_vertex << endl;
		
		do
		{			
			if (next_vertex == start)
			{
				if (edge_count == 3) // cycle length 3
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph:   right turning cycle involving edge " << last_vertex << " to " << next_vertex << " has length 3" << endl;
	print (incidence,debug, 2,"valid_4_gon_graph: ");
}	
					valid = false;
					break;
				}
				else
				{
					/* here edge_count should be 4, look for a new start in edge_pair */
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:   returned to start vertex " << next_vertex << " edge_count == " << edge_count << endl;
		
					bool found = false;
					
					for (int i=num_boundary_vertices; i< num_edges; i++)
					{
						if (edge_pair[i].first == num_vertices || edge_pair[i].second == num_vertices)
						{
							/* don't want an edge involving the infinite vertex */
							continue;
						}
						else if (edge_pair[i].first > 0 )
						{
							start = edge_pair[i].first;
							edge_pair[i].first *= -1;
							last_vertex = start;
							next_vertex = abs(edge_pair[i].second);
							edge_count = 1;
							found = true;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: found next start vertex " << start << " at edge_pair[" << i << "].first" << endl;
	debug << "valid_4_gon_graph:   last_vertex = " << last_vertex << ", next_vertex = " << next_vertex << " edge_count = " << edge_count << endl;
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].first negative" << endl;
}

							break;
						}
						else if (edge_pair[i].second > 0)
						{
							start = edge_pair[i].second;
							edge_pair[i].second *=-1;
							last_vertex = start;
							next_vertex = abs(edge_pair[i].first);
							edge_count = 1;
							found = true;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: found next start vertex " << start << " at edge_pair[" << i << "].second" << endl;
	debug << "valid_4_gon_graph:   last_vertex = " << last_vertex << ", next_vertex = " << next_vertex << " edge_count = " << edge_count << endl;
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].second negative" << endl;
}

							break;
						}						
					}

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "valid_4_gon_graph: adjusted edge_pair: " << endl;
	for (int i = 0; i< 2* num_crossings; i++)
		debug << "valid_4_gon_graph:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}
					
					if (found)
					{
						cycle_index++;
						cycle_vertices[cycle_index][0] = start;
					}
					else
					{
						break;
					}
				}
			}
			else
			{
				if (edge_count == 4)  // cycle length greater than 4
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:   right turning cycle involving edge " << last_vertex << " to " << next_vertex << " has length greater than 4" << endl;
					valid = false;
					break;
				}
				else
				{
					cycle_vertices[cycle_index][edge_count] = next_vertex;
					
					edge_count++;
					
					/* turn right at next_vertex and set next_vertex in edge_pair negative */
					list<int>::iterator pptr = find_if(petal_list[next_vertex].begin(),petal_list[next_vertex].end(),vertex_equals(last_vertex));
					if (pptr == petal_list[next_vertex].end())
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: Error! could not find " << last_vertex << " in the petal list of vertex " << next_vertex << endl;
						cerr << "Error! valid_4_gon_graph could not find " << last_vertex << " in the petal list of vertex " << next_vertex << endl;
						exit(0);
					}
					else
					{
						pptr++;
						if (pptr == petal_list[next_vertex].end())
							pptr = petal_list[next_vertex].begin();
														
						last_vertex = next_vertex;
						next_vertex = *pptr;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:   last_vertex = " << last_vertex << ", next_vertex " << next_vertex << " edge_count = " << edge_count << endl;
						
						/* find the edge last_vertex -- next_vertex in edge pair and set last_vertex negative */
						for (int i=num_boundary_vertices; i< num_edges; i++)
						{
							if (edge_pair[i].first == last_vertex && abs(edge_pair[i].second) == next_vertex)
							{
								edge_pair[i].first *= -1;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].first negative" << endl;
								break;
							}
							else if (edge_pair[i].second == last_vertex && abs(edge_pair[i].first) == next_vertex)
							{
								edge_pair[i].second *= -1;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].second negative" << endl;
								break;
							}
						}
					}
				}
			}			
		} while (valid);		

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: cycle:" << endl;
	for (int i=0; i<num_4_gons_in_disc; i++)
	{
		debug << "valid_4_gon_graph:   " << i+1 << ": ";
		for (int j=0; j< 4; j++)
			debug << setw(3) << cycle_vertices[i][j];
		debug << endl;
	}
	
}					

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: returning valid = " << valid << endl;
		
	if (valid)
	{		
		gauss_code = doodle_gauss_code(edge_pair,cycle_vertices,2*num_crossings,num_boundary_vertices);
	}
	
	return valid;
}

/* doodle_gauss_code produces a Gauss code for the doodle determined by the planar dual graph described by edge_pair and cycle_vertices.
   Each of the 4-gon regions of the dual graph (including the infinite 4-gon regions outside the disc) correspond to a crossing, so we
   first identify the edges corresponding to each region, recording them in a clockwise manner with respect to the chosen point at infinity
   and then we can trace the components of the doodle, as the strands across each 4-gon join opposite edges, forming a cross.
*/
string doodle_gauss_code(vector<pair<int,int> >& edge_pair,matrix<int>& cycle_vertices, int num_edges, int num_boundary_edges)
{
	string left_preferred = "";
	
	int num_edges_in_infinite_cycle = num_boundary_edges/2;
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: presented with edge_pair: " << endl;
	for (int i = 0; i< num_edges; i++)
		debug << "doodle_gauss_code:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}

	/* identify the region edges, where the edges are numbered according to the order in edge_pair.  The boundary edges are listed
	   first, then the interior edges, then the edges to the infinite vertex.  We reset the sign of the interior edge pairs since the 
	   entries will be negative after valid_4_gon_graph.
	*/
	for (int i= num_boundary_edges; i< num_edges-num_edges_in_infinite_cycle; i++)
	{
		edge_pair[i].first *= -1;
		edge_pair[i].second *= -1;
	}
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: presented with edge_pair: " << endl;
	for (int i = 0; i< num_edges; i++)
		debug << "doodle_gauss_code:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}

	/* calculate the region edges, where the edges are numbered by the corresponding index in edge_pair. The region edges will be 
	   recorded clockwise, since that is the orientation in which the cycle_vertices are recorded. 
	*/
	int num_4_gons_in_disc = (num_edges - num_boundary_edges)/2;
	matrix<int> region_edges(num_edges/2,4);
	
	for (int i=0; i< num_4_gons_in_disc; i++)
	{
		for (int j=0; j< 4; j++)
		{
			int v1 = cycle_vertices[i][j];
			int v2 = cycle_vertices[i][(j+1)%4];
			
			for (int k=0; k< num_edges; k++)
			{
				if ( (edge_pair[k].first == v1 && edge_pair[k].second == v2) || (edge_pair[k].second == v1 && edge_pair[k].first == v2) )
				{
					region_edges[i][j] = k;
					break;
				}
			}
		}
	}
	
	/* Add the infinite regions to region_edges in the same clockwise manner used for the interior regions */
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
	{
		region_edges[num_4_gons_in_disc+i][0] = 2*i;
		region_edges[num_4_gons_in_disc+i][1] = (2*i-1+num_boundary_edges)%num_boundary_edges;
		region_edges[num_4_gons_in_disc+i][2] = num_edges - num_edges_in_infinite_cycle +(i-1+num_edges_in_infinite_cycle )%num_edges_in_infinite_cycle;
		region_edges[num_4_gons_in_disc+i][3] = num_edges - num_edges_in_infinite_cycle + i;
	}
	
	
	
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: region_edges: " << endl;
	for (int i = 0; i< num_edges/2; i++)
	{
		debug << "doodle_gauss_code:   " << i << ": ";	
		for (int j=0; j< 4; j++)
			debug << setw(3) << region_edges[i][j];
		debug << endl;
	}
}

	/* trace the region edges to build up a list of the consecutive edges encountered as we traverse the doodle.  Each interior edge appears twice in 
	   region_edges and each boundary edge apears once.  We always traverse regions by leaving on the opposite edge (i.e i <-->3 and 2 <--> 4).  If we 
	   hit the disc boundary we have to traverse an "infinite 4-gon" and cross an infinite edge.
	   
	   If we are leaving the disc on an odd numbered edge, we cross the suceeding infinite edge and re-enter the disc three boundary edges along.
	   If we are leaving the disc on an even numbered edge, we cross the preceding infinite edge and re-enter the disc three boundary edges back.
	   
	   There are num_boundary_vertices/2 infinite edges and they appear at the end of edge_pair, attached to vertices 1,3,...,num_boundary_vertices-1.
	   Consider the infinite edges numbered from zero. Then, the succeeding infinite edge to edge 2i+1 is the (i+1)-st infinite edge and the preceding 
	   infinite edge to edge 2i is the (i-1)-st infinite edge, for i=0,...
	*/
	list<list<int> > component_edges;
	list<list<int> > component_regions;
	
	vector<bool> edge_flag(num_edges); // recoords whether we have visited the edge, initializes to false	
	bool finished = false;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: search for components:" << endl;
	
	do
	{
		/* look for the start of another component */
		finished = true;
		int start = -1;
		for (int i=0; i< num_edges; i++)
		{
			if (edge_flag[i] == false)
			{
				finished = false;
				start = i;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: found start of component at edge " << start;
				break;
			}
		}
		
		if (!finished)
		{
			list<int> next_component_edge;
			list<int> next_component_region;
			edge_flag[start] = true;
			next_component_edge.push_back(start);
			int region = -1;  // records the region we crossed last
			int next_edge = succeeding_edge(region_edges,num_4_gons_in_disc, start,region);
			next_component_region.push_back(region);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << ", next_edge = " << next_edge << " in region " << region << endl;
					
			/* trace the component */
			bool complete = false;
			do
			{				
				if (next_edge == start)
				{				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "\ndoodle_gauss_code:   next_edge = start edge " << start << " component complete" << endl;
		
					complete = true;
				}
				else 
				{
					edge_flag[next_edge] = true;
					next_component_edge.push_back(next_edge);  
					
					if (next_edge < num_boundary_edges)
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   edge " << next_edge << " is on the disc boundary" << endl;
			
						if (next_edge%2 == 1)
						{
							int infinite_edge = num_edges - num_edges_in_infinite_cycle + (next_edge/2 +1)%num_edges_in_infinite_cycle;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     infinite_edge = " << infinite_edge << endl;
			
							edge_flag[infinite_edge] = true;
							next_component_edge.push_back(infinite_edge);
							next_component_region.push_back(((next_edge+1)%num_boundary_edges)/2+num_4_gons_in_disc);
							
							next_edge = (next_edge+3)%num_boundary_edges;
							next_component_region.push_back(next_edge/2+num_4_gons_in_disc);
							
							if (next_edge == start)
							{				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   next_edge = start edge " << start << " component complete" << endl;
			
								complete = true;
							}
							else
							{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next boundary edge = " << next_edge << endl;
								edge_flag[next_edge] = true;
								next_component_edge.push_back(next_edge);
								region = -1; // re-entering the disc, want to check all regions
								next_edge = succeeding_edge(region_edges,num_4_gons_in_disc,next_edge,region);
								next_component_region.push_back(region);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next interior edge = " << next_edge << " in region " << region << endl;
							}
						}
						else
						{
							int infinite_edge = num_edges - num_edges_in_infinite_cycle +(next_edge/2 - 1 + num_edges_in_infinite_cycle )%num_edges_in_infinite_cycle;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     infinite_edge = " << infinite_edge << endl;
	
							edge_flag[infinite_edge] = true;
							next_component_edge.push_back(infinite_edge);
							next_component_region.push_back(next_edge/2+num_4_gons_in_disc);
							next_edge = (next_edge-3+num_boundary_edges)%num_boundary_edges;
							next_component_region.push_back(((next_edge+1)%num_boundary_edges)/2+num_4_gons_in_disc);
		
							if (next_edge == start)
							{				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   next_edge = start edge " << start << " component complete" << endl;
			
								complete = true;
							}
							else
							{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next boundary edge = " << next_edge << endl;
								edge_flag[next_edge] = true;
								next_component_edge.push_back(next_edge);
								region = -1; // re-entering the disc, want to check all regions
								next_edge = succeeding_edge(region_edges,num_4_gons_in_disc,next_edge,region);
								next_component_region.push_back(region);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next interior edge = " << next_edge << " in region " << region << endl;
							}
						}
					}
					else
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   edge " << next_edge << " is in the disc interior in region " << region;
						/* move across the region to the next edge */
						next_edge = succeeding_edge(region_edges,num_4_gons_in_disc,next_edge,region);
						next_component_region.push_back(region);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << ", next_edge  = " << next_edge << " in region " << region << endl;
					}
				}
			} while(!complete);
	
			component_edges.push_back(next_component_edge);
			component_regions.push_back(next_component_region);
		}
	} while (!finished);
	
	int num_components = component_edges.size();
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: num_components = " << num_components << endl;
	debug << "doodle_gauss_code: component edges: " << endl;
	
	list<list<int> >::iterator llptr = component_edges.begin();
	int index=0;
	
	while (llptr != component_edges.end())
	{
		debug << "doodle_gauss_code:   " << index++ << ": ";
		
		list<int>::iterator lptr = (*llptr).begin();
		while (lptr != (*llptr).end())
		{
			debug << setw(3) << *lptr;
			lptr++;
		}
		
		debug << endl;
		llptr++;
	}

	debug << "doodle_gauss_code: component regions: " << endl;
	
	llptr = component_regions.begin();
	index=0;

	while (llptr != component_regions.end())
	{
		debug << "doodle_gauss_code:   " << index++ << ": ";
		
		list<int>::iterator lptr = (*llptr).begin();
		while (lptr != (*llptr).end())
		{
			debug << setw(3) << *lptr;
			lptr++;
		}
		
		debug << endl;
		llptr++;
	}
	
}

	/* Write the doodle Gauss code to a string. */
	ostringstream oss;
	list<list<int> >::iterator elptr = component_edges.begin();
	list<list<int> >::iterator rlptr = component_regions.begin();
	unsigned int component = 1;
	
	while (elptr != component_edges.end())
	{	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: write component " << distance(component_edges.begin(),elptr) << endl;

		list<int>::iterator eptr = (*elptr).begin();
		list<int>::iterator rptr = (*rlptr).begin();
		
		unsigned int term = 1;
		
		while (eptr != (*elptr).end())
		{			
			/* determine whether the strand transverse to e1 = *eptr -- *(eptr+1) = e2 is a left or right strand.  We use the fact that the 
			   region_edges are recorded clockwise, so each region has the form shown below, where n is the region(crossing number):
			   
			                               b
			                            +------+
			                            |      |
			                         a  |  n   | c
			                            +------+
			                               d
			                               
			   the strand we are following is oriented from e1 to e2, so given the transvere strand, oriented from t1 to t2 we have:

			   reverse_e_edges = false			   
			   e1--e2 = a--c    t1--t2 = b--d  => Ln      t1--t2 = d--b  => Rn  
			   e1--e2 = b--d    t1--t2 = a--c  => Rn      t1--t2 = c--a  => Ln  

			   reverse_e_edges = true
			   e1--e2 = c--a    t1--t2 = b--d  => Rn      t1--t2 = d--b  => Ln  
			   e1--e2 = d--b    t1--t2 = a--c  => Ln      t1--t2 = c--a  => Rn  
			   
			   These eight cases account for all possible cyclic permutations of a b c and d in region_edges, since they correspond to a 
			   rotation of the above diagram.
			*/			
			int e1 = *eptr;
			int e2;
			list<int>::iterator next_eptr = eptr;
			next_eptr++;
			if (next_eptr == (*elptr).end())
				e2 = *(*elptr).begin();
			else
				e2 = *next_eptr;
				
			int region = *rptr;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: edges " << e1 << " and " << e2 << ", region " << region << endl;
	
			bool reverse_e_edges = false;
			bool reverse_t_edges = false;
			bool e_edges_position_a_c = false;
			bool left_transverse = false;
		
			int t1=-1;
			int t2=-1;
			
			for (int i=0; i< 4; i++)
			{
				if (region_edges[region][i] == e1)
				{
					if (i/2 == 1)
						reverse_e_edges = true;
					else
						reverse_e_edges = false;
						
					if (i%2 == 0)
						e_edges_position_a_c = true;						
					else
						e_edges_position_a_c = false;
				}
				else if (region_edges[region][i] != e2)
				{
					if (t1 == -1)
					{
						t1 = region_edges[region][i];

						if (i/2 == 1)
							reverse_t_edges = true;
						else
							reverse_t_edges = false;
					}
					else
					{
						t2 = region_edges[region][i];
					}
				}
			}
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code:   transverse edges " << t1 << " and " << t2 << endl;
	debug << "doodle_gauss_code:   e_edges_position_a_c = " << e_edges_position_a_c << " reverse_e_edges = " << reverse_e_edges << " reverse_t_edges = " << reverse_t_edges << endl;
}				

			/* order_transverse_edges ensures that t1 preceeds t2 in the orientation implicit in component_edges,
			   re-ordering their values if necessary.  Look for t1 in component_edges, each edge appears exactly once, 
			   then look to see whether t1 appears in front or after t1
			*/
			
			list<list<int> >::iterator llptr = component_edges.begin();
			bool found_t1 = false;
			int index = 0;
			while (llptr != component_edges.end())
			{			
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   looking for transverse edges " << t1 << " and " << t2 << " in component  " << index++ << endl;
					
				list<int>::iterator lptr = (*llptr).begin();
				while (lptr != (*llptr).end())
				{
					if (*lptr == t1)
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   found at offset " << distance((*llptr).begin(),lptr) << endl;
	
						found_t1 = true;
						list<int>::iterator next_ptr = lptr;
						list<int>::iterator prev_ptr = lptr;
						
						if (prev_ptr == (*llptr).begin())
							prev_ptr = (*llptr).end();

						prev_ptr--;
							
						next_ptr++;
						if (next_ptr == (*llptr).end())
							next_ptr = (*llptr).begin();

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   prev_ptr offset " << distance((*llptr).begin(),prev_ptr) << " value " << *prev_ptr << " next_ptr offset " << distance((*llptr).begin(),next_ptr) << " value " << *next_ptr << endl;
						
						if (*prev_ptr == t2)
						{
							swap(t1, t2);
							reverse_t_edges = !reverse_t_edges;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   transverse edges found in reverse order in component_edges, new t1 = " << t1 << ", new t2 = " << t2 << " new reverse_t_edges = " << reverse_t_edges << endl;

						}
						else if (*next_ptr != t2)
						{
							cerr << "Error! Transverse edges " << t1 << " and " << t2 << " are not consecutive in component_edges, aborting." << endl;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: Error! transverse edges " << t1 << " and " << t2 << " are not consecutive in component_edges" << endl;
							exit(0);
						}
						
						break;
					}
					
					lptr++;
				}
								
				if (found_t1)
					break;
				else
					llptr++;
			}

			if ((reverse_e_edges && e_edges_position_a_c & reverse_t_edges) || 
			    (reverse_e_edges && !e_edges_position_a_c & !reverse_t_edges) || 
			    (!reverse_e_edges && e_edges_position_a_c && !reverse_t_edges) ||
			    (!reverse_e_edges && !e_edges_position_a_c && reverse_t_edges)
			   )
				left_transverse = true;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code:   transverse edges " << t1 << " and " << t2 << endl;
	debug << "doodle_gauss_code:   e_edges_position_a_c = " << e_edges_position_a_c << " reverse_e_edges = " << reverse_e_edges << " reverse_t_edges = " << reverse_t_edges << endl;
	debug << "doodle_gauss_code:   left_transverse = " << left_transverse << endl;
}				
			
			if(left_transverse)
				oss << 'L';
			else
				oss << 'R';
				
			oss << region+1;  // Gauss code crossing snumbered from 1.
			
			if (term < (*elptr).size())
				oss << ' ';
				
			term++;
			
			eptr++;
			rptr++;
		}
		
		if (component < component_edges.size())
			oss << ',';

		component++;
		
		elptr++;
		rlptr++;
	}

	ostringstream code_trailer;
	
	code_trailer << '/';
	
	for (int i=0; i< num_edges/2; i++)
	{
		code_trailer << '#';
		if (i < num_edges/2 - 1)
			code_trailer << ' ';
	}
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   gauss code " << oss.str() << code_trailer.str() << endl;

	ofstream tmp("tmpfile");
	tmp << "[raw-output,silent,gauss,ulpgd]" << endl;
	tmp << oss.str() << code_trailer.str() << endl;
	
	system("braid tmpfile");
	ifstream braid("braid.out");
	
	if (!getline(braid, left_preferred))
	{
		cerr << "\nError! Braid programe failed to calculate left_preferred Gauss data." << endl;
		exit(0);
	}
	braid.close();

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   left preferred gauss code " << left_preferred << code_trailer.str() << endl;

	return left_preferred + code_trailer.str();
}

int succeeding_edge(matrix<int>& region_edges, int num_4_gons_in_disc, int edge, int& region)
{
	int next_edge = -1;
	
	for (int i=0; i< num_4_gons_in_disc; i++)
	{
		if (i == region)
			continue;
		
		bool found = false;
			
		for (int j=0; j<4; j++)
		{
			if (region_edges[i][j] == edge)
			{
				next_edge = region_edges[i][(j+2)%4]; // 0 <-> 2, 1 <-> 3
				region = i;
				found = true;
				break;
			}
		}
		
		if (found)
			break;
	}
	
	return next_edge;
}

/* four_connected returns true if the planar doodle described by gauss_code is 4-connected.  For each turning cycle, we enumerate the boundary regions
   and then look for two boundary regions that share an edge or vertex indicating that the doodle has three crossings that separate.
*/
bool four_connected (string gauss_code)
{
	generic_code_data gauss_code_data;
	read_gauss_code(gauss_code_data,gauss_code);
	generic_code_data peer_code_data;
	gauss_to_peer_code(gauss_code_data, peer_code_data, true, 0, false); // optimal=true, evaluate_gauss_crossing_perm = false, address of gauss_crossing_map = 0
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "four_connected: peer_code_data ";
	write_peer_code(debug,peer_code_data);	
	debug << endl;
	print_code_data (debug, peer_code_data, "four_connected:   ");
}
	
	int num_left_cycles;
	int num_cycles;	
	int num_crossings = peer_code_data.num_crossings;
	matrix<int> cycle(num_crossings+2, 2*num_crossings+1);

	calculate_turning_cycles(peer_code_data, cycle, num_left_cycles, num_cycles);

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "four_connected: left turning cycles:" << endl;   
	for (int i=0; i<num_left_cycles; i++)
	{
		debug << "four_connected:   cycle " << i << ": ";
		for (int j=1; j<=cycle[i][0]; j++)
			debug << cycle[i][j] << " ";
		debug << endl;
	}

	debug << "four_connected: right turning cycles:" << endl;   
	for (int i=num_left_cycles; i<num_cycles; i++)
	{
		debug << "four_connected:   cycle " << i << ": ";
		for (int j=1; j<=cycle[i][0]; j++)
			debug << cycle[i][j] << " "	;
		debug << endl;
	}
}
	
	bool four_connected_doodle = true;
	
	for (int i=0; i< num_cycles && four_connected_doodle; i++)
	{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected: cycle " << i << endl;
	
		/* identify the boundary regions when cycle i is chosen as the infinte region */
		vector<int> boundary_region(2*cycle[i][0]);
		
		for (int j=1; j<= cycle[i][0]; j++)
		{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected:   edge " << cycle[i][j];
	
			/* find the other cycle containing the j-th edge of the i-th turning cycle */
			int edge_adjacent_cycle;
			for (int k=0; k< num_cycles; k++)
			{
				if (k == i)
					continue;
					
				for (int l=1; l<= cycle[k][0]; l++)
				{
					if (cycle[k][l] == cycle[i][j]) // both will be negative if this is an odd edge
					{
						edge_adjacent_cycle = k;
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << " appears in cycle " << k << endl;
						boundary_region[2*(j-1)] = k;
					}
				}
			}
			
			/* find the v-boundary cycle containing the succeeding crossing of the j-th edge, first identify the e-boundary region incident with the next
			   edge in the infinite turning cycle, since we need to avoid this region when identifying the next v-boundary region
			*/
			int succeeding_crossing = (cycle[i][j] <0 ? peer_code_data.orig_crossing[abs(cycle[i][j])] : peer_code_data.term_crossing[cycle[i][j]]);
			int succeeding_edge = (j == cycle[i][0]? cycle[i][1] : cycle[i][j+1]);

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected:     succeeding_crossing = " << succeeding_crossing << " succeeding_edge = " << succeeding_edge;
			
			int succeeding_adjacent_cycle;
			
			
			for (int k=0; k< num_cycles; k++)
			{
				if (k == i)
					continue;
					
				for (int l=1; l<= cycle[k][0]; l++)
				{
					if (cycle[k][l] == succeeding_edge) // both will be negative if this is an odd edge
					{
						succeeding_adjacent_cycle = k;
						
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << " succeeding_adjacent_cycle = " << succeeding_adjacent_cycle << endl;
					}
				}
			}

			
			for (int k=0; k< num_cycles; k++)
			{
				if (k == i || k == edge_adjacent_cycle || k == succeeding_adjacent_cycle)
					continue;
					
				
				for (int l=1; l<= cycle[k][0]; l++)
				{
					if ( (cycle[k][l] < 0 &&  peer_code_data.orig_crossing[abs(cycle[k][l])] == succeeding_crossing) 
					    || (cycle[k][l] >= 0 && peer_code_data.term_crossing[cycle[k][l]] == succeeding_crossing)
					   )
					 // both will be negative if this is an odd edge
					{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected:     succeeding crossing " << succeeding_crossing;
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << " appears in cycle " << k << endl;
	
						boundary_region[2*j-1] = k;
	
					}
				}
			}
		}
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "four_connected:   boundary_region: ";
	for (int j=0; j < 2*cycle[i][0]; j++)
		debug << boundary_region[j] << ' ';
	debug << endl;
}

		/* if the doodle is not 4-connected:
		   an odd v-boundary region j could share a vertex with odd v-boundary region, other than the two adjacent v-boundary regions (j-2 and j+2)
		   an odd v-boundary region j could share an edge with an edge-boundary region, other than the two adjacent e-boundary regions (j-1 and j+1)
		   an even e-boundary region j could share a vertex with an e-boundary region, other than the two adjacent e-boundary regions (j-2 and j+2)
		*/
		for (int j=0; j < 2*cycle[i][0] && four_connected_doodle; j++)
		{
			int j_region = boundary_region[j];
			vector<int> j_region_vertex(cycle[j_region][0]);
			for (int k = 1; k<= cycle[j_region][0]; k++)
			{
				j_region_vertex[k-1] = (cycle[j_region][k] < 0 ? peer_code_data.orig_crossing[abs(cycle[j_region][k])] : peer_code_data.term_crossing[cycle[j_region][k]]);

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "four_connected:   region " << j_region << " index k = " << k << ", edge " << cycle[j_region][k] << endl;

			}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "four_connected:   region " << j_region << " vertices : ";
	for (int k=0; k < cycle[j_region][0]; k++)
		debug << j_region_vertex[k] << ' ';
	debug << endl;
}			

/*
			int preceding_region = (j > 0 ? boundary_region[j-1] : boundary_region[2*cycle[i][0]-1]);
			int succeeding_region = (j < 2*cycle[i][0]-1? boundary_region[j+1] : boundary_region[0]);

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected: preceding region " << preceding_region << " succeeding_region " << succeeding_region << endl;
*/

			int start = (j%2 == 0 ? 4: 3);
			int end = (j%2 == 0 ? 2*cycle[i][0]-2: 2*cycle[i][0]-2);
			int step = (j%2 == 0 ? 2: 1);

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected:     j = " << j << ", start region " << boundary_region[(j+start)%(2*cycle[i][0])] << ", end region " << boundary_region[(j+end)%(2*cycle[i][0])] << ", step " << step << endl;
			
			for (int k=start; k< end && four_connected_doodle; k+=step)
			{					
				int k_region = boundary_region[(j+k)%(2*cycle[i][0])];				

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected:     checking region " << k_region << endl;
				
				for (int l=1; l <= cycle[k_region][0]; l++)
				{
					int k_region_vertex = (cycle[k_region][l] < 0 ? peer_code_data.orig_crossing[abs(cycle[k_region][l])] : peer_code_data.term_crossing[cycle[k_region][l]]);
					if (find(j_region_vertex.begin(), j_region_vertex.end(), k_region_vertex) != j_region_vertex.end())
					{
						four_connected_doodle= false;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "four_connected:     found vertex " << k_region_vertex << " in region " << k_region << endl;
						break;
					}
				}
			}
		}
	}
	
	
	return four_connected_doodle;
}
