/**************************************************************************
doodle_code evaluates all possible solutions to

n_3+n_4+...+n_k = n+2
\Sum i*n_i = 2*num_edges = 4*n

for a gven n, thereby determining all possible doodle codes n n_3 n_4,...,n_k

Based on the value of the boolean SHOW_CODES_ONLY it then calls 4-gon with each
doodle dode, and the appropriate parameters, as determiend from the command line

Command line syntax described in helpinfo function in this module.  Running doodle with 
no parameters displays usage information.

January 2023: initial version
August 2023: first released version

**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>

/******************** Global Variables **********************/
bool	DEVELOPMENT_MODE = false;
bool	SHOW_CODES_ONLY = false;
bool	OMIT_BOUNDARY_CHORDS = false;
bool	BOUNDARY_CHORDS_ONLY = false;
bool	SHOW_INTERIOR_CHOICES = false;

bool	termination_request = false;
bool 	sigusr1_request = false;

ofstream 	output;
extern ofstream    debug;

#include <4-gon.h>


/********************* Function prototypes ***********************/
void helpinfo();
void debug_help();
void set_programme_long_option(char* cptr, string source);
void set_programme_short_option(char* cptr);
bool debug_setup(char* argv_ptr);
void check_debug_option_parameters(char* pptr, string option);
void print_prog_params (ostream& os, string level, string prefix);

namespace util
{
	string itos(long n);
}

int diagramnum = 1;

void sigterm_handler (int sig) // kill -15, or just kill
{
	cout << "\nvlist programme received SIGTERM, setting termination_request=true" << endl;
	termination_request = true;
}

/* sigusr1_handler causes vlist to report progress on the current task and continue. */
void sigusr1_handler (int sig) // kill -10
{
	cout << "\nvlist programme received SIGUSR1, setting sigusr1_request=true" << endl;
	sigusr1_request = true;
}


/******************* Main Function ************************/

int main (int argc, char* argv[])
{

	if (argc == 1)
	{
		helpinfo();
		exit(0);
	}
	
	signal(SIGTERM, sigterm_handler); // register the SIGTERM handler
	signal(SIGUSR1, sigusr1_handler); // register the SIGUSR1 handler
	
	/* take a copy of the original command line in case we're debugging and remove
	   the debug details.  We will want to record the original command line for the debug file.
	*/
	
	vector<string> original_programme_arguments(argc); 
	for (int i = 0; i< argc; i++)
		original_programme_arguments[i] = string(argv[i]);
	
	int num_crossings = 0;
	int max_code_index = 0;
	
	
    /* Read the programme parameters and input files */
	for (int i=1; i < argc; i++)
	{
		if (*(argv[i]) == '-')
		{
			if (*(argv[i]+1) == '-')
				set_programme_long_option(argv[i], "command line");
			else 
				set_programme_short_option(argv[i]);
		}
		else if (num_crossings == 0) 
		{
			get_number(num_crossings, argv[i]);
			max_code_index = num_crossings/2;
		}
	}	

/*DEVELOPMENT_MODE is used for testing individual functions during development */
if (DEVELOPMENT_MODE)
{
	exit(0);
}
/* End of DEVELOPMENT_MODE code */
	
	output.open("doodle_code.out");

    if (!output)
    {
		cerr << "\nError opening output file" << endl;
		exit(0);
    }
	
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	print_prog_params (debug, "basic", "");
}

	output << "\nnum_crossings = " << num_crossings << " max_code_index = " << max_code_index << endl;


if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << original_programme_arguments[i] << ' ';
	debug << endl;
}
		
int doodle_code_count = 0;

for (int passes=0; passes<2; passes++)
{	

	/* the doodle code has to satisfy n3+n4+...+nk = num_crossings+2 and 3n3+4n4+...+knk = 4*num_crossings */
	vector<int> doodle_code(max_code_index-2);
	doodle_code[max_code_index-3] = num_crossings+2;
	vector<int> max(max_code_index-2,num_crossings+2);
	bool complete = false;	
	
	do
	{			
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "doodle_code: ";
		
	for (int i=0; i< max_code_index-2; i++)
		debug << doodle_code[i] << ' ';
		
	debug << "  max: ";
	
	for (int i=0; i< max_code_index-2; i++)
		debug << max[i] << ' ';
}
			
		int weighted_sum = 0;
		for (int i=0; i< max_code_index-2; i++)
			weighted_sum += (3+i)*doodle_code[i];

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  weighted_sum = " << weighted_sum << endl;


		if (weighted_sum == 4*num_crossings)
		{
			int max_non_zero_index = -1;
			for (int i=max_code_index-3; i>=0; i--)
			{
				if (doodle_code[i] >0)
				{
					max_non_zero_index = i;
					break;
				}				
			}
			
			if (passes == 0)
			{
				if (SHOW_CODES_ONLY)
				{
					cout << "doodle code ";
					for (int i=0; i<= max_non_zero_index; i++)
						cout  << doodle_code[i] << ' ';				
					cout << endl;
				}
				
				doodle_code_count++;
				
				output << "doodle code ";
				for (int i=0; i<= max_non_zero_index; i++)
					output  << doodle_code[i] << ' ';				
				output << endl;
			}
			else
			{
				ostringstream command;
				if (SHOW_INTERIOR_CHOICES)
					command << "4-gon -i " << num_crossings << ' ';
				else if (BOUNDARY_CHORDS_ONLY)
					command << "4-gon -b " << num_crossings << ' ';
				else if (OMIT_BOUNDARY_CHORDS)
					command << "4-gon -B " << num_crossings << ' ';
				else
					command << "4-gon -Q " << num_crossings << ' ';
					
				for (int i=0; i<= max_non_zero_index; i++)
					command  << doodle_code[i] << ' ';
				
				cout << command.str() << endl;
				
				system(command.str().c_str());
			}
		}

		/* increment doodle_code */
		complete = true;
		for (int i=max_code_index-3; i >=0; i--)
		{
			if (doodle_code[i] < max[i])
			{
				complete = false;
				doodle_code[i]++;
				
				int cumulative_count=0;
				for (int j=0; j<=i; j++)
					cumulative_count += doodle_code[j];

				for (int j=i+1; j< max_code_index-2; j++)
				{
					max[j] = num_crossings+2-cumulative_count;					
					if (j<max_code_index-3)
						doodle_code[j] = 0;
					else
						doodle_code[j] = num_crossings+2-cumulative_count;
				}
				break;
			}
		}
	} while (!complete);	
	
	if (SHOW_CODES_ONLY) // for testing only
	{
		output << "number of doodle codes = " << doodle_code_count << endl;
		break;
	}
}
	
}

void set_main_debug_option_parameter(char* pptr, string option);
void set_main_debug_option(char* start, char* end)
{
	char  loc_buf[end-start+2];
	char* c1 = start;
	char* c2 = loc_buf;

	/* if both start and end are zero, display debug help information.
	   this has been included here in this manner so that each time a debug option
	   is added, the help will be updated (hopefully!)
	*/
	if (start == 0 && end == 0)
	{
		set_main_debug_option_parameter(0,"4-gon");
		return;
	}

	do
	{
		*c2++ = *c1++;
	} while (c1 <= end);
	
	*c2 = '\0';

	char* pptr = strchr(loc_buf,'{');
	if (pptr)
	{
		*pptr++ = '\0';
	}

	/* now, even if there are parameters, the first part of loc_buf 
	   is a C-string that identifies the option */
	
	if (!strcmp(loc_buf,"4-gon"))
	{
		debug_control::DEBUG = debug_control::SUMMARY;
		debug << "main::set_main_debug_option: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
		
		if (pptr)
		{
			/* check for any parameters */
			check_debug_option_parameters(pptr, "4-gon");
		}
	}

	debug.flush();

}

void set_main_debug_option_parameter(char* pptr, string option)
{
	if (option == "4-gon")
	{
		if (!pptr)
		{
			cout << "\t\t4-gon{summary|1:basic|2:intermediate|3:detail|4}, integer: default 0=off, no parameters sets summary" << endl;
		}
		else
		{
			if (!strcmp(pptr,"summary") || !strcmp(pptr,"1") )
			{
				debug_control::DEBUG = debug_control::SUMMARY;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::SUMMARY" << endl;		
			}
			if (!strcmp(pptr,"basic") || !strcmp(pptr,"2") )
			{
				debug_control::DEBUG = debug_control::BASIC;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::BASIC" << endl;		
			}
			else if (!strcmp(pptr,"intermediate") || !strcmp(pptr,"3"))
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			else if (!strcmp(pptr,"detail") || !strcmp(pptr,"4"))
			{
				debug_control::DEBUG = debug_control::DETAIL;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::DETAIL" << endl;		
			}
			else if (!strcmp(pptr,"exhaustive") || !strcmp(pptr,"5"))
			{
				debug_control::DEBUG = debug_control::EXHAUSTIVE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::EXHAUSTIVE" << endl;		
			}
		}
	}
}

void set_programme_short_option(char* cptr)
{
	char* dptr = strchr(cptr, '#');
	if (dptr)
	{
		debug.open ("doodle-code.dbg"); 

		if (!debug)
		{
			cerr << "\nError opening debug file\n";
			exit(0);
		}
		else
		{
			debug << "Debug information from vlist\n\n";
			debug << boolalpha;
		}

		if (!debug_setup(cptr))
		{
			debug_control::DEBUG = debug_control::SUMMARY;
			debug << "main: default debug options set" << endl;
		}
		
	}

	if (strchr(cptr, 'b'))
		BOUNDARY_CHORDS_ONLY = true;
	
	if (strchr(cptr, 'B'))
		OMIT_BOUNDARY_CHORDS = true;

	if (strchr(cptr, 'c'))
		SHOW_CODES_ONLY = true;

	if (strchr(cptr, 'i'))
		SHOW_INTERIOR_CHOICES = true;
}

void set_programme_long_option(char* cptr, string source)
{
}

void print_prog_params (ostream& os, string level, string prefix)
{
    os << prefix << "DEVELOPMENT_MODE = " << DEVELOPMENT_MODE << endl;
    os << prefix << "SHOW_CODES_ONLY = " << SHOW_CODES_ONLY << endl;
    os << prefix << "SHOW_INTERIOR_CHOICES = " << SHOW_INTERIOR_CHOICES << endl;
}

void main_display_default_options()
{
	cout << "\t\t4-gon{summary}" << endl;
}

void helpinfo()
{
	cout << "\nUsage: doodle-code [-<short-option>] [--<long-option>] <num-crossings>";
	cout << "\n<short-option>:";
	cout << "\n       #: debug, see #H! for details";
	cout << "\n       b: use boundary chords only option for 4-gon";
	cout << "\n       B: use omit boundary chords option for 4-gon";
	cout << "\n       c: show codes only, don't call 4-gon";
	cout << "\n       i: show 4-gon interior choices";
	cout << "\n<long-option>:";
	cout << endl;
	exit(0);
}
