/**********************************************************************************************************************
                               4-gon

4-gon takes a doodle code of the form n n_3 n_4,...,n_k,  where n is the number of crossings and n_i is the number of i-gon regions
and enumerates all prime (3-connected) minimal planar doodles, with that code.  For an overview of the algorithm, see the paper
"Planar Doodles: Their Properties, Codes and Classification"

By Euler's formula, n_3+n_4+...+n_k = n+2 and counting edges we have \Sum i*n_i = 2*num_edges = 4*n                   (*)
but not every such code corresponds to a realizable minimal planar doodle, see e.g. min_num_interior_vertices below.

A 3-connected doodle diagram has a dual graph, \Gamma that divides the 2-sphere into 4-gons.  Moreover, the dual graph corresponding 
to the compact regions of the doodle complement, \Gamma', is a the 1-skeleton of a "4-gon disc" disc, \Delta, with the property that every 
alternate vertex of \Gamma' that lies in the boundary of \Delta is joined by an edge to the "infinite" vertex of \Gamma.

The strategy for the 4-gon search is to choose a region with the maximal number of edges as the infinite region and attempt 
to construct a 4-gon disc.  The disc has:

    n-p      4-gons (since there are p 4-gons outside the disc, incident with the infinite vertex)
    n+1-2p   interior vertices
    2n-3p    interior edges (2n edges in total, 2p in the boundary of the disc and p incident with the infinite vertex)

The enumeration proceeds by considering all possible patterns of boundary chords and, for each pattern, each assignment of interior vertices
to chord regions and all possible interior choices for the inside edges.  Finally, the remaining interior edges are added by considering all
combinations of assignment to the available interior vertices (including boundary to interior vertex assignments).

Doodles are identified up to reflection in a line in the plane so the dual graph need only be constructed up to symmetry described by the 
dihedral group on $2p$ elements.

The vertices in boundary of the disc are numbered 0,...,2p-1 clockwise and we assume, without loss in generality, that vertex 0 is a v-boundary 
vertex.  Interior vertices are numbered as described below.  

An incidence matrix for the dual graph is initialised with the edges in the boundary of the disc and then there are three stages to complete a
candidate matrix.

Stage 1, within the main function.

The first stage of the enumeration is to identify all possible configurations of boundary chords, including the case where there are none, 
and to divide \Delta into chord regions that have the property that a vertex in the chord region's boundary only connects to an interior vertex.  
We do not need to consider boundary chords between adjacent vertices since we are searching for 3-connected doodles.

For the case there are no boundary chords, we just call the stag 2 function enumerate_interior_choices.  Where there are boundary chords, we call 
enumerate_interior_vertices to consider all possible assignments of interior vertices across the chord regions.

Stage 2, enumerate_interior_vertices enumerate_interior_choices

The inside_vertices of a chord region must be joined to an interior vertex in order to have valency three, so the number of inside_vertices and the 
max_valency determine the min_num_interior_vertices required by each region.  Provided suffient interior vertices are available, the remaining interior 
vertices are allocated to the chord regions in all possible ways by enumerate_interior_vertices and then enumerate_interior_choices evaluates all possible 
ways to join an inside vertex in the boundary of a chord region to an interior vertex.

Each pattern of boundary chords, interior vertex assignment and interior choice are then passed to stage 3.

Stage 3, add_interior_edges.

The remaining edges in an incidence matrix produced by the first two stages either join a boundary vertex to an interior vertex or join two interior vertices.  
Since the dual graph has at most one edge between a pair of vertices, it is straightforward to identify the zero locations of I in the right-most n+1-2p columns 
that lie to the right of the leading diagonal and then select the appropriate number of locations to set to 1 in all possible ways.

add_interior_edges iterates through the possibilities, rejecting any that result in the creation of triangular regions or if they result in a vertex having valency 
less than three.

Each completed matrix I is checked to ensure the valency requirements of the doodle code are met and that the corresponding graph \Gamma is planar (planar_graph). 
For a planar graph, in valid_4_gon_graph, we may trace the boundary of a region by moving along an edge towards a crossing and turning right, since there is a unique 
choice for the subsequent edge.  In this manner, the regions of planar graphs are checked to ensure they are all four-sided and, if so, we say that the incidence 
matrix is admissible.  The dual graph represented by an admissible incidence matrix I is connected, (see the above paper for a proof).

Finally, doodle_gauss_code evaluates the dual doodle and determines the unoriented left preferred Gauss code for the doodle to provides a unique representationthat 
allows a list of distinct dual doodles to be determined.

Resultant doodles are classified as 3-connected or 4-connected by the function four_connected

Note that the only solution to the equations (*) above with n_i = 0 for i>3 is v=6, since we have to have n_3 = v+2 and 3*n_3 = 4v.  That means that
for more than 6 crossings we shall always choose an infinite region with at least 4 edges and so there will be at least 8 vertices in the boundary of
the dual graph's disc.

Command line syntax described in helpinfo function in this module.  Running 4-gon with no parameters displays usage information.

January 2023: initial version
August 2023 first release version

*******************************************************************************************************/

using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>

/******************** Global Variables **********************/
bool	CHORD_PATTERNS_ONLY = false;
bool	DEVELOPMENT_MODE = false;
bool	DRAW_GRAPHS = false;
bool	INTERIOR_CHOICES = false;
bool	QUIET_MODE = false;
bool	BOUNDARY_CHORDS_ONLY = false;
bool	OMIT_BOUNDARY_CHORDS = false;  // used for testing
bool	TEST_MODE = false;

bool	termination_request = false;
bool 	sigusr1_request = false;

ofstream 	output;
extern ofstream    debug;

#include <4-gon.h>


/********************* Function prototypes ***********************/
void helpinfo();
void debug_help();
void set_programme_long_option(char* cptr, string source);
void set_programme_short_option(char* cptr);
bool debug_setup(char* argv_ptr);
void check_debug_option_parameters(char* pptr, string option);
void print_prog_params (ostream& os, string level, string prefix);
int min_num_edges(int n2, int n1, int n0, int n2j2=0, int n2j1=0, int n1j1=0, int n1j2j1=0, int n1j2j2=0, int n1j2j2j1=0);
void add_interior_edges(matrix<int> incidence, int num_edges_to_add, result_data& search_results);
void enumerate_interior_choices(chord_data& pattern, vector<int>& valency, int max_valency, result_data& search_results);
void enumerate_interior_vertices(chord_data& pattern, int num_interior_vertices, vector<int>& region_allocation, int remaining_vertices, int first_region, vector<int>& valency, int max_valency,result_data& search_results);
bool four_connected (string gauss_code);

namespace util
{
	string itos(long n);
}

int diagramnum = 1;

/* when reading codes from an input file, sigterm_handler causes vlist to get to the end of the 
   current seed code, write out the list files and exit gracefully.  The signal has no effect when 
   enumerating peer codes or Gauss codes, use kill -9 (SIGKILL) to abort in these cases.
*/
void sigterm_handler (int sig) // kill -15, or just kill
{
	clog << "\n4-gon programme received SIGTERM, setting termination_request=true" << endl;
	termination_request = true;
}

/* sigusr1_handler causes vlist to report progress on the current task and continue. */
void sigusr1_handler (int sig) // kill -10
{
	clog << "\n4-gon programme received SIGUSR1, setting sigusr1_request=true" << endl;
	sigusr1_request = true;
}


/******************* Main Function ************************/

int main (int argc, char* argv[])
{

	if (argc == 1)
	{
		helpinfo();
		exit(0);
	}
	
	signal(SIGTERM, sigterm_handler); // register the SIGTERM handler
	signal(SIGUSR1, sigusr1_handler); // register the SIGUSR1 handler
	
	/* take a copy of the original command line in case we're debugging and remove
	   the debug details.  We will want to record the original command line for the debug file.
	*/
	
	vector<string> original_programme_arguments(argc); 
	for (int i = 0; i< argc; i++)
		original_programme_arguments[i] = string(argv[i]);
	
	vector<int> doodle_code;
	int num_crossings = 0;
	int max_code_index = 0;
	int last_code_index = 2; // acts as an index into doodle_code when reading input and then tells us the highest n-gon in the code.
	
	time_t start_time = time(0);
	
    /* Read the programme parameters and input files */
	for (int i=1; i < argc; i++)
	{
		if (*(argv[i]) == '-')
		{
			if (*(argv[i]+1) == '-')
				set_programme_long_option(argv[i], "command line");
			else 
				set_programme_short_option(argv[i]);
		}
		else if (num_crossings == 0) 
		{
			get_number(num_crossings, argv[i]);
			
			
			max_code_index = num_crossings/2;

			doodle_code = vector<int>(max_code_index+1);
		}
		else if (last_code_index < max_code_index) 
		{
			/* read the number of n-gons where n=last_code_index */
			last_code_index++;
			get_number(doodle_code[last_code_index],argv[i]);
		}
	}	

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << argv[i] << ' ';
	debug << endl;
}

	output.open("4-gon.out",ios_base::app);

    if (!output)
    {
		cerr << "\nError opening output file" << endl;
		exit(0);
    }

/*DEVELOPMENT_MODE is used for testing individual functions during development */
if (DEVELOPMENT_MODE)
{
		
	/* test of boundary vertex symmetry 
	cout << "num_chord_regions: ";
	int num_chord_regions;
	cin >> num_chord_regions;

	cout << "num_distributed_vertices: ";
	int num_distributed_vertices;
	cin >> num_distributed_vertices;

	chord_data pattern;
	pattern.chords = list<pair<int,int> >(num_chord_regions-1);
	pattern.min_num_interior_vertices = vector<int>(num_chord_regions);
	pattern.region = vector<region_data>(num_chord_regions);
	
	int num_boundary_vertices = 0;
	int num_interior_vertices = 0;
	
	vector<int> region_allocation(num_chord_regions);
*/


	cout << "input file: ";
	string filename;
	cin >> filename;
	
	ifstream input(filename);
	
	string gauss_code;
	while (getline(input,gauss_code))
	{
		cout << gauss_code;
		output << gauss_code;
		if (four_connected (gauss_code))
		{
			cout << "; 4-connected ";
			output << "; 4-connected ";
		}
		else
		{
			cout << "; 3-connected";
			output << "; 3-connected";
		}	
		cout << endl;
		output << endl;
	}
	
exit(0);

	bool reset_interior_choice(region_data& region, vector<int>& valency, int max_valency);
	bool increment_interior_choice(region_data& region, vector<int>& valency, int max_valency);

	region_data region;
	
	region.index = 0;
	
	cout << "num_boundary_vertices: ";
	cin >> region.num_boundary_vertices;

	cout << "num_inside_vertices: ";
	cin >> region.num_inside_vertices;

	cout << "num_interior_vertices: ";
	cin >> region.num_interior_vertices;

	cout << "first_interior_vertex: ";
	cin >> region.first_interior_vertex;

	region.boundary_vertices = vector<int>(region.num_boundary_vertices);
	region.inside_vertices = vector<int>(region.num_inside_vertices);
	region.interior_choice = vector<int>(region.num_inside_vertices);
	region.max_interior_choice = vector<int>(region.num_inside_vertices, region.first_interior_vertex+region.num_interior_vertices-1);
	cout << "boundary vertices: ";
	for (int i=0; i<region.num_boundary_vertices; i++)
		cin >> region.boundary_vertices[i];
	
	cout << "inside vertices: ";
	for (int i=0; i<region.num_inside_vertices; i++)
		cin >> region.inside_vertices[i];
	
	vector<int> valency(region.first_interior_vertex+region.num_interior_vertices);
//	for (int i=0; i<region.first_interior_vertex+region.num_interior_vertices; i++)
//		valency[i] = 1;
		
	int max_valency = 4;  // test value
	
	reset_interior_choice(region, valency, max_valency);
	
	while(increment_interior_choice(region, valency, max_valency)) {}
	
//    enumerate_interior_vertices(pattern, num_interior_vertices, region_allocation, num_distributed_vertices, 0, valency, max_valency, search_results);

	exit(0);
}

/* End of DEVELOPMENT_MODE code */
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	print_prog_params (debug, "basic", "");
}

	output << "\ndoodle_code: ";
	for (int i=0; i <= last_code_index; i++)
		output << doodle_code[i] << ' ';
	output << endl;
	output << "num_crossings = " << num_crossings << " max_code_index = " << max_code_index << endl;
	
	if (!QUIET_MODE)
	{
		cout << "\nnum_crossings = " << num_crossings << " max_code_index = " << max_code_index << ", doodle_code: ";
		for (int i=0; i <= last_code_index; i++)
			cout << doodle_code[i] << ' ';
		cout << endl;
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << original_programme_arguments[i] << ' ';
	debug << endl;
}
		
	int num_vertices_in_dual_graph = num_crossings+1; // we don't include the infinite vertex
	int num_edges_in_infinite_cycle = last_code_index;
	int num_4_gons_in_disc = num_crossings - num_edges_in_infinite_cycle;
	int num_boundary_vertices = 2*num_edges_in_infinite_cycle;
	int num_interior_vertices = num_vertices_in_dual_graph - num_boundary_vertices;
	int num_interior_edges = 2*num_crossings - 3*num_edges_in_infinite_cycle;
	int num_test_block_places = (num_boundary_vertices * num_interior_vertices - num_boundary_vertices/2) + (num_interior_vertices * (num_interior_vertices -1) / 2);
	
	if (TEST_MODE)
	{
		cout << "default num_edges_in_infinite_cycle = " << num_edges_in_infinite_cycle << ", do you want to change it? y/n: ";
		char ch;
		cin >> ch;
		
		if (ch == 'y')
		{
			cout << "new num_edges_in_infinite_cycle: ";
			cin >> num_edges_in_infinite_cycle;
			num_boundary_vertices = 2*num_edges_in_infinite_cycle;
			
if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << "TEST_MODE: specified " << num_edges_in_infinite_cycle << " num_edges_in_infinite_cycle" << endl;
		}
	}

	output << "num_vertices_in_dual_graph = " << num_vertices_in_dual_graph << endl;
	output << "num_edges_in_infinite_cycle = " << num_edges_in_infinite_cycle << endl;
	output << "num_4_gons_in_disc = " << num_4_gons_in_disc << endl;
	output << "num_boundary_vertices = " << num_boundary_vertices << endl;
	output << "num_interior_vertices = " << num_interior_vertices << endl;
	output << "num_interior_edges = " << num_interior_edges << endl;
	output << "test block size = " << num_interior_edges-num_edges_in_infinite_cycle << endl;
	output << "num_test_block_places = " << num_test_block_places << endl;

	if (!QUIET_MODE)
	{
		cout << "num_vertices_in_dual_graph = " << num_vertices_in_dual_graph << endl;
		cout << "num_edges_in_infinite_cycle = " << num_edges_in_infinite_cycle << endl;
		cout << "num_4_gons_in_disc = " << num_4_gons_in_disc << endl;
		cout << "num_boundary_vertices = " << num_boundary_vertices << endl;
		cout << "num_interior_vertices = " << num_interior_vertices << endl;
		cout << "num_interior_edges = " << num_interior_edges << endl;
		cout << "test block size = " << num_interior_edges-num_edges_in_infinite_cycle << endl;
		cout << "num_test_block_places = " << num_test_block_places << endl;
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)		
{
	debug << "num_vertices_in_dual_graph = " << num_vertices_in_dual_graph << endl;
	debug << "num_edges_in_infinite_cycle = " << num_edges_in_infinite_cycle << endl;
	debug << "num_4_gons_in_disc = " << num_4_gons_in_disc << endl;
	debug << "num_boundary_vertices = " << num_boundary_vertices << endl;
	debug << "num_interior_vertices = " << num_interior_vertices << endl;
	debug << "num_interior_edges = " << num_interior_edges << endl;
	debug << "test block size = " << num_interior_edges-num_edges_in_infinite_cycle << endl;
	debug << "num_test_block_places = " << num_test_block_places << endl;
}
	
	/* decrement doodle_code[last_code_index] so doodle code indicates the number of i-valent vertices in the dual graph */
	doodle_code[last_code_index]--;
	int max_valency = last_code_index;
	while (doodle_code[max_valency] == 0)
		max_valency--;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << "max_valency of vertices on the 4-gon disc = " << max_valency << endl;	

	/* we track the valency of interior vertices so we can correctly reset the interior choice of all inside vertices after 
	   the one whose choice has been incremented by increment_interior_choice to the smallest available interior vertex.
	*/
	vector<int> valency(num_vertices_in_dual_graph);
	
	matrix<int> boundary_incidence(num_vertices_in_dual_graph,num_vertices_in_dual_graph);
	for (int i=0; i< num_boundary_vertices; i++)
	{
		boundary_incidence[i][(i+1)%num_boundary_vertices] = 1;
		boundary_incidence[(i+1)%num_boundary_vertices][i] = 1;
	}


	result_data search_results(doodle_code,num_boundary_vertices);

	/* Identify the endpoints of all possible boundary chord positions 
	
	   Consider a disc with 2n vertices in the boundary and a set of disjoint chords drawn between a subset of those vertices.  Assume that is it possible to extend the disc 
	   and chords to the dual graph of a doodle D by adding outside edges and interior vertices as necessary.  We get the same doodle, up to the usual reflective symmetry, 
	   if we act upon the disc by any element of the dihedral group D_{2n}.  That means we may fix a cyclic numbering the of the boundary vertices, assume that the vertex at the
	   12 o'clock position is not joined to the infinite vertex, and consider the possible chord endpoints with respect to that numbering and orientation.  In particular all 
	   single chord patterns may be considered to be connected to the top vertex and connect to a boundary vertex in the right-hand semi-cirle of the disc boundary.
	
	   Any given set of chords divides the disc into polygonal regions, at least two of which contain only one chord in its boundary.  Orienting the boundary vertices clockwise, 
	   rotate the first such region we encounter so that the v-boundary vertex endpoint is at the 12 o'clock position and reflect if the e-boundary endpoint does not lie in the 
	   right hand semicircle of the disc.
	   
	   Thus, we need consider only patterns of chords where the first chord is incident with the boundary vertex in the 12 o'clock position and lies in the right-hand semi-circle 
	   and subsequent chords are always added to the left of the pattern.  Clearly boundary chords may intersect only at their endloints. 
	   
	   The interior of the chord regions are further subdivided (if necessary) to create the four-sided regions of a dual graph.  Suppose such a region has k boundary edges 
	   and is divided into r four-sided regions by a total of e edges and some number of interior vertices.   We have 4r = 2e - k, which has a solution only if k is even.
	   
	   Consider for example the case p=6 so that the disc has 12 vertices (i.e. a clock face) and consider the possible chords that we can construct from the vertex at 12 o'clock,
	   which we regard as vertex 0.  By the above observation on the number of edges in a region boundary, we could construct a chord only to the odd numbered vertices and, up to 
	   symmetry we only need consider chords from vertex 0 to vertices 3 or 5.
	
	   For the 2-chord cases, by symmetry we need only consider the ways to add a second chord amongst higher numbered vertices.  Thus, writing x-y if x is joined to y we have to 
	   consider the cases:         

	   0-3:  0-5, 0-7, 0-9
	         3-6, 3-8, 3-10
	         4-7, 4-9, 4-11
	         5-8, 5-10
	         6-9, 6-11
	         7-10
	         8-11
	         
	   0-5:  0-7, 0-9
	         5-8, 5-10
	         6-9, 6-11
	         7-10
	         8-11
		 
	   One can continue with three-chords etc. in the same vein.  The nett effect is that the only boundary chords that we need to consider are:		
			
	         0-3, 0-5, 0-7, 0-9
	         1-4, 1-6, 1-8, 1-10
	         2-5, 2-7, 2-9, 2-11
	         3-6, 3-8, 3-10
	         4-7, 4-9, 4-11
	         5-8, 5-10
	         6-9, 6-11
	         7-10
	         8-11
	         
		These are exactly the boundary chords joining an e-boundary vertex to a non-adjacent v-boundary vertex.

		In general, for 2p boundary vertices there are 2 * \Sum_1^{p-2} + p-2  = p(p-2) possible boundary chords.
	*/
	
	if (!BOUNDARY_CHORDS_ONLY)
	{
		/* First, determine the doodles that have no boundary chords */
		chord_data no_boundary_edge_chord_data;
		no_boundary_edge_chord_data.num_edges_to_add = num_interior_edges - num_edges_in_infinite_cycle;
		no_boundary_edge_chord_data.total_num_inside_vertices = num_boundary_vertices/2;
		no_boundary_edge_chord_data.incidence = boundary_incidence;
		no_boundary_edge_chord_data.min_num_interior_vertices = vector<int>(1);
		no_boundary_edge_chord_data.min_num_interior_vertices[0] = num_interior_vertices;
		no_boundary_edge_chord_data.region = vector<region_data>(1);
		
		region_data single_region_data;
		single_region_data.index = 0;
		single_region_data.num_boundary_vertices = num_boundary_vertices;
		single_region_data.num_boundary_edges = num_boundary_vertices;
		single_region_data.num_inside_vertices = num_boundary_vertices/2;
		single_region_data.num_interior_vertices = num_interior_vertices;
		single_region_data.first_interior_vertex = num_boundary_vertices;
		single_region_data.boundary_vertices = vector<int>(num_boundary_vertices);
		for (int i=0; i< num_boundary_vertices; i++)
			single_region_data.boundary_vertices[i] = i;
	
		single_region_data.inside_vertices = vector<int>(single_region_data.num_inside_vertices);
		for (int i=0; i< num_edges_in_infinite_cycle; i++)
			single_region_data.inside_vertices[i] = 2*i;
		
		single_region_data.interior_choice = vector<int>(single_region_data.num_inside_vertices);
		single_region_data.max_interior_choice = vector<int>(single_region_data.num_inside_vertices,num_vertices_in_dual_graph-1);
	
		no_boundary_edge_chord_data.region[0] = single_region_data;
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "  no-boundary_edge_chord_data:" << endl;
	write_chord_data(no_boundary_edge_chord_data,debug,"  ");
}
	
		if (INTERIOR_CHOICES)
			cout << "Interior choices for single region interior vertex allocation of " << num_interior_vertices << " vertices" << endl;			
	
		enumerate_interior_choices(no_boundary_edge_chord_data, valency, max_valency, search_results);
	}
	
	if (!OMIT_BOUNDARY_CHORDS && num_interior_vertices > 1) // If there is only one interior vertex we cannot have any boundary chords
	{
		/* Now determine possible boundary chord patterns */
		
		int num_boundary_chords = num_edges_in_infinite_cycle*(num_edges_in_infinite_cycle-2);

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): num_boundary_chords = " << num_boundary_chords << endl;

		vector<pair<int,int> > boundary_chords(num_boundary_chords);
		int index=0;

		for (int i=0; i < num_boundary_vertices; i++)
		{
			if (i>0)
			{
				for (int j=1; i+2*j+1 < num_boundary_vertices; j++)
					boundary_chords[index++] = pair<int,int>(i,i+2*j+1);
			}
			else
			{
				for (int j=1; 2*j+1 < num_boundary_vertices-1; j++)
					boundary_chords[index++] = pair<int,int>(0,2*j+1);
			}
		}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration): boundary_chords: ";
	
	for (size_t i=0; i< boundary_chords.size(); i++)
		debug << '(' << boundary_chords[i].first << ',' << boundary_chords[i].second << ") ";
	
	debug << endl;
}	
				
		if (TEST_MODE)
		{			
			if (CHORD_PATTERNS_ONLY && TEST_MODE)
			{
				cout << "default num_interior_vertices = " << num_interior_vertices << ", do you want to change it? ";
				char ch;
				cin >> ch;
				
				if (ch == 'y')
				{
					cout << "new num_interior_vertices: ";
					cin >> num_interior_vertices;
			
if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << "TEST_MODE: specified " << num_interior_vertices << " num_interior_vertices" << endl;								
				}
			}
			else
			{
				cout << "boundary_chords: ";
				
				for (size_t i=0; i< boundary_chords.size(); i++)
					cout << i << ":(" << boundary_chords[i].first << ',' << boundary_chords[i].second << ") ";
				
				cout << endl;
			}
		}


		/* Each chord pattern is represented by a chord_data class. Adding a boundary chord increases the numnber of chord regions by one.  We therefore 
		   create a list of all valid single-chord patterns and then successively take the first on the list, extend it in all possible ways to form two-chord 
		   patterns, which we push onto the back of the list, then we process the first pattern of the list and finally pop it from the list.
		*/
			
		list<chord_data> patterns;

		/* initialize patterns with single chord patterns that do not require too many interior vertices and record the min_num_interior_vertices
		   for each region.
		*/
		for (int i=3; i<= num_edges_in_infinite_cycle; i+=2) // i<= num_boundary_vertices/2
		{
			if ( (i==1 && num_interior_vertices < 7) || (i==3 && num_interior_vertices < 5) )
			   continue;
			   
			list<pair<int,int> > local_list;
			local_list.push_back(pair<int,int>(0,i));

			chord_data pattern(boundary_incidence,local_list);

			/* record the boundary chord in the incidence matrix of the pattern */
			pattern.incidence[0][i] = 1;
			pattern.incidence[i][0] = 1;

			pattern.num_edges_to_add = num_interior_edges - num_edges_in_infinite_cycle; 
			pattern.total_num_inside_vertices = num_edges_in_infinite_cycle - 1; // everything other than vertex zero is an inside vertex

			pattern.min_num_interior_vertices = vector<int>(2);
			if (i==1)
				pattern.min_num_interior_vertices = {6,1};
			else if (i==3)
				pattern.min_num_interior_vertices = {4,1};
			else
				pattern.min_num_interior_vertices = {1,1};


			pattern.region = vector<region_data>(2);
			
			pattern.region[0].index = 0;
			pattern.region[0].num_boundary_vertices = i+1;
			pattern.region[0].num_boundary_edges = i;
			pattern.region[0].num_inside_vertices = i/2;
			
			pattern.region[0].boundary_vertices = vector<int>(pattern.region[0].num_boundary_vertices);
			for (int j=0; j< pattern.region[0].num_boundary_vertices; j++)
				pattern.region[0].boundary_vertices[j] = j;
			
			pattern.region[0].inside_vertices = vector<int>(pattern.region[0].num_inside_vertices);
			for (int j=0; j< pattern.region[0].num_inside_vertices; j++)
				pattern.region[0].inside_vertices[j] = 2*j+2;

			pattern.region[0].interior_choice = vector<int>(pattern.region[0].num_inside_vertices);
			
			pattern.region[1].index = 1;
			pattern.region[1].num_boundary_vertices = num_boundary_vertices-i+1;
			pattern.region[1].num_boundary_edges = num_boundary_vertices-i;
			pattern.region[1].num_inside_vertices = (num_boundary_vertices-i)/2;
			
			pattern.region[1].boundary_vertices = vector<int>(pattern.region[1].num_boundary_vertices);
			for (int j=0; j< pattern.region[1].num_boundary_vertices; j++)
				pattern.region[1].boundary_vertices[j] = (i+j)%num_boundary_vertices;
			
			pattern.region[1].inside_vertices = vector<int>(pattern.region[1].num_inside_vertices);
			for (int j=0; j< pattern.region[1].num_inside_vertices; j++)
				pattern.region[1].inside_vertices[j] = 2*j+i+1;

			pattern.region[1].interior_choice = vector<int>(pattern.region[1].num_inside_vertices);

			pattern.region_map = vector<int>(num_boundary_vertices);
			for (int j=0; j< i; j++)
				pattern.region_map[j] = 0;
			for (int j=i; j< num_boundary_vertices; j++)
				pattern.region_map[j] = 1;
			
			patterns.push_back(pattern);
		}
	
		list<chord_data>::iterator pptr = patterns.begin();
		
		int pattern_count = 0;
		
		while(pptr != patterns.end())
		{
			pattern_count++;
			
			if (TEST_MODE && !CHORD_PATTERNS_ONLY)
			{
				cout << "num boundary chords ";
				int num_chords=0;
				cin >> num_chords;
	
				pptr->chords.pop_front();  // remove the default boundary chord
				
				if (num_chords >0)
				{
					int index;
					cout << "chord indices: ";
					
					for (int i=0; i< num_chords; i++)
					{
						cin >> index;
						pptr->push(boundary_chords[index]);
					}
				}

if (debug_control::DEBUG >= debug_control::SUMMARY)		

	debug << "TEST_MODE: specified " << num_chords << " boundary chords: " << endl;

				/* reset pptr incidence matrix */
				pptr->incidence = boundary_incidence;
	
				list<pair<int,int> >::iterator next_cptr = pptr->chords.begin();
				while (next_cptr != pptr->chords.end())
				{
					cout << "boundary chords: " << '(' << next_cptr->first << ',' << next_cptr->second << ") ";
					pptr->incidence[next_cptr->first][next_cptr->second] = 1;
					pptr->incidence[next_cptr->second][next_cptr->first] = 1;
					next_cptr++;
				}
				cout << endl;

				vector<region_data> new_regions(num_chords+1);
				for (int i=0; i<= num_chords; i++)
				{
					new_regions[i].index = i;
					cout << "region " << i << " num_boundary_vertices: ";
					cin >> new_regions[i].num_boundary_vertices;
					
					new_regions[i].boundary_vertices = vector<int>(new_regions[i].num_boundary_vertices);
					cout << "region " << i << " boundary_vertices: ";
					for (int j=0; j< new_regions[i].num_boundary_vertices; j++)
						cin >> new_regions[i].boundary_vertices[j];

					cout << "region " << i << " num_inside_vertices: ";
					cin >> new_regions[i].num_inside_vertices;

					new_regions[i].inside_vertices = vector<int>(new_regions[i].num_inside_vertices);
					new_regions[i].interior_choice = vector<int>(new_regions[i].num_inside_vertices);
					
					if (new_regions[i].num_inside_vertices> 0)
					{
						cout << "region " << i << " inside_vertices: ";
						for (int j=0; j< new_regions[i].num_inside_vertices; j++)
							cin >> new_regions[i].inside_vertices[j];
					}

				}				
				pptr->region = new_regions;

				pptr->total_num_inside_vertices = 0;
				for (int j=0; j<= num_chords; j++)
					pptr->total_num_inside_vertices += pptr->region[j].num_inside_vertices;

				pptr->min_num_interior_vertices = vector<int>(num_chords+1);
				cout << "min_num_interior_vertices for regions: ";
				for (int j=0; j<=num_chords; j++)
					cin >> pptr->min_num_interior_vertices[j];				
			}
		
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration): next chord pattern: " << endl;
		write_chord_data(*pptr,debug,"main (chord pattern enumeration):   ");
}


			/* Extend the pattern list before processing the current chord pattern by looking for all boundary chords greater than the last
			   chord in the pattern that do not intersect any current chord.  We require that no second endpoint in the current list lie
			   strictly between the endpoints of any prospective chord.
			   
			*/
			if (!TEST_MODE || CHORD_PATTERNS_ONLY)
			{										
				list<pair<int,int> >::iterator last_cptr = pptr->chords.end();
				last_cptr--;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): last boundary chord (" << last_cptr->first << ',' << last_cptr->second << ')' << endl;
					
				for (int i=0; i< num_boundary_chords; i++)
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): i = " << i << ": boundary chord (" << boundary_chords[i].first << ',' << boundary_chords[i].second << ")" << endl;
					
					if ((boundary_chords[i].first > last_cptr->first) || (boundary_chords[i].first == last_cptr->first && boundary_chords[i].second > last_cptr->second) )
					{
						bool acceptable_chord = true;							
						list<pair<int,int> >::iterator cptr = pptr->chords.begin();
						while (cptr != pptr->chords.end())
						{
							if (boundary_chords[i].first > cptr->first && cptr->second > boundary_chords[i].first && cptr->second < boundary_chords[i].second)
							{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): boundary chord (" << boundary_chords[i].first << ',' << boundary_chords[i].second << ") intersects (" << cptr->first << ',' << cptr->second << ')' << endl;
								acceptable_chord = false;
								break;
							}
							cptr++;
						}
						
						if (acceptable_chord)
						{																
							/* note the total minimum number of interior vertices in pptr */
							int total_min_interior_vertices = 0;
							for (size_t j=0; j< pptr->chords.size()+1; j++)
								total_min_interior_vertices += pptr->min_num_interior_vertices[j];
									
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): total_min_interior_vertices = " << total_min_interior_vertices << endl;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): boundary chord (" << boundary_chords[i].first << ',' << boundary_chords[i].second << ") does not intersect the pattern" << endl;
	
							/* Find the region of pptr contianing the new chord and determine the new  min_num_interior_vertices 
							
							   The boundary chord lies within a host region, which includes both ends of the chord in its boundary vertices.
							   
							   The host region must meet the disc boundary in an edge, since a region bounded entirely by boundary chords cannot be the host region for an additional acceptable chord.
							*/							
							
							int host_region;

							for (size_t j=0; j< pptr->chords.size()+1; j++)
							{
								if (find(pptr->region[j].boundary_vertices.begin(),pptr->region[j].boundary_vertices.end(),boundary_chords[i].first) != pptr->region[j].boundary_vertices.end() &&
								    find(pptr->region[j].boundary_vertices.begin(),pptr->region[j].boundary_vertices.end(),boundary_chords[i].second) != pptr->region[j].boundary_vertices.end())
								{
									host_region = j;
									break;
								}
							}
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): boundary chord lies within host_region " << host_region << endl;
	
							/* Look along region_map from edge 1 for and edge incident with the host_region and set first_endpoint to be boundary_chords[i].first 
							   or boundary_chords[i].second depending on which endpoint we encounter first.  We need only look from edge 1 since edge 0 is always 
							   part of region 0 and all subsequent boundary chords lie in region >=1.
							   
							   We will encounter the second end of the boundary chord before the first end if boundary_chords[i].first == 0 and also in cases like 
							   the pattern (2p,3), (3.5) to which we add boundary chord (3,7).
							*/

							int first_endpoint = boundary_chords[i].first;
							int second_endpoint = boundary_chords[i].second;

							/* in the following, if j+1 is an endpoint of a boundary chord then j != 2p-1, since 
							   the chord (0,2p-1) is one of the initializing set of chords.
							*/
							for (int j=1; j< num_boundary_vertices; j++)
							{
								if (pptr->region_map[j] == host_region)
								{
									if (j == boundary_chords[i].first || j+1 == boundary_chords[i].first) 
									{
										break;  // endpoints set correctly
									}
									else if (j+1 == boundary_chords[i].second)
									{
										swap(first_endpoint,second_endpoint);											
										break;
									}
									
								}
							}
														
							int num_sub_region_a_vertices = 0;
							int num_sub_region_b_vertices = 0;
							vector<int> sub_region_a_vertices(pptr->region[host_region].num_boundary_vertices);
							vector<int> sub_region_b_vertices(pptr->region[host_region].num_boundary_vertices);

							bool processing_sub_region_a = true;
							int num_host_region_boundary_vertices = pptr->region[host_region].num_boundary_vertices;

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration):  num_host_region_boundary_vertices = " << num_host_region_boundary_vertices << endl;
								
							/* trace around the host region, recording sub_region_a_vertices until we reach a boundary chord endpoint.  Switch to recording 
							   sub_region_b vertices until we get to the other end of the chord, then revert to recording sub_region_a_vertices.
							*/
							for (int j=0; j< num_host_region_boundary_vertices; j++)
							{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "main (chord pattern enumeration):  boundary_vertex = " << pptr->region[host_region].boundary_vertices[j] << ", processing_sub_region_a = " << processing_sub_region_a << endl;
								if (processing_sub_region_a)
									sub_region_a_vertices[num_sub_region_a_vertices++] = pptr->region[host_region].boundary_vertices[j];
								else
									sub_region_b_vertices[num_sub_region_b_vertices++] = pptr->region[host_region].boundary_vertices[j];
								
								
								if (pptr->region[host_region].boundary_vertices[j] == boundary_chords[i].first || pptr->region[host_region].boundary_vertices[j] == boundary_chords[i].second)
								{
									if (processing_sub_region_a)
									{
										processing_sub_region_a = false;
										sub_region_b_vertices[num_sub_region_b_vertices++] = pptr->region[host_region].boundary_vertices[j];
									}
									else
									{
										processing_sub_region_a = true;											
										sub_region_a_vertices[num_sub_region_a_vertices++] = pptr->region[host_region].boundary_vertices[j];
									}
								}
							}

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration):  num_sub_region_a_vertices = " << num_sub_region_a_vertices << ", num_sub_region_b_vertices = " << num_sub_region_b_vertices << endl;

							sub_region_a_vertices.resize(num_sub_region_a_vertices);
							sub_region_b_vertices.resize(num_sub_region_b_vertices);

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration):  sub_region_a_vertices: ";
	for (int j=0; j< num_sub_region_a_vertices; j++)
		debug << sub_region_a_vertices[j] << ' ';
	debug << endl;
	debug << "main (chord pattern enumeration):  sub_region_b_vertices: ";
	for (int j=0; j< num_sub_region_b_vertices; j++)
		debug << sub_region_b_vertices[j] << ' ';
	debug << endl;
	
}

							/* identify the inside vertices of each sub-region: the even boundary vertices of each sub-region that are not chord endpoints */
							vector<int>::iterator bptr;
							vector<int> sub_region_a_inside_vertices = sub_region_a_vertices;
							cptr = pptr->chords.begin();
							while (cptr != pptr->chords.end())
							{
								bptr = find(sub_region_a_inside_vertices.begin(),sub_region_a_inside_vertices.end(),cptr->first);
								if (bptr != sub_region_a_inside_vertices.end())
									*bptr = 1;  // something non-even
								bptr = find(sub_region_a_inside_vertices.begin(),sub_region_a_inside_vertices.end(),cptr->second);
								if (bptr != sub_region_a_inside_vertices.end())
									*bptr = 1;
								cptr++;
							}
							bptr = find(sub_region_a_inside_vertices.begin(),sub_region_a_inside_vertices.end(),boundary_chords[i].first);
							if (bptr != sub_region_a_inside_vertices.end())
								*bptr = 1; 
							bptr = find(sub_region_a_inside_vertices.begin(),sub_region_a_inside_vertices.end(),boundary_chords[i].second);
							if (bptr != sub_region_a_inside_vertices.end())
								*bptr = 1; 

							vector<int> sub_region_b_inside_vertices = sub_region_b_vertices;
							cptr = pptr->chords.begin();
							while (cptr != pptr->chords.end())
							{
								bptr = find(sub_region_b_inside_vertices.begin(),sub_region_b_inside_vertices.end(),cptr->first);
								if (bptr != sub_region_b_inside_vertices.end())
									*bptr = 1;  // something non-even
								bptr = find(sub_region_b_inside_vertices.begin(),sub_region_b_inside_vertices.end(),cptr->second);
								if (bptr != sub_region_b_inside_vertices.end())
									*bptr = 1;
								cptr++;
							}
							bptr = find(sub_region_b_inside_vertices.begin(),sub_region_b_inside_vertices.end(),boundary_chords[i].first);
							if (bptr != sub_region_b_inside_vertices.end())
								*bptr = 1; 
							bptr = find(sub_region_b_inside_vertices.begin(),sub_region_b_inside_vertices.end(),boundary_chords[i].second);
							if (bptr != sub_region_b_inside_vertices.end())
								*bptr = 1; 

							int num_sub_region_a_inside_vertices = 0;
							for (int j=0; j< num_sub_region_a_vertices; j++)
							{
								if (sub_region_a_inside_vertices[j]%2 == 0)
									num_sub_region_a_inside_vertices++;
							}

							int num_sub_region_b_inside_vertices = 0;
							for (int j=0; j< num_sub_region_b_vertices; j++)
							{
								if (sub_region_b_inside_vertices[j]%2 == 0)
									num_sub_region_b_inside_vertices++;
							}
																
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration):  num_sub_region_a_inside_vertices = " << num_sub_region_a_inside_vertices << endl;
	debug << "main (chord pattern enumeration):  bleached sub_region_a_vertices: ";
	for (int j=0; j< num_sub_region_a_vertices; j++)
		debug << sub_region_a_inside_vertices[j] << ' ';
	debug << endl;
	debug << "main (chord pattern enumeration):  num_sub_region_b_inside_vertices = " << num_sub_region_b_inside_vertices << endl;
	debug << "main (chord pattern enumeration):  bleached sub_region_b_vertices: ";
	for (int j=0; j< num_sub_region_b_vertices; j++)
		debug << sub_region_b_inside_vertices[j] << ' ';
	debug << endl;	
}


							/* check whether sub_region_a is adjacent to the last_chord */
							bool sub_region_a_adjacent_to_last_chord;
							if (find(sub_region_a_vertices.begin(),sub_region_a_vertices.end(),last_cptr->first) != sub_region_a_vertices.end() &&
							    find(sub_region_a_vertices.begin(),sub_region_a_vertices.end(),last_cptr->second) != sub_region_a_vertices.end())
							{
								sub_region_a_adjacent_to_last_chord = true;
							}
							else
							{
								sub_region_a_adjacent_to_last_chord = false;
							}

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration):  sub_region_a_adjacent_to_last_chord = " << sub_region_a_adjacent_to_last_chord << endl;
	
							/* assign new_region_map, start by marking the boundary edges in sub_region_a and sub_region_b as belonging to regions -1 and -2 respectively */
							vector<int> new_region_map = pptr->region_map;
							/* look for adjacent vertices in the boundary of sub-region a */
							for (int j=0; j< num_sub_region_a_vertices; j++)
							{
								int next_boundary_vertex = sub_region_a_vertices[(j+1)%num_sub_region_a_vertices];
								if (next_boundary_vertex == 0)
									next_boundary_vertex = num_boundary_vertices;

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "main (chord pattern enumeration):  sub_region_a vertex " << sub_region_a_vertices[j] << ", next_boundary_vertex = " << next_boundary_vertex << endl;
										
								if (next_boundary_vertex - sub_region_a_vertices[j] == 1)
								{
									/* adjacent boundary vertices */
									pair<int,int> test_chord = {sub_region_a_vertices[j],next_boundary_vertex%num_boundary_vertices};
									if (next_boundary_vertex == num_boundary_vertices)
										swap(test_chord.first,test_chord.second);

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "main (chord pattern enumeration):  adjacent boundary vertices "<< endl;
											
									new_region_map[sub_region_a_vertices[j]] = -1;
								}
							}

							/* look for adjacent vertices in the boundary of sub-region b */
							for (int j=0; j< num_sub_region_b_vertices; j++)
							{
								int next_boundary_vertex = sub_region_b_vertices[(j+1)%num_sub_region_b_vertices];
								if (next_boundary_vertex == 0)
									next_boundary_vertex = num_boundary_vertices;

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "main (chord pattern enumeration):  sub_region_b vertex " << sub_region_b_vertices[j] << ", next_boundary_vertex = " << next_boundary_vertex << endl;
										
								if (next_boundary_vertex - sub_region_b_vertices[j] == 1)
								{
									/* adjacent boundary vertices */
									pair<int,int> test_chord = {sub_region_b_vertices[j],next_boundary_vertex%num_boundary_vertices};
									if (next_boundary_vertex == num_boundary_vertices)
										swap(test_chord.first,test_chord.second);
											
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "main (chord pattern enumeration):  adjacent boundary vertices "<< endl;

									new_region_map[sub_region_b_vertices[j]] = -2;
								}
							}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration):  initial new_region_map: ";
	for (int j=0; j< num_boundary_vertices; j++)
		debug << new_region_map[j] << ' ';
	debug << endl;
}
							/* count the number of disc boundary edges in each sub_region */
							int num_sub_region_a_disc_edges = 0;
							int num_sub_region_b_disc_edges = 0;
							for (int j=0; j< num_boundary_vertices; j++)
							{
								if (new_region_map[j] == -1)
									num_sub_region_a_disc_edges++;
								else if (new_region_map[j] == -2)
									num_sub_region_b_disc_edges++;
							}

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration):  num_sub_region_a_disc_edges = " << num_sub_region_a_disc_edges << ", num_sub_region_b_disc_edges = " << num_sub_region_b_disc_edges << endl;
	
							int min_num_sub_region_a_interior_vertices = 0;
							int min_num_sub_region_b_interior_vertices = 0;
							
							total_min_interior_vertices -= pptr->min_num_interior_vertices[host_region];

							if (num_sub_region_a_vertices >4)
								min_num_sub_region_a_interior_vertices = 1;
							else if (num_sub_region_a_vertices == 4 && num_sub_region_a_inside_vertices > 0)
								min_num_sub_region_a_interior_vertices = 4;

							if (num_sub_region_b_vertices >4)
								min_num_sub_region_b_interior_vertices = 1;
							else if (num_sub_region_b_vertices == 4 && num_sub_region_b_inside_vertices > 0)
								min_num_sub_region_b_interior_vertices = 4;

							total_min_interior_vertices += (min_num_sub_region_a_interior_vertices + min_num_sub_region_b_interior_vertices);
									
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration):  min_num_sub_region_a_interior_vertices = " << min_num_sub_region_a_interior_vertices << ", min_num_sub_region_b_interior_vertices = " << min_num_sub_region_b_interior_vertices << endl;
	debug << "main (chord pattern enumeration):  new total_min_interior_vertices = " << total_min_interior_vertices << endl;
}											
							region_data sub_region_a;
							region_data sub_region_b;
							
							if (total_min_interior_vertices > num_interior_vertices)										
							{
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): new pattern requires too many interior vertices, ignoring" << endl;
							}
							else
							{
								/* first new sub-region */
								sub_region_a.num_boundary_vertices = num_sub_region_a_vertices;									
								sub_region_a.num_boundary_edges = num_sub_region_a_disc_edges;									
								sub_region_a.num_inside_vertices = num_sub_region_a_inside_vertices;
								sub_region_a.boundary_vertices = sub_region_a_vertices;
								sub_region_a.inside_vertices = vector<int>(num_sub_region_a_inside_vertices);
								int inside_index = 0;
								
								for (int k=0; k < num_sub_region_a_vertices; k++)
								{
									if (sub_region_a_inside_vertices[k]%2 == 0)
										sub_region_a.inside_vertices[inside_index++] = sub_region_a_inside_vertices[k];
								}									

								sub_region_a.interior_choice = vector<int>(sub_region_a.num_inside_vertices);
								
								/* second new sub-region */
								sub_region_b.num_boundary_vertices = num_sub_region_b_vertices;									
								sub_region_b.num_boundary_edges = num_sub_region_b_disc_edges;									
								sub_region_b.num_inside_vertices = num_sub_region_b_inside_vertices;
								sub_region_b.boundary_vertices = sub_region_b_vertices;
								sub_region_b.inside_vertices = vector<int>(num_sub_region_b_inside_vertices);
								inside_index = 0;
								
								for (int k=0; k < num_sub_region_b_vertices; k++)
								{
									if (sub_region_b_inside_vertices[k]%2 == 0)
										sub_region_b.inside_vertices[inside_index++] = sub_region_b_inside_vertices[k];
								}																		

								sub_region_b.interior_choice = vector<int>(sub_region_b.num_inside_vertices);

								/* complete the re-evaluation of new_region_map: set region -1 back to host_region and -2 to pptr->chords.size()+2, note that
								   this may not be the correct final numbering for these regions; the correct numbering will be determined by region_perm below.
								*/
								int new_num_regions = pptr->chords.size()+2;
								
								for (int j=0; j< num_boundary_vertices; j++)
								{
									if (new_region_map[j] == -1)
										new_region_map[j] = host_region;
									else if (new_region_map[j] == -2)
										new_region_map[j] = new_num_regions-1;
								}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration):  final new_region_map: ";
	for (int j=0; j< num_boundary_vertices; j++)
		debug << new_region_map[j] << ' ';
	debug << endl;
}
								
								vector<int> region_perm(new_num_regions,-1);
								int region_index = 0;
								
								for (int j=0; j< num_boundary_vertices; j++)
								{
									if (region_perm[new_region_map[j]] == -1)
									{
										region_perm[new_region_map[j]] = region_index++;
									}
								}							
								
								if (region_index < new_num_regions)
								{
									for (size_t j=0; j< pptr->chords.size()+1; j++)
									{
										if (pptr->region[j].num_boundary_edges == 0) // this can't be the host_region, nor can it be new_num_regions-1
										{
											region_perm[j] = region_index++;
										}
									}
									
									if (num_sub_region_a_disc_edges == 0)
										region_perm[host_region] = region_index++;
									
									if (num_sub_region_b_disc_edges == 0)
										region_perm[new_num_regions-1] = region_index++;
								}		

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration):  region_perm: ";
	for (int j=0; j< new_num_regions; j++)
		debug << region_perm[j] << ' ';
	debug << endl;
}

								vector<region_data> new_regions(pptr->chords.size()+2);
								vector<int> new_min_num_interior_vertices(pptr->chords.size()+2);
								vector<int> inverse_region_perm(new_num_regions);
								for (int j=0; j< new_num_regions; j++)
									inverse_region_perm[region_perm[j]] = j;
								
								for (int j=0; j< new_num_regions; j++)
								{
									if (inverse_region_perm[j] == host_region)
									{
										new_regions[j] = sub_region_a;
										new_min_num_interior_vertices[j] = min_num_sub_region_a_interior_vertices;
									}
									else if (inverse_region_perm[j] == new_num_regions-1)
									{
										new_regions[j] = sub_region_b;
										new_min_num_interior_vertices[j] = min_num_sub_region_b_interior_vertices;
									}
									else
									{
										new_regions[j] = pptr->region[inverse_region_perm[j]];										
										new_min_num_interior_vertices[j] = pptr->min_num_interior_vertices[inverse_region_perm[j]];
									}
								}
								
								for (int j=0; j< new_num_regions; j++)
								{
									new_regions[j].index = j;
								}

								chord_data next_pattern = *pptr;
								next_pattern.push(boundary_chords[i]);
								
								/* record the new boundary chord in the incidence matrix of the next pattern */
								next_pattern.incidence[boundary_chords[i].first][boundary_chords[i].second] = 1;
								next_pattern.incidence[boundary_chords[i].second][boundary_chords[i].first] = 1;
								
								next_pattern.region = new_regions;
								next_pattern.min_num_interior_vertices = new_min_num_interior_vertices;
								next_pattern.region_map = vector<int>(num_boundary_vertices);
								
								for (int j=0; j< num_boundary_vertices; j++)
									next_pattern.region_map[j] = region_perm[new_region_map[j]];
								
								next_pattern.total_num_inside_vertices = 0;
								for (size_t j=0; j< next_pattern.region.size(); j++)
									next_pattern.total_num_inside_vertices += next_pattern.region[j].num_inside_vertices;

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "main (chord pattern enumeration): updated pattern: " << endl;
		write_chord_data(next_pattern,debug,"main (chord pattern enumeration):   ");
}

								patterns.push_back(next_pattern);
							}
						}
					}
					
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): end processing i = " << i << ": boundary chord (" << boundary_chords[i].first << ',' << boundary_chords[i].second << ")" << endl;
						
				}
			}
			
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "main (chord pattern enumeration): chord_data before interior vertices:" << endl;
	write_chord_data(*pptr,debug,"  ");
}
			int num_chord_regions = pptr->chords.size()+1;
			int num_vertices_to_allocate = num_interior_vertices;
			for (int i=0; i< num_chord_regions;i++)
				num_vertices_to_allocate -= pptr->min_num_interior_vertices[i];

if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "main (chord pattern enumeration): num_vertices_to_allocate = " << num_vertices_to_allocate << endl;

			if (TEST_MODE)
			{
				cout << "chord regions min_num_interior_vertices: ";
				for (int i=0; i< num_chord_regions;i++)
					cout << pptr->min_num_interior_vertices[i] << ' ';
				cout << endl;
			}
	
			vector<int> region_allocation(num_chord_regions);

			if (INTERIOR_CHOICES)
			{
				cout << "Interior choices for " << num_chord_regions << " regions, boundary chords:";				
				list<pair<int,int> >::iterator next_cptr = pptr->chords.begin();
				while (next_cptr != pptr->chords.end())
				{
					cout << " (" << next_cptr->first << ',' << next_cptr->second << ')';
					next_cptr++;
				}
								
				cout << ", with inside vertex count ";				
				for (int i=0; i< num_chord_regions;i++)
					cout << pptr->region[i].num_inside_vertices << ' ';
				cout << endl;
			}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "enumerate interior vertices for " << num_chord_regions << " regions, with inside vertex count ";
	for (int i=0; i< num_chord_regions;i++)
		debug << pptr->region[i].num_inside_vertices << ' ';
	debug << endl;
}			

			
			if (!CHORD_PATTERNS_ONLY)
				enumerate_interior_vertices(*pptr, num_interior_vertices, region_allocation,num_vertices_to_allocate,0, valency, max_valency,search_results); //first_region=0, where to allocate the first additional interior vertex			
			
			pptr++;
			
			if (TEST_MODE && !CHORD_PATTERNS_ONLY)
				break;
		}

if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << "total of " << pattern_count << " patterns" << endl;
	
	}
		
	if (!INTERIOR_CHOICES)
		cout << "found " << search_results.doodle_list.size() << " distinct doodles" << endl;
		
	output << "found " << search_results.doodle_list.size() << " distinct doodles" << endl;
	list<string>::iterator sptr = search_results.doodle_list.begin();
	while (sptr != search_results.doodle_list.end())
	{
		cout << *sptr;
		output << *sptr;
		if (four_connected (*sptr))
		{
			cout << "; 4-connected ";
			output << "; 4-connected ";
		}
		else
		{
			cout << "; 3-connected";
			output << "; 3-connected";
		}		
		cout << endl;
		output << endl;
		
		sptr++;
	}
	

	time_t end_time = time(0);
	double time_exp = difftime(end_time, start_time);
	int dd = static_cast<int>(time_exp/3600/24);
	time_exp -= dd*3600*24;
	int hh = static_cast<int>(time_exp/3600);
	time_exp -= hh*3600;
	int mm = static_cast<int>(time_exp/60);
	time_exp -= mm*60;
	int ss = static_cast<int>(time_exp);
	output << "\nRunning time: " << dd << " days " << hh << " hours " << mm	<< " minutes " << ss << " seconds" << endl;

	output.close();
	

	return 0;
} /* End of main */

void helpinfo()
{
	cout << "\nUsage: 4-gon [-<short-option>] [--<long-option>] <num-crossings> <doodle-code>";
	cout << "\n<short-option>:";
	cout << "\n       b: boundary chords only (used for testing)";
	cout << "\n       c: chords patterns only (used for testing), turns on b";
	cout << "\n       B: omit boundary chords (used for testing)";
	cout << "\n       i: show initial interior vertex choices only";
	cout << "\n       d: draw graphs of valid 4-gon graphs";
	cout << "\n       Q: quiet mode (only write doodles to cout)";
	cout << "\n       T: test mode";
	cout << "\n       X: development mode";
	cout << "\n       #: debug, see #H! for details";
	cout << "\n<long-option>:";
	cout << "\n       development: same as -X";
	cout << "\n       draw: same as -d";
	cout << "\n       test: same as -T";
	cout << endl;
	exit(0);
}

/***************  Functions required for setting up vlist specific debug **************/

void set_main_debug_option_parameter(char* pptr, string option);
void set_main_debug_option(char* start, char* end)
{
	char  loc_buf[end-start+2];
	char* c1 = start;
	char* c2 = loc_buf;

	/* if both start and end are zero, display debug help information.
	   this has been included here in this manner so that each time a debug option
	   is added, the help will be updated (hopefully!)
	*/
	if (start == 0 && end == 0)
	{
		set_main_debug_option_parameter(0,"f-gon");
		return;
	}

	do
	{
		*c2++ = *c1++;
	} while (c1 <= end);
	
	*c2 = '\0';

	char* pptr = strchr(loc_buf,'{');
	if (pptr)
	{
		*pptr++ = '\0';
	}

	/* now, even if there are parameters, the first part of loc_buf 
	   is a C-string that identifies the option */
	
	if (!strcmp(loc_buf,"f-gon"))
	{
		debug_control::DEBUG = debug_control::INTERMEDIATE;
		debug << "main::set_main_debug_option: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
		
		if (pptr)
		{
			/* check for any parameters */
			check_debug_option_parameters(pptr, "f-gon");
		}
	}

	debug.flush();

}

void set_main_debug_option_parameter(char* pptr, string option)
{
	if (option == "f-gon")
	{
		if (!pptr)
		{
			cout << "\t\tf-gon{summary|1:basic|2:intermediate|3:detail|4}, integer: default 0=off, no parameters sets summary" << endl;
		}
		else
		{
			if (!strcmp(pptr,"summary") || !strcmp(pptr,"1") )
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			if (!strcmp(pptr,"basic") || !strcmp(pptr,"2") )
			{
				debug_control::DEBUG = debug_control::BASIC;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::BASIC" << endl;		
			}
			else if (!strcmp(pptr,"intermediate") || !strcmp(pptr,"3"))
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			else if (!strcmp(pptr,"detail") || !strcmp(pptr,"4"))
			{
				debug_control::DEBUG = debug_control::DETAIL;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::DETAIL" << endl;		
			}
			else if (!strcmp(pptr,"exhaustive") || !strcmp(pptr,"5"))
			{
				debug_control::DEBUG = debug_control::EXHAUSTIVE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::EXHAUSTIVE" << endl;		
			}
		}
	}
}

void main_display_default_options()
{
	cout << "\t\tf-gon{summary}" << endl;
}

void set_programme_short_option(char* cptr)
{
	if (strchr(cptr, 'H') && strchr(cptr, '!'))
	{
		if (strchr(cptr,'#'))
		{
			debug_help();
		}
		else
		{					
			helpinfo();
		}
	}

	char* dptr = strchr(cptr, '#');
	if (dptr)
	{
		debug.open ("4-gon.dbg"); 

		if (!debug)
		{
			cerr << "\nError opening debug file\n";
			exit(0);
		}
		else
		{
			debug << "Debug information from 4-gon\n\n";
			debug << boolalpha;
		}

		if (!debug_setup(cptr))
		{
			debug_control::DEBUG = debug_control::SUMMARY;
			debug << "main: default debug options set" << endl;
		}
		
	}

	if (strchr(cptr, 'B'))
	{
		OMIT_BOUNDARY_CHORDS = true;
	}

	if (strchr(cptr, 'b'))
	{
		BOUNDARY_CHORDS_ONLY = true;
	}

	if (strchr(cptr, 'c'))
	{
		BOUNDARY_CHORDS_ONLY = true;
		CHORD_PATTERNS_ONLY = true;
	}

	if (strchr(cptr, 'd'))
	{
		DRAW_GRAPHS = true;
	}

	if (strchr(cptr, 'i'))
	{
		INTERIOR_CHOICES = true;
	}

	if (strchr(cptr, 'Q'))
	{
		QUIET_MODE = true;
	}

	if (strchr(cptr, 'T'))
	{
		TEST_MODE = true;
	}
	
	if (strchr(cptr, 'X'))
	{
		DEVELOPMENT_MODE = true;
	}
}

void set_programme_long_option(char* cptr, string source)
{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "set_programme_long_option: passed " << cptr << " from " << source << endl;

	char  loc_buf[40];
	char* c1;
	char* c2;

	c1 = cptr;

	/* take out any leading -- */
	while (*c1 == '-')
		c1++;
	
	c2 = loc_buf;
	while (isalpha(*c1) || isdigit(*c1) || *c1 == '-' || *c1 == '_')
		*c2++ = *c1++;
    *c2 = '\0';
	
	if (!strcmp(loc_buf,"development"))
	{
		DEVELOPMENT_MODE = true;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "set_programme_long_option: DEVELOPMENT_MODE read from " << source << ", DEVELOPMENT_MODE = " << DEVELOPMENT_MODE << endl;
	}
	else if (!strcmp(loc_buf,"draw"))
	{
		DRAW_GRAPHS = true;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "set_programme_long_option: DRAW_GRAPHS read from " << source << ", DRAW_GRAPHS = " << DRAW_GRAPHS << endl;
	}
	else if (!strcmp(loc_buf,"test"))
	{
		TEST_MODE = true;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "set_programme_long_option: TEST_MODE read from " << source << ", TEST_MODE = " << TEST_MODE << endl;
	}
}

void print_prog_params (ostream& os, string level, string prefix)
{
    os << prefix << "DEVELOPMENT_MODE = " << DEVELOPMENT_MODE << endl;
    os << prefix << "TEST_MODE = " << TEST_MODE << endl;
    os << prefix << "BOUNDARY_CHORDS_ONLY = " << BOUNDARY_CHORDS_ONLY << endl;
    os << prefix << "CHORD_PATTERNS_ONLY = " << CHORD_PATTERNS_ONLY << endl;
    os << prefix << "OMIT_BOUNDARY_CHORDS = " << OMIT_BOUNDARY_CHORDS << endl;
}

