/**********************************************************************************************************************
                               4-gon

A minimal planar doodle may be characterized by a "doodle code" that lists the number of n-gon regions in a diagram of 
the doodle, which is unique up to the choice of the point at infinity.  Thus a doodle code has the form 

   n n_3 n_4,...,n_k,  where n is the number of crossings and n_i is the number of i-gon regions.
   
Note that we only support codes with less than 19 crossings, as described below.  

Also, we support searches only for prime minimal planar doodles, those for which the dual graph \Gamma' is a disc

By Euler's formula, n_3+n_4+...+n_k = n+2 and counting edges we have \Sum i*n_i = 2*num_edges = 4*n                   (*)
but not every such code corresponds to a realizable minimal planar doodle.

Given a doodle code, 4-gon looks for minimal planar doodles with that code.  Any doodle diagram has a dual graph, \Gamma 
that divides the 2-sphere into 4-gons.  Moreover, the dual graph corresponding to the compact regions of the doodle 
complement, \Gamma', is a the 1-skeleton of a "4-gon disc" disc, D', with the property that every alternate vertex of \Gamma' 
that lies in the boundary of D' is joined by an edge to the "infinite" vertex of \Gamma.

If we select a region with p edges as the infinite region we have a 4-gon disc with the following properties:

0. \Gamma has n+2 vertices, 2n edges and n regions

1. There are n+1 vertices in \Gamma'

2. There are 2p edges and 2p vertices in the boundary of the 4-gon disc, so there are n+1-2p non-boundary vertices in the 4-gon disc

3. There are n-p 4-gons in the disc

4. There are n_p-1 vertices of valency p in \Gamma' and for i=3,4,5..., i != p, there are n_i vertices of valency i in \Gamma'

5. Every 4-gon in the disc must be adjacent to at least two other 4-gons in the disc.  Otherwise, the 4-gon has three edges in the 
   disc boundary and we would have a bigon in the doodle.

The strategy for the 4-gon search is therefore to choose a region with the maximal number of edges as the infinite region and attempt 
to construct a 4-gon disc.  The disc has:

    n-p      4-gons (since there are p 4-gons outside the disc, incident with the infinite vertex)
    n+1-2p   interior vertices
    2n-3p    interior edges (2n edges in total, 2p in the boundary of the disc and p incident with the infinite vertex)

As noted above, every alternate vertex in the 4-gon disc boundary is joined by an edge to the infinite vertex and therefore has valency
at least three.  The other vertices in the disc boundary must therefore be joined to an interior vertex if the number of crossings is 
less than 19.  This is because, by 5 above, the smallest 4-gon disc in which two boundary vertices are joined by an edge is the one shown 
below

           |     |     |     |         In this graph the edges above and below the 4-gons (those that do not terminate at a +) connect 
           +--+--+     +--+--+         to the "infinite" vertex
           |  |  |  |  |  |  |
           +--+--+--+--+--+--+         Any smaller graph would include a 2-valent vertex, which corresponds to a doodle bigon
           |  |  |  |  |  |  |         and therefore cannot correspond to a minimal doodle.
           +--+--+--+--+--+--+
           |     |     |     |         There are 19 regions to this 4-gon graph, hence it corresponds to a doodle with 19 crossings
   

We assume that the vertices of the dual graph are numbered from zero and that vertex zero is not joined to the infinite vertex.  The
4-gon disc boundary vertices are therefore 0,...,2p-1.  We than number the n+1-2p interior vertices 2p,...,n and finally number the 
infinite vertex n+1.

We enumerate all possible choices of interior vertex for the p vertices in the disc boundary that are not joined to the infinite vertex in 
such a way  as to avoid effectively renumbering interior vertices.  Thus, boundary vertex zero is always joined to interior vertex 2p 
(see the comment below for further details).  Using 4. above we also check that we can connect the even vertices in the disc boundary to 
interior edges without exceeding the maximum valency for any vertex in the dual graph.  

For each set of interior vertex choices, we check that there are sufficient remaining edges (within the 2n-3p limit above) to bring all
the interior vertices to be at least 3-valent.  If there are, we call add_interior_edges to test systematically adding the remaining edges
to the interior graph.  Given that we are assuming the number of crossings is less than 19, these additional edges must join a boundary vertex 
to an interior vertex.  Each possible assignment of additional interior edges is checked to ensure the valency requirements are met and then the
Open Graph Drawing Framework library is used to test whether the resultant graph is planar.  If it is, we check that all the regions of the graph 
are 4-gons and, if that is the case, proceed to evaluate the left prefrred doodle Gauss data of the dual doodle and produce a list of distinct
such doodles, up to reflection (in a line in the plane).

Note that the only solution to the equations (*) above with n_i = 0 for i>3 is v=6, since we have to have n_3 = v+2 and 3*n_3 = 4v.  That means that
for more than 6 crossings we shall always choose an infinite region with at least 4 edges and so there will be at least 8 vertices in the boundary of
the dual graph's disc.

The dual graph constructed by the enumeration will always be connected.  Otherwise there would have to be a sub-graph with interior vertices disconected from the 
interior choices.  If the interior choice joins all boundary vertices to the same vertex, there would be n-2p disconnected interior vertices and there are 2n-4p 
additional interior edges.  Thus, the disconnected interior subgraph would have at least twice the number of edges than vertices and satisfies the search criteria 
of being planar and having four-sided regions.  Having four-sided regions, each component of this disconnected subgraph would determine a cellular subdivision of 
the sphere by itself (the component's "infinite" region would also be four-sided), so each component has Euler characteristic 2 and satisfies 4f=2e, since the 
regions are all four-sided. Thus each component has e=2f and v=f+2 from Euler's formula but then f = v-2 and e = 2v-4, so there  cannot be at least twice the number
of edges than vertices.


Jan 2023: initial version

*******************************************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>

/******************** Global Variables **********************/
bool	DEVELOPMENT_MODE = false;
bool	DRAW_GRAPHS = false;
bool	INTERIOR_CHOICES = false;
bool	QUIET_MODE = false;
bool	TEST_MODE = false;

bool	termination_request = false;
bool 	sigusr1_request = false;

ofstream 	output;
extern ofstream    debug;

#include <4-gon.h>


/********************* Function prototypes ***********************/
void helpinfo();
void debug_help();
void set_programme_long_option(char* cptr, string source);
void set_programme_short_option(char* cptr);
bool debug_setup(char* argv_ptr);
void check_debug_option_parameters(char* pptr, string option);
void print_prog_params (ostream& os, string level, string prefix);
int min_num_edges(int n2, int n1, int n0, int n2j2=0, int n2j1=0, int n1j1=0, int n1j2j1=0, int n1j2j2=0, int n1j2j2j1=0);
void add_interior_edges(matrix<int> incidence, int num_interior_edges, int num_edges_to_add, int num_boundary_vertices, vector<int>& max_valency_count, 
                        int& solution_count, list<string>& doodle_list, vector<int>& doodle_code, vector<int>& interior_choice);
void set_interior_choices(chord_data pattern);

namespace util
{
	string itos(long n);
}

int diagramnum = 1;

/* when reading codes from an input file, sigterm_handler causes vlist to get to the end of the 
   current seed code, write out the list files and exit gracefully.  The signal has no effect when 
   enumerating peer codes or Gauss codes, use kill -9 (SIGKILL) to abort in these cases.
*/
void sigterm_handler (int sig) // kill -15, or just kill
{
	clog << "\nvlist programme received SIGTERM, setting termination_request=true" << endl;
	termination_request = true;
}

/* sigusr1_handler causes vlist to report progress on the current task and continue. */
void sigusr1_handler (int sig) // kill -10
{
	clog << "\nvlist programme received SIGUSR1, setting sigusr1_request=true" << endl;
	sigusr1_request = true;
}


/******************* Main Function ************************/

int main (int argc, char* argv[])
{

	if (argc == 1)
	{
		helpinfo();
		exit(0);
	}
	
	signal(SIGTERM, sigterm_handler); // register the SIGTERM handler
	signal(SIGUSR1, sigusr1_handler); // register the SIGUSR1 handler
	
	/* take a copy of the original command line in case we're debugging and remove
	   the debug details.  We will want to record the original command line for the debug file.
	*/
	
	vector<string> original_programme_arguments(argc); 
	for (int i = 0; i< argc; i++)
		original_programme_arguments[i] = string(argv[i]);
	
	vector<int> doodle_code;
	int num_crossings = 0;
	int max_code_index = 0;
	int last_code_index = 2; // acts as an index into doodle_code when reading input and then tells us the highest n-gon in the code.
	
	time_t start_time = time(0);
	
    /* Read the programme parameters and input files */
	for (int i=1; i < argc; i++)
	{
		if (*(argv[i]) == '-')
		{
			if (*(argv[i]+1) == '-')
				set_programme_long_option(argv[i], "command line");
			else 
				set_programme_short_option(argv[i]);
		}
		else if (num_crossings == 0) 
		{
			get_number(num_crossings, argv[i]);
			
			
			max_code_index = num_crossings/2;
//			max_code_index = (num_crossings+1)/2;

			doodle_code = vector<int>(max_code_index+1);
		}
		else if (last_code_index < max_code_index) 
		{
			/* read the number of n-gons where n=last_code_index */
			last_code_index++;
			get_number(doodle_code[last_code_index],argv[i]);
		}
/*		else
		{
			cout << "Usage vlist [-<short-options>] [--<long-options>] [<input_invariant_file>] [<input_codes_file>]\n";
			exit(0);
		}
*/
	}	

/*DEVELOPMENT_MODE is used for testing individual functions during development */
if (DEVELOPMENT_MODE)
{
	/* test of boundary vertex symmetry */
	cout << "Number of boundary vertices: ";
	int num_boundary_vertices;
	cin >> num_boundary_vertices;
	int num_edges_in_infinite_cycle = num_boundary_vertices/2;
	
	int num_boundary_chords = (num_edges_in_infinite_cycle-2)*(num_edges_in_infinite_cycle-2);
	vector<pair<int,int> > boundary_chords(num_boundary_chords);
	int index=0;
	for (int i=num_boundary_vertices; i >= 6; i--)
	{
		for (int j=0; j< i/2 -2; j++)
		{
			if (i%2 == 1)
			{
//				boundary_chords[index++] = pair<int,int>(i%num_boundary_vertices,3+2*j+1);
				boundary_chords[index++] = pair<int,int>(i,3+2*j+1);
			}
			else
			{
//				boundary_chords[index++] = pair<int,int>(i%num_boundary_vertices,3+2*j);
				boundary_chords[index++] = pair<int,int>(i,3+2*j);
			}
		}
	}

	cout << "\nboundary_chords: ";
	
	for (size_t i=0; i< boundary_chords.size(); i++)
		cout << '(' << boundary_chords[i].first << ',' << boundary_chords[i].second << ") ";
	
	cout << endl;
	
		
	list<chord_data> patterns;
	
//	for (int i=3; i< num_boundary_vertices-1; i+=2)
	for (int i=3; i<= num_edges_in_infinite_cycle; i+=2) // i<= num_boundary_vertices/2
	{
		list<pair<int,int> > local_list;
//		local_list.push_back(pair<int,int>(0,i));
		local_list.push_back(pair<int,int>(num_boundary_vertices,i));
		patterns.push_back(chord_data(local_list));
	}
	
	cout << "patterns size = " << patterns.size() << endl;
	
	list<chord_data>::iterator pptr = patterns.begin();
	
	int count = 0;
		
	while(pptr != patterns.end())
	{
		count++;
		
		cout << "  pattern length = " << pptr->chords.size() << ": ";

		/* count the minimum number of interior vertices required by this pattern of boundary chords */	
		list<pair<int,int> >::iterator next_cptr = pptr->chords.begin();
		cout << '(' << next_cptr->first << ',' << next_cptr->second << ") ";

		int num_interior_vertices_required = (next_cptr->second == 3? 4:1);
		list<pair<int,int> >::iterator last_cptr = next_cptr;
		next_cptr++;
		
		while (next_cptr!= pptr->chords.end())
		{

			/* if the next chord is parallel to the last, we don't necessarily need any interior vertices, though
			   we could add four to such a region if enough are available.
			*/
			if (next_cptr->first != last_cptr->first -1 || next_cptr->second != last_cptr->second +1)
				num_interior_vertices_required++;
			
//			if (next_cptr->first == last_cptr->first -1 && next_cptr->second == last_cptr->second +1)
//				num_interior_vertices_required+=4;
//			else
//				num_interior_vertices_required++;
			
			cout << '(' << next_cptr->first << ',' << next_cptr->second << ") ";
			last_cptr++;
			next_cptr++;
		}
		
		if (last_cptr->first - last_cptr->second == 3)
			num_interior_vertices_required+=4;
		else
			num_interior_vertices_required++;
			
		cout << ", requires " << num_interior_vertices_required << " interior vertices" << endl;
		
		for (int i=0; i< num_boundary_chords; i++)
		{
			if (boundary_chords[i].first <= last_cptr->first)
			{
				if (boundary_chords[i].second >= last_cptr->second && boundary_chords[i] != *last_cptr)
				{
					chord_data next_pattern = *pptr;
					next_pattern.push(boundary_chords[i]);
					patterns.push_back(next_pattern);
				}
			}
		}

		pptr++;
	}
		
	cout << "total of " << count << " patterns" << endl;
	exit(0);
}

/* End of DEVELOPMENT_MODE code */
	
	output.open("4-gon.out");

    if (!output)
    {
		cerr << "\nError opening output file" << endl;
		exit(0);
    }
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	print_prog_params (debug, "basic", "");
}

	if (!QUIET_MODE)
	{
		cout << "\nnum_crossings = " << num_crossings << " max_code_index = " << max_code_index << ", doodle_code: ";
		for (int i=0; i <= last_code_index; i++)
			cout << doodle_code[i] << ' ';
		cout << endl;
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "full command line: ";
	for (int i = 0; i< argc; i++)
		debug << original_programme_arguments[i] << ' ';
	debug << endl;
}
		
	int num_vertices_in_dual_graph = num_crossings+1; // we don't include the infinite vertex
	int num_edges_in_infinite_cycle = last_code_index;
	int num_4_gons_in_disc = num_crossings - num_edges_in_infinite_cycle;
	int num_boundary_vertices = 2*num_edges_in_infinite_cycle;
	int num_interior_vertices = num_vertices_in_dual_graph - num_boundary_vertices;
	int num_interior_edges = 2*num_crossings - 3*num_edges_in_infinite_cycle;
	int num_test_block_places = (num_boundary_vertices * num_interior_vertices - num_boundary_vertices/2) + (num_interior_vertices * (num_interior_vertices -1) / 2);

	if (!QUIET_MODE)
	{
		cout << "num_vertices_in_dual_graph = " << num_vertices_in_dual_graph << endl;
		cout << "num_edges_in_infinite_cycle = " << num_edges_in_infinite_cycle << endl;
		cout << "num_4_gons_in_disc = " << num_4_gons_in_disc << endl;
		cout << "num_boundary_vertices = " << num_boundary_vertices << endl;
		cout << "num_interior_vertices = " << num_interior_vertices << endl;
		cout << "num_interior_edges = " << num_interior_edges << endl;
		cout << "test block size = " << num_interior_edges-num_edges_in_infinite_cycle << endl;
		cout << "num_test_block_places = " << num_test_block_places << endl;
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)		
{
	debug << "num_vertices_in_dual_graph = " << num_vertices_in_dual_graph << endl;
	debug << "num_edges_in_infinite_cycle = " << num_edges_in_infinite_cycle << endl;
	debug << "num_4_gons_in_disc = " << num_4_gons_in_disc << endl;
	debug << "num_boundary_vertices = " << num_boundary_vertices << endl;
	debug << "num_interior_vertices = " << num_interior_vertices << endl;
	debug << "num_interior_edges = " << num_interior_edges << endl;
	debug << "test block size = " << num_interior_edges-num_edges_in_infinite_cycle << endl;
	debug << "num_test_block_places = " << num_test_block_places << endl;
}
	
	/* decrement doodle_code[last_code_index] so doodle code indicates the number of i-valent vertices in the dual graph */
	doodle_code[last_code_index]--;
	int max_valency = last_code_index;
	while (doodle_code[max_valency] == 0)
		max_valency--;
		
	if (!QUIET_MODE)
	{
		cout << "max_valency of vertices on the 4-gon disc = " << max_valency << endl;	
		cout << endl;
	}
	
	matrix<int> incidence(num_vertices_in_dual_graph,num_vertices_in_dual_graph);
	for (int i=0; i< num_boundary_vertices; i++)
	{
		incidence[i][(i+1)%num_boundary_vertices] = 1;
		incidence[(i+1)%num_boundary_vertices][i] = 1;
	}


	/* Identify the endpoints of all possible boundary chord positions 
	
	
	   Consider a disc with 2n vertices in the boundary and a set of disjoint chords drawn between a subset of those vertices.  Assume that is it possible to extend the disc 
	   and chords to the dual graph of a doodle D by adding outside edges and interior vertices as necessary.  We get the same doodle, up to the usual reflective symmetry, 
	   if we act upon the disc by any element of the dihedral group D_{2n}.  That means we may fix a cyclic numbering the of the boundary vertices and consider the possible 
	   chord endpoints with respect to that numbering.
	
	   Any given set of chords divides the disc into polygonal regions whose interiors are then further subdivided (if necessary) to create the four-sided regions of a 
	   dual graph.  Suppose such a polygonal region has k boundary edges and is divided into r four-sided regions by a total of e edges and some number of interior vertices.  
	   We have 4r = 2e - k, which has a solution only if k is even.
	
	   Consider for example the case p=6 so that the disc has 12 vertices (i.e. a clock face) and consider the possible chords that we can construct from the vertex at 12 o'clock.  
	   By the above, we could construct a chord only to the odd numbered vertices.
	
	   In fact, for doodles up to 14 crossings we don't need to consider constructing a chord between adjacent vertices, so 12 can't connect to 1 or 11.  This is because to subdivide 
	   a disc with two boundary edges into four-sided regions, where all the interior vertices have valency 3, requires adding 7 interior edges.   With 14 crossings we have at least 
	   eight vertices in the boundary of the dual graph and therefore only 5 additional vertices to play with.  
	
	   So, if we ignore chords between adjacent vertices we can join vertex 12 to 3,5,7 or 9, however up to symmetry we only need consider 12 to 3 or 5.
	
	   For the 2-chord cases, by symmetry we need only consider the ways to add a second chord amongst higher numbered vertices.  Thus, writing x-y if x is joined to y we have to 
	   consider the cases:
	
	   12-3: 12-5, 12-7, or 12-9
	         11-4, 11-6, or 11-8
	         10-3, 10-5, or 10-7
	         9-4, or 9-6
	         8-3 or 8-5
	         7-4
	         6-3
	       
	   12:5: 12-7 or 12-9
	         11-6 or 11-8
	         10-5 or 10-7
	         9-6 
	         8-5
	      
	   12:7: 12-9
	         11-8
	         10-7
		 
	   One can continue with three-chords etc. in the same vein.  The nett effect is that the only boundary chords that we need to consider are:
		
			12-3 12-5 12-7 12-9
			11-4 11-6 11-8
			10-3 10-5 10-7
			9-4 9-6
			8-3 8-5
			7-4
			6-3	
			
		Similarly for any other p, we have a set of 2 * \Sum_1^{p-3} + p-1  = (p-2)^2 possible boundary chords, which have a maximum number of 2^{(p-2)^2} combinations, many of which are
		invalid due to the planarity requirement and the fact that we may have insufficent interior vertices for all the regions created by the boundary chords.
		
		If there are a maximum of 14 crossings in the doodle, there are 16 vertices in the dual graph and at least 8 vertices in the boundary of \Gamma^*, leaving a maximum of 7 interior vertices.
		That means there acnnot be more than one region with four vertices in its boundary, since any 4--boundary-vertex region requires at least 4 interior vertices to sub-divide it into four-
		sided regions.
	*/
	
	/* First, determine the doodles that have no boundary chords */
	chord_data no_boundary_edge_chord_data;
	no_boundary_edge_chord_data.min_num_interior_vertices = vector<int>(1);
	no_boundary_edge_chord_data.min_num_interior_vertices[0] = 0;
	no_boundary_edge_chord_data.region = vector<region_data>(1);
	
	region_data single_region_data;
	single_region_data.num_boundary_vertices = num_boundary_vertices;
	single_region_data.num_interior_vertices = num_interior_vertices;
	single_region_data.first_interior_vertex = num_boundary_vertices;
	single_region_data.boundary_vertices = vector<int>(num_boundary_vertices);
	for (int i=0; i< num_boundary_vertices; i++)
		single_region_data.boundary_vertices[i] = i;
	
	no_boundary_edge_chord_data.region[0] = single_region_data;
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "  no-boundary_edge_chord_data:" << endl;
	write_chord_data(no_boundary_edge_chord_data,debug);
	debug << endl;
}
	
	set_interior_choices(no_boundary_edge_chord_data);

	/* Now determine possible boundary chord patterns */
	int num_boundary_chords = (num_edges_in_infinite_cycle-2)*(num_edges_in_infinite_cycle-2);
	vector<pair<int,int> > boundary_chords(num_boundary_chords);
	int index=0;
	for (int i=num_boundary_vertices; i >= 6; i--)
	{
		for (int j=0; j< i/2 -2; j++)
		{
			if (i%2 == 1)
				boundary_chords[index++] = pair<int,int>(i,3+2*j+1);
			else
				boundary_chords[index++] = pair<int,int>(i,3+2*j);
		}
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "\nboundary_chords: ";
	
	for (size_t i=0; i< boundary_chords.size(); i++)
		debug << '(' << boundary_chords[i].first << ',' << boundary_chords[i].second << ") ";
	
	debug << endl;
}	
		
	list<chord_data> patterns;
	
	for (int i=3; i<= num_edges_in_infinite_cycle; i+=2) // i<= num_boundary_vertices/2
	{
		list<pair<int,int> > local_list;
		local_list.push_back(pair<int,int>(num_boundary_vertices,i));
		patterns.push_back(chord_data(local_list));
	}
	
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "patterns size = " << patterns.size() << endl;
	
	list<chord_data>::iterator pptr = patterns.begin();
	
	int count = 0;
	int valid_pattern_count = 0;
	
	while(pptr != patterns.end())
	{
		count++;
		
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << "  pattern length = " << pptr->chords.size() << ": ";

		/* count the minimum number of interior vertices required by this pattern of boundary chords */	
		list<pair<int,int> >::iterator next_cptr = pptr->chords.begin();
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << '(' << next_cptr->first << ',' << next_cptr->second << ") ";

		int num_chord_regions = pptr->chords.size()+1;
		vector<int> min_num_interior_vertices(num_chord_regions);
		int num_interior_vertices_required = (next_cptr->second == 3? 4:1);
		min_num_interior_vertices[0] = num_interior_vertices_required;
		int index = 1;
		
		list<pair<int,int> >::iterator last_cptr = next_cptr;
		next_cptr++;
		
		
		while (next_cptr!= pptr->chords.end())
		{

			/* if the next chord is parallel to the last, we don't necessarily need any interior vertices, though
			   we could add four to such a region if enough are available.
			*/
			if (next_cptr->first != last_cptr->first -1 || next_cptr->second != last_cptr->second +1)
			{
				num_interior_vertices_required++;
				min_num_interior_vertices[index] = 1;
			}
			index++;
			
if (debug_control::DEBUG >= debug_control::SUMMARY)
	debug << '(' << next_cptr->first << ',' << next_cptr->second << ") ";
			last_cptr++;
			next_cptr++;
		}
		
		if (last_cptr->first - last_cptr->second == 3)
		{
			min_num_interior_vertices[index] = 4;
			num_interior_vertices_required+=4;
		}
		else
		{
			min_num_interior_vertices[index] = 1;
			num_interior_vertices_required++;
		}


		bool valid_chord_pattern = (num_interior_vertices_required <= num_interior_vertices);

if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << ", requires " << num_interior_vertices_required << " interior vertices, valid_chord_pattern = " << valid_chord_pattern << endl;
		
		if (valid_chord_pattern)
		{
			valid_pattern_count++;

			/* extend the pattern list before processing the current chord pattern */
			for (int i=0; i< num_boundary_chords; i++)
			{
				if (boundary_chords[i].first <= last_cptr->first)
				{
					if (boundary_chords[i].second >= last_cptr->second && boundary_chords[i] != *last_cptr)
					{
						chord_data next_pattern = *pptr;
						next_pattern.push(boundary_chords[i]);
						patterns.push_back(next_pattern);
					}
				}
			}
			
			
			
			
		}

		pptr++;
	}

if (debug_control::DEBUG >= debug_control::SUMMARY)		
	debug << "total of " << count << " patterns" << endl;
	cout << valid_pattern_count << " valid patterns" << endl;

exit(0);

	/* The first num_boundary_vertices rows and columns of the incidence matrix correspond to the boundary vertices and we assume that vertex 0 is
	   a middle vertex and that vertex 1 is joined to the infinite vertex, so that the even numbered vertices must have an edge connecting them.
	   These even numbered vertices must be joined to an interior edge.
	   
	   There are num_interior_vertices and num_edges_in_infinite_cycle middle vertices, so there are num_interior_vertices^ num_edges_in_infinite_cycle
	   possible combinations, though we avoid effectively renumbering vertices as we enumerate them, so there are fewer combinations to consider.
	   We can only have max_valency edges at any one vertex of the interior, so  we initialize the interior choice by joining as many vertices as posible 
	   to each interior vertex.  If we run out of interior vertices, we can reject doodle_code as unrealizable.   
	*/
	vector<int> interior_choice(num_edges_in_infinite_cycle);
	vector<int> max_interior_choice(num_edges_in_infinite_cycle,num_vertices_in_dual_graph-1);
	vector<int> valency(num_vertices_in_dual_graph);
	
	bool realizable = true;
	
	
	/* Initialize interior_choice and check as we do that we can join all the 2-valent boundary vertices to interior vertices without
	   exceeding the maximum allowed valencey.  Take a note of the valency of the interior vertices based on the interior_choice.
	*/
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
	{
		int interior_vertex = i/max_valency;
		if (interior_vertex == num_interior_vertices)
		{
			cout << "Doodle code is unrealizable.  Cannot join all 2-valent boundary vertices to interior vertices without exceeding maximum allowed valency" << endl;
			realizable = false;
			break;
		}
		else
		{
			interior_choice[i] = num_boundary_vertices+interior_vertex;
			valency[num_boundary_vertices+interior_vertex]++;
			
			if (i < num_interior_vertices)
				max_interior_choice[i] = num_boundary_vertices+i;
		}
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "max_interior_choice: ";
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
		debug << max_interior_choice[i] << ' ';
	debug << endl;
}	
	if (realizable)
	{
		list<string> doodle_list;
		int solution_count = 0;
		bool finished = false;
		do
		{
			
if (TEST_MODE)
{
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
		valency[interior_choice[i]]--;
	
	cout << "Interior_choice: ";
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
		cin >> interior_choice[i];

	for (int i=0; i< num_edges_in_infinite_cycle; i++)
		valency[interior_choice[i]]++;
}			
			
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "interior_choice: ";
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
		debug << interior_choice[i] << ' ';
	debug << endl;
}
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valency: ";
	for (int i=0; i< num_vertices_in_dual_graph; i++)
		debug << valency[i] << ' ';
	debug << endl;
}
			if (INTERIOR_CHOICES)
			{
				cout << "interior_choice: ";
				for (int i=0; i< num_edges_in_infinite_cycle; i++)
					cout << setw(3) << interior_choice[i];

				cout << "   max: ";
				for (int i=0; i< num_edges_in_infinite_cycle; i++)
					cout << setw(3) << max_interior_choice[i];
				cout << endl;
			}
			else
			{
				/* determine how many i-valent vertices there are in the interior graph, given the initial interior choice
				vector<int> valency_count(max_code_index+1);
				valency_count[3] = num_boundary_vertices;
				for (int v = num_boundary_vertices; v<num_vertices_in_dual_graph; v++)
				{
					int vcount = count(interior_choice.begin(),interior_choice.end(),v);
					if (vcount>2)
						valency_count[vcount]++;
				}
				*/
				
				matrix<int> local_incidence = incidence;
				for (int i=0; i< num_edges_in_infinite_cycle; i++)
				{
					local_incidence[2*i][interior_choice[i]] = 1;
					local_incidence[interior_choice[i]][2*i] = 1;
				}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "sufficent edges available" << endl;
			
					add_interior_edges(local_incidence,num_interior_edges, num_interior_edges-num_edges_in_infinite_cycle,num_boundary_vertices, doodle_code, solution_count, doodle_list, doodle_code, interior_choice);
			}
			
if (TEST_MODE)
	break;
			


			/* set the new max_interior_choice to avoid effectively re-numbering the interior vertices.  If there are unattached interior vertices and we are going to increment 
			   interior_choice so that an unattached vertex becomes joined to the boundary, then we do not need to select each available unattached interior vertex in turn, since
			   doing so effectively renumbers the unattached interior vertices.  We would get the same graph layouts with the interior vertices renumbered.  Therefore, we can set 
			   the maximum interior choice for boundary vertex i (well, 2*i but the ith element if interior_choice) to be one more than the current value of interior_choice[i-1].
			   Note however, that if we have a current interior choice such as 8 9 8 9, with three interior vertices 8,9,10 available, then we have to allow the next interior_choice
			   8 9 8 10, so the vector max_interior_choice must always be monotonically increasing.
			*/		
			for (int i=1; i<num_edges_in_infinite_cycle; i++)
				max_interior_choice[i] = max(max_interior_choice[i-1],min(interior_choice[i-1]+1,num_vertices_in_dual_graph-1));


			/* increment interior_choice for the even numbered vertices in the disc boundary */
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "increment interior_choice" << endl;
	
			finished = true;
			for (int i = num_edges_in_infinite_cycle-1; i>0; i--) // always fix vertex zero to the first interior edge
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "  i = " << i << " interior_choice[" << i << "] = " << interior_choice[i] << endl;
				if (interior_choice[i] < max_interior_choice[i])
				{
					for (int j=i; j<num_edges_in_infinite_cycle; j++)
						valency[interior_choice[j]]--;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valency: ";
	for (int i=0; i< num_vertices_in_dual_graph; i++)
		debug << valency[i] << ' ';
	debug << endl;
}
					/* we could already have max_valency edges at interior_choice[i]+1 in amongst the vertices 0,...,i, if not, this is a valid next choice 
					   for the ith even numbered boundary vertex
					*/
					if(valency[interior_choice[i]+1] < max_valency)
					{
						
						interior_choice[i]++;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "  new interior_choice[" << i << "] = " << interior_choice[i] << endl;
	
						valency[interior_choice[i]]++;
						
						/* we know there are enough interior vertics to join all the even edges of the disc boundary to an interior verrtex */
						for (int j=0; j< num_edges_in_infinite_cycle-i-1; j++)
						{
							int interior_vertex = 0;
							while (valency[num_boundary_vertices+interior_vertex] >= max_valency)
								interior_vertex++;
							interior_choice[i+1+j] = num_boundary_vertices+interior_vertex;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "  new interior_choice[" << i+1+j << "] = " << interior_choice[i+1+j] << endl;
							valency[num_boundary_vertices+interior_vertex]++;
						}
						
						finished = false;
						break;
					}
					else
					{
						/* restore the valency counts ready for testing the preceeding even numbered boundary vertex's interior choice.*/
						for (int j=i; j<num_edges_in_infinite_cycle; j++)
							valency[interior_choice[j]]++;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "restored valency: ";
	for (int i=0; i< num_vertices_in_dual_graph; i++)
		debug << valency[i] << ' ';
	debug << endl;
}
					}
				}
			}
							
			if (finished)
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "finished, no more initial interior choices" << endl;
			}
									
		} while (!finished);
		
		cout << "found " << doodle_list.size() << " distinct solutions" << endl;
		list<string>::iterator sptr = doodle_list.begin();
		while (sptr != doodle_list.end())
		{
			cout << *sptr << endl;
			output << *sptr << endl;
			sptr++;
		}
	}
	

	time_t end_time = time(0);
	double time_exp = difftime(end_time, start_time);
	int dd = static_cast<int>(time_exp/3600/24);
	time_exp -= dd*3600*24;
	int hh = static_cast<int>(time_exp/3600);
	time_exp -= hh*3600;
	int mm = static_cast<int>(time_exp/60);
	time_exp -= mm*60;
	int ss = static_cast<int>(time_exp);
	output << "\nRunning time: " << dd << " days " << hh << " hours " << mm	<< " minutes " << ss << " seconds" << endl;

	output.close();
	

	return 0;
} /* End of main */

void helpinfo()
{
	cout << "\nUsage: 4-gon [-<short-option>] [--<long-option>] <num-crossings> <doodle-code>";
	cout << "\n<short-option>:";
	cout << "\n       i: show initial interior vertex choices only";
	cout << "\n       d: draw graphs of valid 4-gon graphs";
	cout << "\n       T: test mode";
	cout << "\n       X: development mode";
	cout << "\n       #: debug, see #H! for details";
	cout << "\n<long-option>:";
	cout << "\n       development: same as -X";
	cout << "\n       draw: same as -d";
	cout << "\n       test: same as -T";
	cout << endl;
	exit(0);
}

/***************  Functions required for setting up vlist specific debug **************/

void set_main_debug_option_parameter(char* pptr, string option);
void set_main_debug_option(char* start, char* end)
{
	char  loc_buf[end-start+2];
	char* c1 = start;
	char* c2 = loc_buf;

	/* if both start and end are zero, display debug help information.
	   this has been included here in this manner so that each time a debug option
	   is added, the help will be updated (hopefully!)
	*/
	if (start == 0 && end == 0)
	{
		set_main_debug_option_parameter(0,"f-gon");
		return;
	}

	do
	{
		*c2++ = *c1++;
	} while (c1 <= end);
	
	*c2 = '\0';

	char* pptr = strchr(loc_buf,'{');
	if (pptr)
	{
		*pptr++ = '\0';
	}

	/* now, even if there are parameters, the first part of loc_buf 
	   is a C-string that identifies the option */
	
	if (!strcmp(loc_buf,"f-gon"))
	{
		debug_control::DEBUG = debug_control::INTERMEDIATE;
		debug << "main::set_main_debug_option: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
		
		if (pptr)
		{
			/* check for any parameters */
			check_debug_option_parameters(pptr, "f-gon");
		}
	}

	debug.flush();

}

void set_main_debug_option_parameter(char* pptr, string option)
{
	if (option == "f-gon")
	{
		if (!pptr)
		{
			cout << "\t\tf-gon{summary|1:basic|2:intermediate|3:detail|4}, integer: default 0=off, no parameters sets summary" << endl;
		}
		else
		{
			if (!strcmp(pptr,"summary") || !strcmp(pptr,"1") )
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			if (!strcmp(pptr,"basic") || !strcmp(pptr,"2") )
			{
				debug_control::DEBUG = debug_control::BASIC;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::BASIC" << endl;		
			}
			else if (!strcmp(pptr,"intermediate") || !strcmp(pptr,"3"))
			{
				debug_control::DEBUG = debug_control::INTERMEDIATE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::INTERMEDIATE" << endl;		
			}
			else if (!strcmp(pptr,"detail") || !strcmp(pptr,"4"))
			{
				debug_control::DEBUG = debug_control::DETAIL;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::DETAIL" << endl;		
			}
			else if (!strcmp(pptr,"exhaustive") || !strcmp(pptr,"5"))
			{
				debug_control::DEBUG = debug_control::EXHAUSTIVE;
				debug << "main::set_main_debug_option_parameter: setting debug option debug_control::DEBUG = debug_control::EXHAUSTIVE" << endl;		
			}
		}
	}
}

void main_display_default_options()
{
	cout << "\t\tf-gon{summary}" << endl;
}

void set_programme_short_option(char* cptr)
{
	if (strchr(cptr, 'H') && strchr(cptr, '!'))
	{
		if (strchr(cptr,'#'))
		{
			debug_help();
		}
		else
		{					
			helpinfo();
		}
	}

	char* dptr = strchr(cptr, '#');
	if (dptr)
	{
		debug.open ("4-gon.dbg"); 

		if (!debug)
		{
			cerr << "\nError opening debug file\n";
			exit(0);
		}
		else
		{
			debug << "Debug information from 4-gon\n\n";
			debug << boolalpha;
		}

		if (!debug_setup(cptr))
		{
			debug_control::DEBUG = debug_control::SUMMARY;
			debug << "main: default debug options set" << endl;
		}
		
	}

	if (strchr(cptr, 'd'))
	{
		DRAW_GRAPHS = true;
	}

	if (strchr(cptr, 'i'))
	{
		INTERIOR_CHOICES = true;
	}

	if (strchr(cptr, 'Q'))
	{
		QUIET_MODE = true;
	}

	if (strchr(cptr, 'T'))
	{
		TEST_MODE = true;
	}
	
	if (strchr(cptr, 'X'))
	{
		DEVELOPMENT_MODE = true;
	}
}

void set_programme_long_option(char* cptr, string source)
{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "set_programme_long_option: passed " << cptr << " from " << source << endl;

	char  loc_buf[40];
	char* c1;
	char* c2;

	c1 = cptr;

	/* take out any leading -- */
	while (*c1 == '-')
		c1++;
	
	c2 = loc_buf;
	while (isalpha(*c1) || isdigit(*c1) || *c1 == '-' || *c1 == '_')
		*c2++ = *c1++;
    *c2 = '\0';
	
	if (!strcmp(loc_buf,"development"))
	{
		DEVELOPMENT_MODE = true;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "set_programme_long_option: DEVELOPMENT_MODE read from " << source << ", DEVELOPMENT_MODE = " << DEVELOPMENT_MODE << endl;
	}
	else if (!strcmp(loc_buf,"draw"))
	{
		DRAW_GRAPHS = true;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "set_programme_long_option: DRAW_GRAPHS read from " << source << ", DRAW_GRAPHS = " << DRAW_GRAPHS << endl;
	}
	else if (!strcmp(loc_buf,"test"))
	{
		TEST_MODE = true;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "set_programme_long_option: TEST_MODE read from " << source << ", TEST_MODE = " << TEST_MODE << endl;
	}
}

void print_prog_params (ostream& os, string level, string prefix)
{
    os << prefix << "DEVELOPMENT_MODE = " << DEVELOPMENT_MODE << endl;
    os << prefix << "TEST_MODE = " << TEST_MODE << endl;
}

