/**************************************************************************
4-gon functions

This module contains:

void add_interior_edges(matrix<int> incidence, int num_interior_edges, int num_edges_to_add, int num_boundary_vertices, vector<int>& max_valency_count, 
                        int& solution_count, list<string>& doodle_list, vector<int>& doodle_code, vector<int>& interior_choice)
int min_num_edges(int n2, int n1, int n0, int n2j2, int n2j1, int n1j1, int n1j2j1, int n1j2j2, int n1j2j2j1)
bool planar_graph(matrix<int>& incidence,int num_boundary_vertices)
bool valid_4_gon_graph(matrix<int>& incidence, int num_boundary_vertices, int num_crossings, int solution_count, string& gauss_code) 
float edge_angle(float v_x, float v_y, float e_x, float e_y)
string doodle_gauss_code(vector<pair<int,int> >& edge_pair,matrix<int>& cycle_vertices, int num_edges, int num_boundary_edges)
int succeeding_edge(matrix<int>& region_edges, int num_4_gons_in_disc, int edge, int& region)
void test_ogdf(matrix<int>& incidence,int num_boundary_vertices, int diagramnum)

**************************************************************************/
using namespace std;
#include <fstream>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>
#include <algorithm>
#include <csignal>
#include <iomanip>
#include <ctype.h>

#include <pthread.h>

/******************** Global Variables **********************/
extern ofstream    debug;
extern ofstream    output;
extern bool	DRAW_GRAPHS;
extern bool	TEST_MODE;
extern bool	sigusr1_request;

#include <4-gon.h>

#include <ogdf/fileformats/GraphIO.h>
#include <ogdf/orthogonal/OrthoLayout.h>
#include <ogdf/planarity/EmbedderMinDepthMaxFaceLayers.h>
#include <ogdf/planarity/PlanarSubgraphFast.h>
#include <ogdf/planarity/PlanarizationLayout.h>
#include <ogdf/planarity/SubgraphPlanarizer.h>
#include <ogdf/planarity/VariableEmbeddingInserter.h>

namespace util
{
	string itos(long n);
}

vector<int> debug_start={10, 17, 25, 28, 41, 42, 43, 44};
vector<int> debug_end={10, 17, 27, 38, 41, 42, 43, 44};
 
using namespace ogdf;

bool comp_petal (petal& p1, petal& p2) 
{
	return (p1.angle < p2.angle);
}
	
bool planar_graph(matrix<int>& incidence,int num_boundary_vertices);
bool valid_4_gon_graph(matrix<int>& incidence, int num_boundary_vertices, int num_crossings, int solution_count, string& gauss_code);
float edge_angle(float v_x, float v_y, float e_x, float e_y);
string doodle_gauss_code(vector<pair<int,int> >& edge_pair,matrix<int>& cycle_vertices, int num_edges, int num_boundary_edges);
int succeeding_edge(matrix<int>& region_edges, int num_4_gons_in_disc, int start, int& region);


/* for i >=3 max_valency_count[i] indicates the maximum number of i-valent vertices permitted in the interior graph */
void add_interior_edges(matrix<int> incidence, int num_interior_edges, int num_edges_to_add, int num_boundary_vertices, vector<int>& max_valency_count, 
                        int& solution_count, list<string>& doodle_list, vector<int>& doodle_code, vector<int>& interior_choice)
{
	bool no_block_increment = false; // used by TEST_MODE
	int num_vertices = incidence.numrows();
	int num_interior_vertices = num_vertices -num_boundary_vertices;

	int max_valency = static_cast<int>(max_valency_count.size())-1;
	while (max_valency_count[max_valency] == 0)
		max_valency--;
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	print (incidence,debug, 2,"add_interior_edges: ");
	debug << "add_interior_edges: num_interior_edges = " << num_interior_edges << ", num_edges_to_add = " << num_edges_to_add << ", num_boundary_vertices = " << num_boundary_vertices << endl;
	debug << "add_interior_edges: max_valency = " << max_valency << ", max valency_count: ";
	for (size_t i=3; i <max_valency_count.size(); i++)
		debug << max_valency_count[i] << ' ';
	debug << endl;
}
	vector<int> valency(num_vertices);
	for (int i=0; i< num_vertices; i++)
	{
		for (int j=0; j< num_vertices; j++)
			valency[i] += incidence[i][j];
	}
	
	for (int i=0; i<num_boundary_vertices/2; i++)
		valency[2*i+1]++; // edge to the infinite vertex.

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: current vertex valency: ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
}	

	vector<int> valency_count(max_valency+1); // we want to use the valency as an index, so +1
	for (int i=0; i< num_vertices; i++)
		valency_count[valency[i]]++;
		
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: initial valency_count: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}	
	
	/* We have to place num_edges_to_add edges in incidence to the intersection of the right side of the leading diagonal and the right side of the
	   top left num_boundary_vertices x num_boundary_vertices square.  Our initial_choice has taken num_boundary_vertices/2 of these spaces, so start
	   by calculating the number of places left and establish a map from the integers 0,...,num_places-1 to the location of the free places.
	*/
	int num_places = (num_boundary_vertices * num_interior_vertices - num_boundary_vertices/2) + (num_interior_vertices * (num_interior_vertices -1) / 2);
	vector<pair<int,int> > place_map(num_places);
	int index = 0;
	for (int i=0; i<num_boundary_vertices; i++)
	{
		for (int j=num_boundary_vertices; j< num_vertices; j++)
		{
			if (incidence[i][j] == 0)
			{
				place_map[index] = {i,j};
				index++;
			}
		}
	}
	
	for (int i=num_boundary_vertices; i< num_vertices; i++)
	{
		for (int j=i+1; j< num_vertices; j++)
		{
			if (incidence[i][j] == 0)
			{
				place_map[index] = {i,j};
				index++;
			}
		}
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: num_places to fill = " << num_places << endl;
	debug << "add_interior_edges: place_map:" << endl;
	for (int i=0; i< num_places; i++)
	debug << "add_interior_edges:   " << setw(2) << i << ": " << place_map[i].first << ' ' << place_map[i].second << endl;
}

	/* work through all options for filling the available places */
	vector<int> block(num_edges_to_add);
	vector<int> max(num_edges_to_add);
	for (int i=0; i< num_edges_to_add; i++)
	{
		block[i] = i;
		max[i] = num_places-num_edges_to_add+i;
	}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: block max: ";
	for (int i=0;i< num_edges_to_add; i++)
		debug << max[i] << ' ';
	debug << endl;
}

if (TEST_MODE)
{
	cout << "test block? y/n/r ";
	char ch;
	cin >> ch;

	if (ch == 'y')
		no_block_increment = true;
	
	if (ch == 'y' || ch == 'r')
	{
		cout << "enter test block: ";
		for (int i=0; i< num_edges_to_add; i++)
			cin >> block[i];
	}
}	

	bool finished = false;
	do // test block
	{		
		if (sigusr1_request)
		{						
			vector<int> original_doodle_code = doodle_code;
			original_doodle_code[num_boundary_vertices/2]++;
			
			clog << "doodle code: ";
			for (size_t i=0; i < original_doodle_code.size(); i++)
				clog << original_doodle_code[i] << ' ';
			clog << endl;
			clog << "interior_choice: ";
			for (size_t i=0; i< interior_choice.size(); i++)
				clog << setw(3) << interior_choice[i];
			clog << endl;
			clog << "num_boundary_vertices = " << num_boundary_vertices << endl;
			clog << "num_interior_vertices = " << num_interior_vertices << endl;
			clog << "num_interior_edges = " << num_interior_edges << endl;
			clog << "block size = " << num_edges_to_add << endl;
			clog << "num_places for test block = " << num_places << endl;
			clog << "test block: ";
			for (int i=0;i< num_edges_to_add; i++)
				clog << block[i] << ' ';
			clog << endl;
			
			sigusr1_request = false;
		}
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: test block ";
	for (int i=0;i< num_edges_to_add; i++)
		debug << block[i] << ' ';
	debug << endl;
}
		
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: current vertex valency : ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges: current valency_count: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}

		bool skip_block = false;
		
		for (int i=0; i< num_edges_to_add; i++)
		{	
			int vr=place_map[block[i]].first;  // vertex row, this will be the row in incidence in which we set the next 1
			int vc=place_map[block[i]].second;  // vertex column, the column in incidence where we set the next 1
			
			/* The regions of the dual graph have to be four-sided, so if u is joined by an edge to v and v is joinred to w, 
			   then u cannot be joined to w.  Here we are proposing to vertex vr to vc, so first check that rows vr and vc
			   do not already have a common set column entry, indicting that vertex vr and vc are both already joined to a
			   common vertex.
			*/
			bool common_vertex = false;
			for (int j=0; j< num_vertices; j++)
			{
				if (incidence[vr][j] == 1 && incidence[vc][j] == 1)
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: vertices " << vr << " and " << vc << " (block[" << i << "]) are joined to common vertex = " << j << endl;
					common_vertex = true;
					break;
				}
			}

			bool valency_limit_reached = false;
			if (valency[vr] == max_valency || valency[vc] == max_valency)
			{				
				valency_limit_reached = true;
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: cannot add another edge to vertex " << (valency[vr] == max_valency? vr:vc) << " (block[" << i << "]) without exceeding max_valency = " << max_valency << endl;
			}
			
			if (common_vertex || valency_limit_reached)
			{				

				/* Un-set any edges we've already set in incidence and adjust valency and 
				   valency _count accordingly, then advance the block if we can. 
				*/
				for (int j=0; j< i; j++)
				{
					int vvr = place_map[block[j]].first;
					int vvc = place_map[block[j]].second;
					valency_count[valency[vvr]]--;
					valency_count[valency[vvc]]--;
					incidence[vvr][vvc] = 0;
					incidence[vvc][vvr] = 0;
					valency[vvr]--;
					valency[vvc]--;
					valency_count[valency[vvr]]++;
					valency_count[valency[vvc]]++;
				}
					
				finished = true;
				for (int j=i; j>=0; j--)
				{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "add_interior_edges: j = " << j << "block[" << j << "] =  " << block[j] << " max[" << j << "] =  " << max[j] << endl;
					if (block[j] < max[j])
					{
						finished = false;
						block[j]++;
						for (int k=j+1; k<num_edges_to_add; k++)
							block[k] = block[k-1]+1;
						break;
					}
				}

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{			
	if (finished)
	{
		debug << "add_interior_edges: cannot advance block, finished" << endl;
	}
	else
	{
		debug << "add_interior_edges: advanced block to ";
		for (int j=0;j< num_edges_to_add; j++)
			debug << block[j] << ' ';
		debug << endl;
		
		debug << "add_interior_edges: vertex valency reset to : ";
		for (int i=0; i< num_vertices; i++)
			debug << valency[i] << ' ';
		debug << endl;
		debug << "add_interior_edges: valency_count reset to: ";
		for (int i=0; i<= max_valency; i++)
			debug << valency_count[i] << ' ';
		debug << endl;
	}
}				
				skip_block = true;
				break;
			}
			else
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: set incidence [" << vr << "][" << vc << "] = 1, together with its transpose" << endl;

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "add_interior_edges:   valency[" << vr << "] = " << valency[vr] << ", valency[" << vc << "] = " << valency[vc] << endl;
	debug << "add_interior_edges:   valency_count[" << valency[vr] << "] = " << valency_count[valency[vr]] << ", valency_count[" << valency[vc] << "] = " << valency_count[valency[vc]] << endl;
}
				valency_count[valency[vr]]--;
				valency_count[valency[vc]]--;
				incidence[vr][vc] = 1;
				incidence[vc][vr] = 1;
				valency[vr]++;
				valency[vc]++;
				valency_count[valency[vr]]++;
				valency_count[valency[vc]]++;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges:   vertex valency stands at: ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges:   valency_count stands at: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}
	
if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "add_interior_edges:   new valency[" << vr << "] = " << valency[vr] << ", new valency[" << vc << "] = " << valency[vc] << endl;
	debug << "add_interior_edges:   new valency_count[" << valency[vr] << "] = " << valency_count[valency[vr]] << ", new valency_count[" << valency[vc] << "] = " << valency_count[valency[vc]] << endl;
}

			}
		}

		if (skip_block)
		{
			continue; // already reset incidence and incremented block
		}
		else
		{
			/* test incidence and valency */
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "add_interior_edges: test vertex valency: ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges: test valency_count: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}
			bool valid = true;

			/* Reject any block that leaves vertices with valency less than 3.  Advance block past these cases where possible.  All
			   vertices up to num_boundary_vertices have valency at least 3, so only need check after that.
			*/
	
			for (int i=num_vertices-1; i >= num_boundary_vertices; i--)
			{
				if (valency[i] < 3)
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: vertex " << i << " has valency less than 3, reject block" << endl;					

					valid = false;
					
					/* look for the greatest block element on or before row i, this is the edge we're going to advance */
					int block_index = -1;
					for (int j=num_edges_to_add-1; j>=0 ; j--)
					{
						if (place_map[block[j]].first <= i)
						{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   block index " << j << ", element " << block[j] << " is the largest index contributing an edge before row " << i << endl;					
	
							block_index = j;
							break;
						}
					}
					
					if (block_index == -1)
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   no block index contributing an edge before row " << i << ", finished" << endl;											
						finished = true;
						break;
					}
					else
					{					
						/* advance the block starting at block_index, re-setting incidence and valency and valency_count */
	
						for (int j=0; j< num_edges_to_add; j++)
						{
							int vvr = place_map[block[j]].first;
							int vvc = place_map[block[j]].second;
							valency_count[valency[vvr]]--;
							valency_count[valency[vvc]]--;
							incidence[vvr][vvc] = 0;
							incidence[vvc][vvr] = 0;
							valency[vvr]--;
							valency[vvc]--;
							valency_count[valency[vvr]]++;
							valency_count[valency[vvc]]++;
						}
	
						/* adjust block */
						finished = true;
						for (int i=block_index; i>=0; i--)
						{
							if (block[i] < max[i])
							{
								finished = false;
								block[i]++;
								for (int j=i+1; j<num_edges_to_add; j++)
									block[j] = block[j-1]+1;
								break;
							}
						}
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{			
	debug << "add_interior_edges: vertex valency reset to : ";
	for (int i=0; i< num_vertices; i++)
		debug << valency[i] << ' ';
	debug << endl;
	debug << "add_interior_edges: valency_count reset to: ";
	for (int i=0; i<= max_valency; i++)
		debug << valency_count[i] << ' ';
	debug << endl;
}				
						break;					
					}
				}
			}

			if (!valid)
				continue;


			for (int i=3; i<= max_valency; i++)
			{
				if (valency_count[i] != max_valency_count[i])
				{
					valid = false;
					break;
				}
			}
			
			if (valid)
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: valency_count correct, block may define a 4-gon graph" << endl;					
		
				if (planar_graph(incidence,num_boundary_vertices))
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: interior graph is planar" << endl;					

					string gauss_code;
					if (valid_4_gon_graph(incidence, num_boundary_vertices, num_vertices-1,solution_count,gauss_code))
					{
						solution_count++;			
						
						/* check whether this soltion is distinct from the others on doodle_list, we check twice, once with 
						   the gauss_code provided and once mirror image
						*/
						bool distinct = true;
						string test_string = gauss_code;
						
						for (int i=0; i < 2; i++)
						{
							if (i == 1)
							{
								for (size_t j=0; j< test_string.length(); j++)
								{
									if (test_string[j] == 'L')
										test_string[j] = 'R';
									else if (test_string[j] == 'R')
										test_string[j] = 'L';
								}
								
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   mirror image of gauss code " << test_string << endl;

								ofstream tmp("tmpfile");
								tmp << "[raw-output,silent,gauss,ulpgd]" << endl;
								tmp << test_string << endl;
								
								system("braid tmpfile");
								ifstream braid("braid.out");
								
								if (!getline(braid, test_string))
								{
									cerr << "\nError! Braid programe failed to calculate left_preferred Gauss data." << endl;
									exit(0);
								}
								braid.close();
								
								test_string += gauss_code.substr(gauss_code.find('/'));
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   left preferred mirror image of gauss code " << test_string << endl;
								
							}
							
							list<string>::iterator sptr = doodle_list.begin();
							while (sptr != doodle_list.end())
							{
								if (*sptr == test_string)
								{
									distinct = false;
									break;
								}
								
								sptr++;
							}
							if (!distinct)
								break;
						}
						
						if (distinct)
						{
							doodle_list.push_back(gauss_code);													
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   gauss code is distinct" << endl;
						}
						else
						{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges:   gauss code is not distinct" << endl;
						}
					}
					else
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: interior graph is not a 4-gon graph" << endl;					
					}
				}
				else
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: interior graph is non-planar" << endl;					
				}										
			}
			else
			{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "add_interior_edges: incorrect valency_count, block does not define a 4-gon graph" << endl;					
			}
		}
	
		if (no_block_increment)
			break;
			
		/* re-set incidence, valency and valency_count for next test */
		for (int i=0; i< num_edges_to_add; i++)
		{
			int vvr = place_map[block[i]].first;
			int vvc = place_map[block[i]].second;
			valency_count[valency[vvr]]--;
			valency_count[valency[vvc]]--;
			incidence[vvr][vvc] = 0;
			incidence[vvc][vvr] = 0;
			valency[vvr]--;
			valency[vvc]--;
			valency_count[valency[vvr]]++;
			valency_count[valency[vvc]]++;
		}
		
		// increment block 
		finished = true;
		for (int i=num_edges_to_add-1; i>=0; i--)
		{
			if (block[i] < max[i])
			{
				finished = false;
				block[i]++;
				for (int j=i+1; j<num_edges_to_add; j++)
					block[j] = block[j-1]+1;
				break;
			}
		}
		
	} while (!finished);	
}


/* min_num_edges calculates the minimum number of edges required to be added to a set of 2-, 1- and 0-valent vertices so that all the vertices are 3-valent.
   We may have some pairs of vertices in the set already joined by an edge and we may have vertices joined by edges to vertices not in the set. We require
   that no pair of vertices are joined by more than one edge.

   There are nine possible configurations to be considered, shown below with the variable name used to indicate the number of each configuration present in
   the set to be considered.  An 'o' indicates a vertex of the set to be considered, an unterminated edge indicates an edge connecting to a vertex not in the set.
   
   
          n0               n1                      n2                  n2j2                   n2j1                  n1j1                 n1j2j1
          
                                                                                                                                         o     o
                            |                     \ /                \        /             \                                             \   /
                            |                      v                  \      /               \                                             \ /
          o                 o                      o                   o----o                 o----o               o----o                   o
   
   0-valent vertex    1-valent vertex        2-valent vertex      a pair of n1             an n1 and an n0       a pair of n0           three n0 vertices
                      not joined to a        not joined to a      vertices in the set      vetex in the set      vertices in the set    joined by two edges
                      member of the set      member of the set    joined by an edge        joined by an edge     joined by an edge
  
  
        n1j2j2                   n1j2j2j1
         
                o             o          o
      \        /               \        /
       \      /                 \      /
        o----o                   o----o
   
   
   When we add an edge between a pair of configurations we may discard any vertices of the set that become 3-valent but must retain any vertices that do not.  The configuratios
   therefore combine as shown below.  Note that some configuration combinations could be joined by an edge in a varity of ways, we are only interested here in the nuber of edges 
   required, not the manner in which they are joined.
   
   n2   --  n2     =  empty (both vertices become 3-valent and may be discarded)
   n2   --  n1     =  n2
   n2   --  n0     =  n1
   n2   --  n2j2   =  n2
   n2   --  n2j1   =  n1
   n2   --  n1j1   =  n2j1
   n2  --  n1j2j1  =  n1 + n1
   n2  --  n1j2j2  =  n2j1                  or n2+n1, 3 edges not 2
   n2 -- n1j2j2j1  =  n1 + n2j1
   
   n1   --  n1     =  n2j2
   n1   --  n0     =  n2j1
   n1   --  n2j2   =  n2 + n2
   n1   --  n2j1   =  n2 + n1
   n1   --  n1j1   =  n1j2j2
   n1  --  n1j2j1  =  n2 + n1 +n1
   n1  --  n1j2j2  =  n2 + n2j1            or  n2 + n1 + n2, 5 edges not 4
   n1 -- n1j2j2j1  =  n2j1 + n1 +n2

   n0   --  n0     =  n1j1
   n0   --  n2j2   =  n2 + n1
   n0   --  n2j1   =  n1 + n1
   n0   --  n1j1   =  n1j2j1
   n0  --  n1j2j1  =  n1 + n1 + n1
   n0  --  n1j2j2  =  n1 + n2j1           or n2 + n1 + n1, 4 edges not 3
   n0 -- n1j2j2j1  =  n2j1 + n1 + n1

   n2j2   --  n2j2   =  n2 + n2
   n2j2   --  n2j1   =  n2 + n1
   n2j2   --  n1j1   =  n2 + n2j1
   n2j2  --  n1j2j1  =  n2 + n1 + n1
   n2j2  --  n1j2j2  =  n2 + n2j1         or n2 + 1 + n2, 5 edges not 4
   n2j2 -- n1j2j2j1  =  n2j1 + n1 + n2
   
   n2j1   --  n2j1   =  n1 + n1
   n2j1   --  n1j1   =  n1 + n2j1
   n2j1  --  n1j2j1  =  n1 + n1 + n1
   n2j1  --  n1j2j2  =  n1 + n2j1         or  n2 + n1 + n1, 4 edges not 3
   n2j1 -- n1j2j2j1  =  n2j1 + n1 +n1

   n1j1   --  n1j1   =  n1j2j2j1
   n1j1  --  n1j2j1  =  n2j1 + n1 + n1
   n1j1  --  n1j2j2  =  n2j1 + n2j1        or  n2 + n1 + n2j1, 5 edges not 4
   n1j1 -- n1j2j2j1  =  n2j1 + n1 + n2j1

   n1j2j1  --  n1j2j1  = n1 + n1 + n1 +n1
   n1j2j1  --  n1j2j2  = n1 + n1 + n2j1     or n2 + n1 + n1  n1, 5 edges not 4
   n1j2j1 -- n1j2j2j1  = n2j1 + n1 + n1 +n1

   n1j2j2  --  n1j2j2  =  n2 + n1 + n2j1    or n2 + n1 + n1 + n2, 6 edges not 5
   n1j2j2 -- n1j2j2j1  =  n2j1 + n1 + n2j1  or n2j1 + n1 + n1 + n2, 6 edges not 5

   n1j2j2j1 -- n1j2j2j1  =  n2j1 + n1 + n1 + n2j1

	We consider the varous combinations in an order that simplifies the scenario as rapidly as possible.  It is likely that some conditions (e.g. n1j2j2j1 > 1) are never seen, given this approach.
	However, no attempt has been made to optimize the number of conditional steps considered in the function.
	
 */
int min_num_edges(int n2, int n1, int n0, int n2j2, int n2j1, int n1j1, int n1j2j1, int n1j2j2, int n1j2j2j1)
{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: n2 = " << n2 << " n1 = " << n1  << " n0 = " << n0 << " n2j2 = " << n2j2 << " n2j1 = " << n2j1 << " n1j1 = " << n1j1 << " n1j2j1 = " << n1j2j1 << " n1j2j2 = " << n1j2j2 << " n1j2j2j1 = " << n1j2j2j1 << endl;
	
	if (n1j2j2j1 > 1) // 8 vertices reduced to 6
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j2j2j1 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2,n1+2,n0,n2j2,n2j1+2,n1j1,n1j2j1,n1j2j2,n1j2j2j1-2);
	}
	else if (n1j2j2 && n1j2j2j1) // 7 vertices reduced to 5
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j2j2 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2,n1+1,n0,n2j2,n2j1+2,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1-1);
	}
	else if (n1j2j1 && n1j2j2j1)  // 7 vertices reduced to 5
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j2j1 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2,n1+3,n0,n2j2,n2j1+1,n1j1,n1j2j1-1,n1j2j2,n1j2j2j1-1);
	}
	else if (n2j1 && n1j2j2j1) // 6 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j1 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2,n1+2,n0,n2j2,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1-1);
	}
	else if (n2j2 && n1j2j2j1) // 6 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j2 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2+1,n1+1,n0,n2j2-1,n2j1+1,n1j1,n1j2j1,n1j2j2,n1j2j2j1-1);
	}
	else if (n1j2j2 > 1) // 6 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j1j2 and n1j2j2" << endl;
		return 1+min_num_edges(n2+1,n1+1,n0,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2-2,n1j2j2j1);
	}
	else if (n1j2j1 && n1j2j2) // 6 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j2j1 and n1j2j2" << endl;
		return 1+min_num_edges(n2,n1+2,n0,n2j2,n2j1+1,n1j1,n1j2j1-1,n1j2j2-1,n1j2j2j1);
	}
	else if (n1j1 && n1j2j2j1) // 6 vertices reduced to 5
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j1 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2,n1+1,n0,n2j2,n2j1+2,n1j1-1,n1j2j1,n1j2j2,n1j2j2j1-1);
	}
	else if (n2 && n1j2j2j1) // 5 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2-1,n1+1,n0,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2,n1j2j2j1-1);
	}
	else if (n2j1 && n1j2j2) // 5 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j1 and n1j2j2" << endl;
		return 1+min_num_edges(n2,n1+1,n0,n2j2,n2j1,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1);
	}
	else if (n2j2 && n1j2j2) // 5 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j2 and n1j2j2" << endl;
		return 1+min_num_edges(n2+1,n1,n0,n2j2-1,n2j1+1,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1);
	}
	else if (n2j2 && n1j2j1) // 5 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j2 and n1j2j1" << endl;
		return 1+min_num_edges(n2+1,n1+2,n0,n2j2-1,n2j1,n1j1,n1j2j1-1,n1j2j2,n1j2j2j1);
	}
	else if (n2j1 && n1j2j1) // 5 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j1 and n1j2j1" << endl;
		return 1+min_num_edges(n2,n1+3,n0,n2j2,n2j1-1,n1j1,n1j2j1-1,n1j2j2,n1j2j2j1);
	}
	else if (n1 && n1j2j2j1) // 5 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n1j2j2j1" << endl;
//		return 1+min_num_edges(n2+1,n1,n0,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2,n1j2j2j1-1);
		return 4+min_num_edges(n2,n1-1,n0,n2j2,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1-1);
	}
	else if (n0 && n1j2j2j1) // 5 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n1j2j2j1" << endl;
		return 1+min_num_edges(n2,n1+2,n0-1,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2,n1j2j2j1-1);
	}
	else if (n1j1 && n1j2j2) // 5 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j1 and n1j2j2" << endl;
		return 1+min_num_edges(n2,n1,n0,n2j2,n2j1+2,n1j1-1,n1j2j1,n1j2j2-1,n1j2j2j1);
	}
	else if (n1j1 && n1j2j1)  // 5 vertices reduced to 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j1 and n1j2j1" << endl;
		return 1+min_num_edges(n2,n1+2,n0,n2j2,n2j1+1,n1j1-1,n1j2j1-1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n1j2j1) // 4 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n1j2j1" << endl;
		return 1+min_num_edges(n2-1,n1+2,n0,n2j2,n2j1,n1j1,n1j2j1-1,n1j2j2,n1j2j2j1);
	}
	else if (n2j2 > 1) // 4 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j2 and n2j2" << endl;
		return 1+min_num_edges(n2+2,n1,n0,n2j2-2,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2j2 && n2j1) // 4 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j2 and n2j1" << endl;
		return 1+min_num_edges(n2+1,n1+1,n0,n2j2-1,n2j1-1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2j1 > 1) // 4 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j1 and n2j1" << endl;
//		return 1+min_num_edges(n2,n1+2,n0,n2j2,n2j1-2,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
		return 3+min_num_edges(n2,n1,n0,n2j2,n2j1-2,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n1j2j2) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n1j2j2" << endl;
//		return 1+min_num_edges(n2-1,n1,n0,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1);
		return 3+min_num_edges(n2-1,n1,n0,n2j2,n2j1,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1);
	}
	else if (n1 && n1j2j2) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n1j2j2" << endl;
		return 1+min_num_edges(n2+1,n1-1,n0,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1);
	}
	else if (n0 && n1j2j2) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n1j2j2" << endl;
		return 1+min_num_edges(n2,n1+1,n0-1,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2-1,n1j2j2j1);
	}
	else if (n1 && n1j2j1) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n1j2j1" << endl;
		return 1+min_num_edges(n2+1,n1+1,n0,n2j2,n2j1,n1j1,n1j2j1-1,n1j2j2,n1j2j2j1);
	}
	else if (n0 && n1j2j1) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n1j2j1" << endl;
		return 1+min_num_edges(n2,n1+3,n0-1,n2j2,n2j1,n1j1,n1j2j1-1,n1j2j2,n1j2j2j1);
	}
	else if (n2j2 && n1j1) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j2 and n1j1" << endl;
		return 1+min_num_edges(n2+1,n1,n0,n2j2-1,n2j1+1,n1j1-1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2j1 && n1j1) // 4 vertices reduced to 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2j1 and n1j1" << endl;
		return 1+min_num_edges(n2,n1+1,n0,n2j2,n2j1,n1j1-1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n2j2) // 3 vertices reduced to 1
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n2j2" << endl;
		return 1+min_num_edges(n2,n1,n0,n2j2-1,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n2j1) // 3 vertices reduced to 1
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n2j1" << endl;
		return 1+min_num_edges(n2-1,n1+1,n0,n2j2,n2j1-1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n1 && n2j2) // 3 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n2j2" << endl;
		return 1+min_num_edges(n2+2,n1-1,n0,n2j2-1,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n0 && n2j2) // 3 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n2j2" << endl;
		return 1+min_num_edges(n2+1,n1+1,n0-1,n2j2-1,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n1 && n2j1) // 3 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n2j1" << endl;
		return 1+min_num_edges(n2+1,n1,n0,n2j2,n2j1-1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n0 && n2j1) // 3 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n2j1" << endl;
		return 1+min_num_edges(n2,n1+2,n0-1,n2j2,n2j1-1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n1j1) // 3 vertices reduced to 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n1j1" << endl;
		return 1+min_num_edges(n2-1,n1,n0,n2j2,n2j1+1,n1j1-1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n0) // 2 vertices reduced to 1
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n0" << endl;
		return 1+min_num_edges(n2-1,n1+1,n0-1,n2j2,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2 && n1) // 2 vertices reduced to 1
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n1" << endl;
		return 1+min_num_edges(n2,n1-1,n0,n2j2,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n1j2j1 > 1) // 4 vertices remains 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j2j1 and n1j2j1" << endl;
		return 1+min_num_edges(n2,n1+4,n0,n2j2,n2j1,n1j1,n1j2j1-2,n1j2j2,n1j2j2j1);
	}
	else if (n1j1 > 1) // 4 vertices remains 4
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1j1 and n1j1" << endl;
		return 1+min_num_edges(n2,n1,n0,n2j2,n2j1,n1j1-2,n1j2j1,n1j2j2,n1j2j2j1+1);
	}
	else if (n1 && n1j1) // 3 vertices remains 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n1j1" << endl;
		return 1+min_num_edges(n2,n1-1,n0,n2j2,n2j1,n1j1-1,n1j2j1,n1j2j2+1,n1j2j2j1);
	}
	else if (n0 && n1j1) // 3 vertices remains 3
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n1j1" << endl;
		return 1+min_num_edges(n2,n1,n0-1,n2j2,n2j1,n1j1-1,n1j2j1+1,n1j2j2,n1j2j2j1);
	}
	else if (n1 && n0) // 2 vertices remains 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n0" << endl;
		return 1+min_num_edges(n2,n1-1,n0-1,n2j2,n2j1+1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n1 > 1) // 2 vertices remains 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n1 and n1" << endl;
		return 1+min_num_edges(n2,n1-2,n0,n2j2+1,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n0 > 1) // 2 vertices remains 2
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n0 and n0" << endl;
		return 1+min_num_edges(n2,n1,n0-2,n2j2,n2j1,n1j1+1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2>1) // 2 vertices reduced to 0
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: join n2 and n2" << endl;
		return 1+min_num_edges(n2-2,n1,n0,n2j2,n2j1,n1j1,n1j2j1,n1j2j2,n1j2j2j1);
	}
	else if (n2)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add edge to n2" << endl;
		return 1;
	}
	else if (n1)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 2 edges to n1" << endl;
		return 2;
	}
	else if (n0)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 3 edges to n0" << endl;
		return 3;
	}
	else if (n2j2)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 2 edges to n2j2" << endl;
		return 2;
	}
	else if (n2j1)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 3 edges to n2j1" << endl;
		return 3;
	}
	else if (n1j1)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 4 edges to n1j1" << endl;
		return 4;
	}
	else if (n1j2j1)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 4 edges to n1j2j1" << endl;
		return 4;
	}
	else if (n1j2j2)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 4 edges to n1j2j2" << endl;
		return 3;
	}
	else if (n1j2j2j1)
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: add 6 edges to n1j2j2j1" << endl;
		return 3;
	}
	else
	{
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "min_num_edges: return 0" << endl;
		return 0;
	}
}

/* planar_graph uses the OGDF library function isPLanar to decide whether the the dual graph determined by incidence is planar.  The function 
   produces a description in DOT format of the graph from incidence, first describing the disc bundary, then the interior edges and finally the 
   infinite edges connecting the disc boundary to the vertex at infinity.  
*/
bool planar_graph(matrix<int>& incidence,int num_boundary_vertices)
{
	int num_vertices = incidence.numrows();

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	print (incidence,debug, 2,"planar_graph: ");
	
	Graph G;

	ostringstream incidence_DOT;
	incidence_DOT << "graph G {0";
	for (int i=0; i< num_boundary_vertices; i++)
		incidence_DOT << "--" << (i+1)%num_boundary_vertices;
	
	for (int i=0; i<num_boundary_vertices; i++)
	{
		for (int j=num_boundary_vertices; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
			{
				incidence_DOT << ' ' << i << "--" << j;
			}
		}
	}
	
	for (int i=num_boundary_vertices; i< num_vertices; i++)
	{
		for (int j=i+1; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
			{
				incidence_DOT << ' ' << i << "--" << j;
			}
		}
	}		
	
	for (int i=0; i<num_boundary_vertices/2; i++)
		incidence_DOT << ' ' << 2*i+1 << "--" << num_vertices;
	
	incidence_DOT << '}';

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: incidence_DOT = " << incidence_DOT.str() << endl;

	istringstream input(incidence_DOT.str());
	
	if (GraphIO::readDOT(G,input))
	{
// 		bool planar = ogdf::isPlanar(G);
 		bool planar = ogdf::planarEmbed(G);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: incidence " << (planar? "verified to " : "does not ") << " define a planar graph" << endl;
	
	
	/* 4-5-23 test of ogdf adjacency list */
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	if (planar)
	{
		for (node n: G.nodes)
		{
			debug << "node index = " << n->index() << " adjEntries: ";
			
			for(adjEntry a: n->adjEntries)
				debug << a->twinNode()->index() << ' ';
			
			debug << endl;
		}				
/*
		int ogdf_num_nodes = G.numberOfNodes();
		
		Graph::node_iterator nptr = G.firstNode();
		
		for (int i=0; i< ogdf_num_nodes; i++)
		{
			debug << "node " << i << " index = " << nptr->index();
			nptr++;
		}
*/		
	}
}	
	
	
 
		return planar;	
	}
	else
	{
		cerr << "planar_graph error! Failed to read graph from DOT string" << endl;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "planar_graph: Error! Failed to read graph from incidence_DOT"<< endl;
		exit(0);
	}
}


/* valid_4_gon_graph checks whether the interior graph determined by incidence has 4-gon regions.  It produces a description in DOT format of the graph
   determined by incidence, first describing the disc bundary, then the interior edges and finally the infinite edges connecting the disc boundary to the 
   vertex at infinity.  
   
   It seems that the OGDF tools in version 2022.02 do not provide the combinatoric information about the planar layout ina simple fashion, so we have to 
   recreate it from a planar layout.   We know that the graph is planar if this function is invoked, so the OGDF orthoganal PlanarizationLayout layout is 
   called to produce a DOT description of the graph (__planar_graph.dot) in the form:
   
	graph G {
	        0 [width="10.000000", height="10.000000", shape="rect", pos="45.000000,115.000000", type="0"]
	        1 [width="10.000000", height="10.000000", shape="rect", pos="105.000000,25.000000", type="0"]
	        ...
	
	
	        0 -- 1 [pos="40.000000,115.000000 20.000000,115.000000 20.000000,25.000000 100.000000,25.000000 ", type="association"]
	        1 -- 2 [pos="110.000000,25.000000 130.000000,25.000000 ", type="association"]
	        2 -- 3 [pos="135.000000,30.000000 135.000000,50.000000 ", type="association"]
	        ...
	        7 -- 0 [pos="45.000000,90.000000 45.000000,110.000000 ", type="association"]
	        
	        0 -- 8 [pos="50.000000,115.000000 70.000000,115.000000 ", type="association"]
	        0 -- 9 [pos="45.000000,120.000000 45.000000,140.000000 135.000000,140.000000 135.000000,120.000000 ", type="association"]
	        ...
	                
	        1 -- 11 [pos="105.000000,30.000000 105.000000,50.000000 ", type="association"]
	        3 -- 11 [pos="130.000000,55.000000 110.000000,55.000000 ", type="association"]
	        5 -- 11 [pos="105.000000,80.000000 105.000000,60.000000 ", type="association"]
	        7 -- 11 [pos="45.000000,80.000000 45.000000,55.000000 100.000000,55.000000 ", type="association"]
	}

   The vertices in this layout are squares with side length 10, identified by a line containing the string "width".  The edge lines are identified by the 
   string "--".  The construction of incidence_DOT means that the __planar_graph.dot lists the boundary edges first, then the interior edges and finally 
   the infinite edges.
   
   The coordinates in the edge lines describe rectilinear paths connecting the vertices in the planar layout and we use the coordinates of the endpoints of these
   paths to determine the anticlockwise order of the edges around each vertex by considering the angle between the vertex centre and the endpoint.  
   
   Note that the PlanarizationLayout generated by the OGDF orthoganal layout algorithm can sometimes renumber vertices, as in the case of the doodle code 14 crossings 
   8 trigons and 8 4-gons, with interion choice 8 8 8 9 and block 1 28 36 50 57 60 61 62 65 66 67 69.  This example includes place map 1, edge 0 -- 10 but the DOT 
   file produced by OGDF connects 0 to 9 as well as 8.  The full permutation from place_map to DOT in this instance renumbers 9 -> 12, 10 -> 9, 11 -> 10 and 12 -> 11,
*/
bool valid_4_gon_graph(matrix<int>& incidence, int num_boundary_vertices, int num_crossings, int solution_count, string& gauss_code) 
{
	bool valid = true;
	
	int num_vertices = incidence.numrows();
	Graph G;
	ostringstream incidence_DOT;
	incidence_DOT << "graph G {0";
	for (int i=0; i< num_boundary_vertices; i++)
		incidence_DOT << "--" << (i+1)%num_boundary_vertices;
	
	for (int i=0; i<num_boundary_vertices; i++)
	{
		for (int j=num_boundary_vertices; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
			{
				incidence_DOT << ' ' << i << "--" << j;
			}
		}
	}
	
	for (int i=num_boundary_vertices; i< num_vertices; i++)
	{
		for (int j=i+1; j< num_vertices; j++)
		{
			if (incidence[i][j] == 1)
			{
				incidence_DOT << ' ' << i << "--" << j;
			}
		}
	}		
	
	for (int i=0; i<num_boundary_vertices/2; i++)
		incidence_DOT << ' ' << 2*i+1 << "--" << num_vertices;
	
	incidence_DOT << '}';

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: incidence_DOT = " << incidence_DOT.str() << endl;

	istringstream iss(incidence_DOT.str());
	
    GraphAttributes GA(G, GraphAttributes::nodeGraphics | 
                          GraphAttributes::nodeType | 
                          GraphAttributes::edgeGraphics | 
                          GraphAttributes::edgeType);
	  
	if (GraphIO::readDOT(GA, G, iss))
	{
	   for (node v : G.nodes)
	    {
	        GA.width(v) /= 2;
	        GA.height(v) /= 2;
	    }
	 
	    PlanarizationLayout pl;
	    pl.call(GA);
	 		
	    GraphIO::write(GA, "__planar_graph.svg", GraphIO::drawSVG); // used for debugging only
	    GraphIO::write(GA, "__planar_graph.dot", GraphIO::writeDOT);
 
	}
	else
	{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: Error! could not read graph from incidence_DOT" << endl;
		cerr << "Error! valid_4_gon_graph could not read planar graph from DOT data" << endl;
		exit(0);
	}
	
	ifstream dotfile("__planar_graph.dot");
	
	vector<pair<int,int> > edge_pair(2*num_crossings);  // here for the call to doodle_gauss_code
	vector<list<petal> > petal_list(num_vertices+1);
	int num_4_gons_in_disc = num_crossings-num_boundary_vertices/2;
	matrix<int> cycle_vertices(num_4_gons_in_disc,4);
	
	if (dotfile)
	{
		/* We work towards determining a set of flowers for each vertex from which to calculate graph regions.
		   To this end, read the coordinates of the orthogonal layout edges from dotfile, of which there are 2*num_+crossings.  
		*/
		vector<pair<float,float> > vertex(num_vertices+1); // include the infinite vertex
		vector<list<float> > coordinates(2*num_crossings);
		vector<int> num_edge_coordinates(2*num_crossings);
		
		string next_line;
		int edge = 0;
		int v_index = 0;
		while (getline(dotfile,next_line))
		{
			if (next_line.find("width") != string::npos)
			{
				next_line.erase(0,next_line.find("pos")+5);
				next_line.erase(next_line.find("\", type="));
				next_line.replace(next_line.find(","),1," ");
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "valid_4_gon_graph: next truncated vertex line: " << next_line << endl;
	
				istringstream iss(next_line);
				iss >> vertex[v_index].first;
				iss >> vertex[v_index].second;				

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "valid_4_gon_graph:   read vertex " << vertex[v_index].first << ',' << vertex[v_index].second << endl;

				v_index++;
			}
			else if (next_line.find("--") != string::npos)
			{
				next_line.erase(next_line.find("\", type="));

if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "valid_4_gon_graph: next truncated edge line: " << next_line << endl;

				istringstream iss(next_line);
				int initial_vertex;
				iss >> initial_vertex;
				
				char ch;
				do {iss >> ch;} while (ch != '-');
				iss >> ch; // the second '-'
				
				int terminal_vertex;
				iss >> terminal_vertex;
				
				edge_pair[edge] = pair<int,int>(initial_vertex, terminal_vertex);

				do {iss >> ch;} while (ch != '"');
				float coordinate;
				list<float>& clist = coordinates[edge];

				while (iss >> coordinate)
				{	
					clist.push_back(coordinate);
					iss >> ch; // skip  the comma
					iss >> coordinate;
					clist.push_back(coordinate);					
				}

				if (clist.size() % 2 ==1)
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: Error! read an odd numbder of coordinates from __planar_graph.dot for edge pair " << edge_pair[edge].first << " -- " << edge_pair[edge].second << endl;
		cerr << "Error! valid_4_gon_graph read an odd numbder of coordinates from __planar_graph.dot for edge pair " << edge_pair[edge].first << " -- " << edge_pair[edge].second << endl;
		exit(0);
				}
				else
				{
					num_edge_coordinates[edge] = clist.size();
if (debug_control::DEBUG >= debug_control::DETAIL)
	debug << "valid_4_gon_graph: num_edge_coordinates for edge " << edge << ", " << edge_pair[edge].first << "--" << edge_pair[edge].second << " = " << num_edge_coordinates[edge] << endl;
				}
				edge++;
			}
		}
		
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: edge_pair: " << endl;
	for (int i = 0; i< 2* num_crossings; i++)
		debug << "valid_4_gon_graph:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
	
	debug << "valid_4_gon_graph: coordinates: " << endl;
	for (int i = 0; i< 2* num_crossings; i++)
	{
		debug << "valid_4_gon_graph:   " << i << ": ";
		list<float>::iterator cptr = coordinates[i].begin();
		while (cptr != coordinates[i].end())
		{
			debug << setprecision(3) << *cptr << ' ';
			cptr++;
		}
		debug << endl;
	}
		
}		

		/* evaluate the vertex petal lists */
		for (int i=0; i< 2*num_crossings; i++)
		{
			int v1 = edge_pair[i].first;
			int v2 = edge_pair[i].second;
			
			
			list<float>::iterator cptr = coordinates[i].begin();
			float x_coord = *cptr++;
			float y_coord = *cptr;
			
			petal_list[v1].push_back(petal(v2,edge_angle(vertex[v1].first,vertex[v1].second,x_coord,y_coord)));
			
			list<float>::reverse_iterator rptr = coordinates[i].rbegin();
			y_coord = *rptr++;
			x_coord = *rptr;
			
			petal_list[v2].push_back(petal(v1,edge_angle(vertex[v2].first,vertex[v2].second,x_coord,y_coord)));
			
		}
		
		for (int i=0; i< num_vertices+1; i++)
			petal_list[i].sort(comp_petal);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: petal_list" << endl;
	for (int i=0; i< num_vertices+1; i++)
	{
		debug << "valid_4_gon_graph:   " << i << ": ";
		list<petal>::iterator pptr = petal_list[i].begin();
		while (pptr != petal_list[i].end())
		{
			debug << pptr->vertex << ' ';
			pptr++;
		}
		debug << endl;
	}
}		
	
		/* calculate the turning cycles on the disc determined by the flowers at each vertex.  We know the number of cycles in the 
		   disc, since num_4_gons_in_disc = num_crossings - num_edges_in_infinite_cycle = num_crossings-num_boundary_vertices/2.
		   We will only complete the record cycles when they all have length 4.
		   
		   We calculate right turning cycles, since this allows us to move anticlockwise around the vertices on the disc, the same
		   order in which the petals are stored.  Moreover, always turning right means that for an interior edge joining u to v, we
		   traverse the edge in the direction u to v and v to u, so we can use this fact to identify the next edge on which to start a cycle.
		   The use of right turning cycles means that the cycle vertices are recorded in a clckwise direction with respect to the chosen point 
		   at infinity.
		   
		   The interior edges are indicated in edge_pair, corresponding to indices [num_boundary_vertices,...,2*num_crossings), we set
		   edge_pair[i].first to be negative if we have traversed the edge in the direction first->second and set edge_pair[i].second
		   negative if we have traversed the edge in the other direction.
		   
		   By construction, the DOT description of incidence, and therefore edge_pair, records the interior edges originating at vertex zero 
		   as the first interior edges, and the calculation of turning cycles starts with vertex zero.  That means that searching edge_pair 
		   from index num_boundary_vertices for a strictly positive vertex from which to start the next cycle we shall never need to find vertex zero.
		   For example, in the case of doodle code 10 (crossings) 8 4, with interior choice 8 9 9 8 and test block 0 14 21 22, there are three 4-gons
		   involving vertex zero.  Starting at zero we first identiy 0 8 6 7, we (try to) set zero negative to record edge 0--8, then find vertex 8 
		   as the first strictly positive vertex, go on to identify cycle 8 0 9 10, (try to) set 0 negative to record edge 0--9, find vertex 9 as the 
		   first strictly positive vertex and find cycle 9 0 1 2.  Thus we start by working our way around vertex zero, so thereafter looking for a 
		   strictly positive starting vertex means that we don't have to worry about trying to set vertex zero negative.
		*/
		int num_edges = 2*num_crossings;

		int start = edge_pair[num_boundary_vertices].first;		
		edge_pair[num_boundary_vertices].first *= -1;	
		int last_vertex = start;
		int next_vertex = edge_pair[num_boundary_vertices].second;
		int edge_count = 1;
		int cycle_index = 0;
		cycle_vertices[0][0] = 0;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: start vertex " << start << " at edge_pair[" << num_boundary_vertices << "].second, next_vertex = " << next_vertex << endl;
		
		do
		{			
			if (next_vertex == start)
			{
				if (edge_count == 3) // cycle length 3
				{
if (debug_control::DEBUG >= debug_control::SUMMARY)
{
	debug << "valid_4_gon_graph:   right turning cycle involving edge " << last_vertex << " to " << next_vertex << " has length 3" << endl;
	print (incidence,debug, 2,"valid_4_gon_graph: ");
}	
					valid = false;
					break;
				}
				else
				{
					/* here edge_count should be 4, look for a new start in edge_pair */
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:   returned to start vertex " << next_vertex << " edge_count == " << edge_count << endl;
		
					bool found = false;
					
					for (int i=num_boundary_vertices; i< num_edges; i++)
					{
						if (edge_pair[i].first == num_vertices || edge_pair[i].second == num_vertices)
						{
							/* don't want an edge involving the infinite vertex */
							continue;
						}
						else if (edge_pair[i].first > 0 )
						{
							start = edge_pair[i].first;
							edge_pair[i].first *= -1;
							last_vertex = start;
							next_vertex = abs(edge_pair[i].second);
							edge_count = 1;
							found = true;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: found next start vertex " << start << " at edge_pair[" << i << "].first" << endl;
	debug << "valid_4_gon_graph:   last_vertex = " << last_vertex << ", next_vertex = " << next_vertex << " edge_count = " << edge_count << endl;
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].first negative" << endl;
}

							break;
						}
						else if (edge_pair[i].second > 0)
						{
							start = edge_pair[i].second;
							edge_pair[i].second *=-1;
							last_vertex = start;
							next_vertex = abs(edge_pair[i].first);
							edge_count = 1;
							found = true;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: found next start vertex " << start << " at edge_pair[" << i << "].second" << endl;
	debug << "valid_4_gon_graph:   last_vertex = " << last_vertex << ", next_vertex = " << next_vertex << " edge_count = " << edge_count << endl;
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].second negative" << endl;
}

							break;
						}						
					}

if (debug_control::DEBUG >= debug_control::DETAIL)
{
	debug << "valid_4_gon_graph: adjusted edge_pair: " << endl;
	for (int i = 0; i< 2* num_crossings; i++)
		debug << "valid_4_gon_graph:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}
					
					if (found)
					{
						cycle_index++;
						cycle_vertices[cycle_index][0] = start;
					}
					else
					{
						break;
					}
				}
			}
			else
			{
				if (edge_count == 4)  // cycle length greater than 4
				{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:   right turning cycle involving edge " << last_vertex << " to " << next_vertex << " has length greater than 4" << endl;
					valid = false;
					break;
				}
				else
				{
					cycle_vertices[cycle_index][edge_count] = next_vertex;
					
					edge_count++;
					
					/* turn right at next_vertex and set next_vertex in edge_pair negative */
					list<petal>::iterator pptr = find_if(petal_list[next_vertex].begin(),petal_list[next_vertex].end(),vertex_equals(last_vertex));
					if (pptr == petal_list[next_vertex].end())
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: Error! could not find " << last_vertex << " in the petal list of vertex " << next_vertex << endl;
						cerr << "Error! valid_4_gon_graph could not find " << last_vertex << " in the petal list of vertex " << next_vertex << endl;
						exit(0);
					}
					else
					{
						pptr++;
						if (pptr == petal_list[next_vertex].end())
							pptr = petal_list[next_vertex].begin();
														
						last_vertex = next_vertex;
						next_vertex = pptr->vertex;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:   last_vertex = " << last_vertex << ", next_vertex " << next_vertex << " edge_count = " << edge_count << endl;
						
						/* find the edge last_vertex -- next_vertex in edge pair and set last_vertex negative */
						for (int i=num_boundary_vertices; i< num_edges; i++)
						{
							if (edge_pair[i].first == last_vertex && abs(edge_pair[i].second) == next_vertex)
							{
								edge_pair[i].first *= -1;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].first negative" << endl;
								break;
							}
							else if (edge_pair[i].second == last_vertex && abs(edge_pair[i].first) == next_vertex)
							{
								edge_pair[i].second *= -1;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph:     interior edge " << last_vertex << " to " << next_vertex << ", set edge_pair[" << i << "].second negative" << endl;
								break;
							}
						}
					}
				}
			}			
		} while (valid);		

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "valid_4_gon_graph: cycle:" << endl;
	for (int i=0; i<num_4_gons_in_disc; i++)
	{
		debug << "valid_4_gon_graph:   " << i+1 << ": ";
		for (int j=0; j< 4; j++)
			debug << setw(3) << cycle_vertices[i][j];
		debug << endl;
	}
	
}					
	}
	else
	{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: Error! could not open DOT file __planar_graph.dot" << endl;
		cerr << "Error! valid_4_gon_graph could not open DOT file __planar_graph.dot" << endl;
		exit(0);
	}	

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "valid_4_gon_graph: returning valid = " << valid << endl;
	
	if (DRAW_GRAPHS && valid)
	{
		string filename = "__planar_graph-"+util::itos(solution_count+1)+".svg";		
	    GraphIO::write(GA, filename, GraphIO::drawSVG); 
	}
	
	if (valid)
	{		
		gauss_code = doodle_gauss_code(edge_pair,cycle_vertices,2*num_crossings,num_boundary_vertices);
	}
	
	return valid;
}

/* edge_angle returns the amplitude of the vector from the vertex at (v_x,v_y) to the start of the edge at (e_x,e_y) */
float edge_angle(float v_x, float v_y, float e_x, float e_y)
{
	double pi = 3.14159265358979;
	
	float x_delta = e_x - v_x;
	float y_delta = e_y - v_y;
	
	double angle = atan2(y_delta,x_delta);
	if (angle < 0)
		angle += 2*pi;
		
	return angle*360/2/pi;
}

/* doodle_gauss_code produces a Gauss code for the doodle determined by the planar dual graph described by edge_pair and cycle_vertices.
   Each of the 4-gon regions of the dual graph (including the infinite 4-gon regions outside the disc) correspond to a crossing, so we
   first identify the edges corresponding to each region, recording them in a clockwise manner with respect to the chosen point at infinity
   and then we can trace the components of the doodle, as the strands across each 4-gon join opposite edges, forming a cross.
*/
string doodle_gauss_code(vector<pair<int,int> >& edge_pair,matrix<int>& cycle_vertices, int num_edges, int num_boundary_edges)
{
	string left_preferred = "";
	
	int num_edges_in_infinite_cycle = num_boundary_edges/2;
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: presented with edge_pair: " << endl;
	for (int i = 0; i< num_edges; i++)
		debug << "doodle_gauss_code:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}

	/* identify the region edges, where the edges are numbered according to the order in edge_pair.  The boundary edges are listed
	   first, then the interior edges, then the edges to the infinite vertex.  We reset the sign of the interior edge pairs since the 
	   entries will be negative after valid_4_gon_graph.
	*/
	for (int i= num_boundary_edges; i< num_edges-num_edges_in_infinite_cycle; i++)
	{
		edge_pair[i].first *= -1;
		edge_pair[i].second *= -1;
	}
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: presented with edge_pair: " << endl;
	for (int i = 0; i< num_edges; i++)
		debug << "doodle_gauss_code:   " << i << ": " << edge_pair[i].first << ", " << edge_pair[i].second <<  endl;	
}

	/* calculate the region edges, where the edges are numbered by the corresponding index in edge_pair. The region edges will be 
	   recorded clockwise, since that is the orientation in whihc the cycle_vertices are recorded. 
	*/
	int num_4_gons_in_disc = (num_edges - num_boundary_edges)/2;
	matrix<int> region_edges(num_edges/2,4);
	
	for (int i=0; i< num_4_gons_in_disc; i++)
	{
		for (int j=0; j< 4; j++)
		{
			int v1 = cycle_vertices[i][j];
			int v2 = cycle_vertices[i][(j+1)%4];
			
			for (int k=0; k< num_edges; k++)
			{
				if ( (edge_pair[k].first == v1 && edge_pair[k].second == v2) || (edge_pair[k].second == v1 && edge_pair[k].first == v2) )
				{
					region_edges[i][j] = k;
					break;
				}
			}
		}
	}
	
	/* Add the infinite regions to region_edges in the same clockwise manner used for the interior regions */
	for (int i=0; i< num_edges_in_infinite_cycle; i++)
	{
		region_edges[num_4_gons_in_disc+i][0] = 2*i;
		region_edges[num_4_gons_in_disc+i][1] = (2*i-1+num_boundary_edges)%num_boundary_edges;
		region_edges[num_4_gons_in_disc+i][2] = num_edges - num_edges_in_infinite_cycle +(i-1+num_edges_in_infinite_cycle )%num_edges_in_infinite_cycle;
		region_edges[num_4_gons_in_disc+i][3] = num_edges - num_edges_in_infinite_cycle + i;
	}
	
	
	
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: region_edges: " << endl;
	for (int i = 0; i< num_edges/2; i++)
	{
		debug << "doodle_gauss_code:   " << i << ": ";	
		for (int j=0; j< 4; j++)
			debug << setw(3) << region_edges[i][j];
		debug << endl;
	}
}

	/* trace the region edges to build up a list of the consecutive edges encountered as we traverse the doodle.  Each interior edge appears twice in 
	   region_edges and each boundary edge apears once.  We always traverse regions by leaving on the opposite edge (i.e i <-->3 and 2 <--> 4).  If we 
	   hit the disc boundary we have to traverse an "infinite 4-gon" and cross an infinite edge.
	   
	   If we are laving the disc on an odd numbered edge, we cross the suceeding infinite edge and re-enter the disc three boundary edges along.
	   If we are laving the disc on an even numbered edge, we cross the preceding infinite edge and re-enter the disc three boundary edges back.
	   
	   There are num_boundary_vertices/2 infinite edges and they appear at the end of edge_pair, attached to vertices 1,3,...,num_boundary_vertices-1.
	   Consider the infinite edges numbered from zero. Then, the succeeding infinite edge to edge 2i+1 is the (i+1)-st infinite edge and the preceding 
	   infinite edge to edge 2i is the (i-1)-st infinite edge, for i=0,...
	*/
	list<list<int> > component_edges;
	list<list<int> > component_regions;
	
	vector<bool> edge_flag(num_edges); // recoords whether we have visited the edge, initializes to false	
	bool finished = false;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: search for components:" << endl;
	
	do
	{
		/* look for the start of another component */
		finished = true;
		int start = -1;
		for (int i=0; i< num_edges; i++)
		{
			if (edge_flag[i] == false)
			{
				finished = false;
				start = i;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: found start of component at edge " << start;
				break;
			}
		}
		
		if (!finished)
		{
			list<int> next_component_edge;
			list<int> next_component_region;
			edge_flag[start] = true;
			next_component_edge.push_back(start);
			int region = -1;  // records the region we crossed last
			int next_edge = succeeding_edge(region_edges,num_4_gons_in_disc, start,region);
			next_component_region.push_back(region);

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << ", next_edge = " << next_edge << " in region " << region << endl;
					
			/* trace the component */
			bool complete = false;
			do
			{				
				if (next_edge == start)
				{				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "\ndoodle_gauss_code:   next_edge = start edge " << start << " component complete" << endl;
		
					complete = true;
				}
				else 
				{
					edge_flag[next_edge] = true;
					next_component_edge.push_back(next_edge);  
					
					if (next_edge < num_boundary_edges)
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   edge " << next_edge << " is on the disc boundary" << endl;
			
						if (next_edge%2 == 1)
						{
							int infinite_edge = num_edges - num_edges_in_infinite_cycle + (next_edge/2 +1)%num_edges_in_infinite_cycle;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     infinite_edge = " << infinite_edge << endl;
			
							edge_flag[infinite_edge] = true;
							next_component_edge.push_back(infinite_edge);
							next_component_region.push_back(((next_edge+1)%num_boundary_edges)/2+num_4_gons_in_disc);
							
							next_edge = (next_edge+3)%num_boundary_edges;
							next_component_region.push_back(next_edge/2+num_4_gons_in_disc);
							
							if (next_edge == start)
							{				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   next_edge = start edge " << start << " component complete" << endl;
			
								complete = true;
							}
							else
							{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next boundary edge = " << next_edge << endl;
								edge_flag[next_edge] = true;
								next_component_edge.push_back(next_edge);
								region = -1; // re-entering the disc, want to check all regions
								next_edge = succeeding_edge(region_edges,num_4_gons_in_disc,next_edge,region);
								next_component_region.push_back(region);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next interior edge = " << next_edge << " in region " << region << endl;
							}
						}
						else
						{
							int infinite_edge = num_edges - num_edges_in_infinite_cycle +(next_edge/2 - 1 + num_edges_in_infinite_cycle )%num_edges_in_infinite_cycle;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     infinite_edge = " << infinite_edge << endl;
	
							edge_flag[infinite_edge] = true;
							next_component_edge.push_back(infinite_edge);
							next_component_region.push_back(next_edge/2+num_4_gons_in_disc);
							next_edge = (next_edge-3+num_boundary_edges)%num_boundary_edges;
							next_component_region.push_back(((next_edge+1)%num_boundary_edges)/2+num_4_gons_in_disc);
		
							if (next_edge == start)
							{				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   next_edge = start edge " << start << " component complete" << endl;
			
								complete = true;
							}
							else
							{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next boundary edge = " << next_edge << endl;
								edge_flag[next_edge] = true;
								next_component_edge.push_back(next_edge);
								region = -1; // re-entering the disc, want to check all regions
								next_edge = succeeding_edge(region_edges,num_4_gons_in_disc,next_edge,region);
								next_component_region.push_back(region);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:     next interior edge = " << next_edge << " in region " << region << endl;
							}
						}
					}
					else
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   edge " << next_edge << " is in the disc interior in region " << region;
						/* move across the region to the next edge */
						next_edge = succeeding_edge(region_edges,num_4_gons_in_disc,next_edge,region);
						next_component_region.push_back(region);
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << ", next_edge  = " << next_edge << " in region " << region << endl;
					}
				}
			} while(!complete);
	
			component_edges.push_back(next_component_edge);
			component_regions.push_back(next_component_region);
		}
	} while (!finished);
	
	int num_components = component_edges.size();
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code: num_components = " << num_components << endl;
	debug << "doodle_gauss_code: component edges: " << endl;
	
	list<list<int> >::iterator llptr = component_edges.begin();
	int index=0;
	
	while (llptr != component_edges.end())
	{
		debug << "doodle_gauss_code:   " << index++ << ": ";
		
		list<int>::iterator lptr = (*llptr).begin();
		while (lptr != (*llptr).end())
		{
			debug << setw(3) << *lptr;
			lptr++;
		}
		
		debug << endl;
		llptr++;
	}

	debug << "doodle_gauss_code: component regions: " << endl;
	
	llptr = component_regions.begin();
	index=0;

	while (llptr != component_regions.end())
	{
		debug << "doodle_gauss_code:   " << index++ << ": ";
		
		list<int>::iterator lptr = (*llptr).begin();
		while (lptr != (*llptr).end())
		{
			debug << setw(3) << *lptr;
			lptr++;
		}
		
		debug << endl;
		llptr++;
	}
	
}

	/* Write the doodle Gauss code to a string. */
	ostringstream oss;
	list<list<int> >::iterator elptr = component_edges.begin();
	list<list<int> >::iterator rlptr = component_regions.begin();
	unsigned int component = 1;
	
	while (elptr != component_edges.end())
	{	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: write component " << distance(component_edges.begin(),elptr) << endl;

		list<int>::iterator eptr = (*elptr).begin();
		list<int>::iterator rptr = (*rlptr).begin();
		
		unsigned int term = 1;
		
		while (eptr != (*elptr).end())
		{			
			/* determine whether the strand transverse to e1 = *eptr -- *(eptr+1) = e2 is a left or right strand.  We use the fact that the 
			   region_edges are recorded clockwise, so each region has the form shown below, where n is the region(crossing number):
			   
			                               b
			                            +------+
			                            |      |
			                         a  |  n   | c
			                            +------+
			                               d
			                               
			   the strand we are following is oriented from e1 to e2, so given the transvere strand, oriented from t1 to t2 we have:

			   reverse_e_edges = false			   
			   e1--e2 = a--c    t1--t2 = b--d  => Ln      t1--t2 = d--b  => Rn  
			   e1--e2 = b--d    t1--t2 = a--c  => Rn      t1--t2 = c--a  => Ln  

			   reverse_e_edges = true
			   e1--e2 = c--a    t1--t2 = b--d  => Rn      t1--t2 = d--b  => Ln  
			   e1--e2 = d--b    t1--t2 = a--c  => Ln      t1--t2 = c--a  => Rn  
			   
			   These eight cases account for all possible cyclic permutations of a b c and d in region_edges, since they correspond to a 
			   rotation of the above diagram.
			*/			
			int e1 = *eptr;
			int e2;
			list<int>::iterator next_eptr = eptr;
			next_eptr++;
			if (next_eptr == (*elptr).end())
				e2 = *(*elptr).begin();
			else
				e2 = *next_eptr;
				
			int region = *rptr;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: edges " << e1 << " and " << e2 << ", region " << region << endl;
	
			bool reverse_e_edges = false;
			bool reverse_t_edges = false;
			bool e_edges_position_a_c = false;
			bool left_transverse = false;
		
			int t1=-1;
			int t2=-1;
			
			for (int i=0; i< 4; i++)
			{
				if (region_edges[region][i] == e1)
				{
					if (i/2 == 1)
						reverse_e_edges = true;
					else
						reverse_e_edges = false;
						
					if (i%2 == 0)
						e_edges_position_a_c = true;						
					else
						e_edges_position_a_c = false;
				}
				else if (region_edges[region][i] != e2)
				{
					if (t1 == -1)
					{
						t1 = region_edges[region][i];

						if (i/2 == 1)
							reverse_t_edges = true;
						else
							reverse_t_edges = false;
					}
					else
					{
						t2 = region_edges[region][i];
					}
				}
			}
				
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code:   transverse edges " << t1 << " and " << t2 << endl;
	debug << "doodle_gauss_code:   e_edges_position_a_c = " << e_edges_position_a_c << " reverse_e_edges = " << reverse_e_edges << " reverse_t_edges = " << reverse_t_edges << endl;
}				

			/* order_transverse_edges ensures that t1 preceeds t2 in the orientation implicit in component_edges,
			   re-ordering their values if necessary.  Look for t1 in component_edges, each edge appears exactly once, 
			   then look to see whether t1 appears in front or after t1
			*/
			
			list<list<int> >::iterator llptr = component_edges.begin();
			bool found_t1 = false;
			int index = 0;
			while (llptr != component_edges.end())
			{			
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   looking for transverse edges " << t1 << " and " << t2 << " in component  " << index++ << endl;
					
				list<int>::iterator lptr = (*llptr).begin();
				while (lptr != (*llptr).end())
				{
					if (*lptr == t1)
					{
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   found at offset " << distance((*llptr).begin(),lptr) << endl;
	
						found_t1 = true;
						list<int>::iterator next_ptr = lptr;
						list<int>::iterator prev_ptr = lptr;
						
						if (prev_ptr == (*llptr).begin())
							prev_ptr = (*llptr).end();

						prev_ptr--;
							
						next_ptr++;
						if (next_ptr == (*llptr).end())
							next_ptr = (*llptr).begin();

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   prev_ptr offset " << distance((*llptr).begin(),prev_ptr) << " value " << *prev_ptr << " next_ptr offset " << distance((*llptr).begin(),next_ptr) << " value " << *next_ptr << endl;
						
						if (*prev_ptr == t2)
						{
							swap(t1, t2);
							reverse_t_edges = !reverse_t_edges;
							
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   transverse edges found in reverse order in component_edges, new t1 = " << t1 << ", new t2 = " << t2 << " new reverse_t_edges = " << reverse_t_edges << endl;

						}
						else if (*next_ptr != t2)
						{
							cerr << "Error! Transverse edges " << t1 << " and " << t2 << " are not consecutive in component_edges, aborting." << endl;
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code: Error! transverse edges " << t1 << " and " << t2 << " are not consecutive in component_edges" << endl;
							exit(0);
						}
						
						break;
					}
					
					lptr++;
				}
								
				if (found_t1)
					break;
				else
					llptr++;
			}

			if ((reverse_e_edges && e_edges_position_a_c & reverse_t_edges) || 
			    (reverse_e_edges && !e_edges_position_a_c & !reverse_t_edges) || 
			    (!reverse_e_edges && e_edges_position_a_c && !reverse_t_edges) ||
			    (!reverse_e_edges && !e_edges_position_a_c && reverse_t_edges)
			   )
				left_transverse = true;

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
{
	debug << "doodle_gauss_code:   transverse edges " << t1 << " and " << t2 << endl;
	debug << "doodle_gauss_code:   e_edges_position_a_c = " << e_edges_position_a_c << " reverse_e_edges = " << reverse_e_edges << " reverse_t_edges = " << reverse_t_edges << endl;
	debug << "doodle_gauss_code:   left_transverse = " << left_transverse << endl;
}				
			
			if(left_transverse)
				oss << 'L';
			else
				oss << 'R';
				
			oss << region+1;  // Gauss code crossing snumbered from 1.
			
			if (term < (*elptr).size())
				oss << ' ';
				
			term++;
			
			eptr++;
			rptr++;
		}
		
		if (component < component_edges.size())
			oss << ',';

		component++;
		
		elptr++;
		rlptr++;
	}

	ostringstream code_trailer;
	
	code_trailer << '/';
	
	for (int i=0; i< num_edges/2; i++)
	{
		code_trailer << '#';
		if (i < num_edges/2 - 1)
			code_trailer << ' ';
	}
	
if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   gauss code " << oss.str() << code_trailer.str() << endl;

	ofstream tmp("tmpfile");
	tmp << "[raw-output,silent,gauss,ulpgd]" << endl;
	tmp << oss.str() << code_trailer.str() << endl;
	
	system("braid tmpfile");
	ifstream braid("braid.out");
	
	if (!getline(braid, left_preferred))
	{
		cerr << "\nError! Braid programe failed to calculate left_preferred Gauss data." << endl;
		exit(0);
	}
	braid.close();

if (debug_control::DEBUG >= debug_control::INTERMEDIATE)
	debug << "doodle_gauss_code:   left preferred gauss code " << left_preferred << code_trailer.str() << endl;

	return left_preferred + code_trailer.str();
}

int succeeding_edge(matrix<int>& region_edges, int num_4_gons_in_disc, int edge, int& region)
{
	int next_edge = -1;
	
	for (int i=0; i< num_4_gons_in_disc; i++)
	{
		if (i == region)
			continue;
		
		bool found = false;
			
		for (int j=0; j<4; j++)
		{
			if (region_edges[i][j] == edge)
			{
				next_edge = region_edges[i][(j+2)%4]; // 0 <-> 2, 1 <-> 3
				region = i;
				found = true;
				break;
			}
		}
		
		if (found)
			break;
	}
	
	return next_edge;
}
